#include <boost/timer.hpp>

#include <symbolic/rules.h>
#include <symbolic/predicates.h>
#include "symbolic/rules.h"

#include "rexd/excuses.h"
#include "rexd/planner_factory.h"
#include "rexd/pasula_learner.h"



#define LOGA(Level_, Logger_, ...)   \
    while (Log::Enabled(Level_)) {  \
      Log_Write(Level_, Logger_, __func__, __FILE__, __LINE__, __VA_ARGS__);  \
      break;  \
    }


void Log_Write(uint level,
               char const* logger,
               char const* function,
               char const* file,
               int line,
               std::string msg)
{
    std::cout << msg;
}


void compare_rules(char* rules1_path, char* rules2_path) {
    RuleSet rules1(rules1_path);
    RuleSet rules2(rules2_path);

    rules1.remove_nop_rules();
    rules2.remove_nop_rules();
//     LOG(INFO) << "Rules1: " << rules1;
//     LOG(INFO) << "Rules2: " << rules2;
    std::sort (rules1.begin(), rules1.end());
    std::sort (rules2.begin(), rules2.end());

//     LOG(INFO) << "Rules1: " << rules1;
//     LOG(INFO) << "Rules2: " << rules2;

    if (rules1 < rules2) {
        std::cout << START_COLOR_RED << "Rules are different!!! rules1 > rules2" << END_COLOR << std::endl;
        std::cout << "Differences: \n" << rules1.get_different_rules(rules2) << std::endl;
    }
    else if (rules2 < rules1) {
        std::cout << START_COLOR_RED << "Rules are different!!! rules2 > rules1" << END_COLOR << std::endl;
        std::cout << "Differences: \n" << rules1.get_different_rules(rules2) << std::endl;
    }
    else {
//         std::cout << "Rules are equal!" << std::endl;
    }
}


void compare_rules_score(char* rules1_path, char* rules2_path) {
    RuleSet rules1(rules1_path);
    RuleSet rules2(rules2_path);
    TransitionList transitions = read_transitions_from_file("transitions.dat");

    float score = rules1.score(transitions);
    LOG(INFO)  << "Rules 1 score: " << score;

    score = rules2.score(transitions);
    LOG(INFO)  << "Rules 2 score: " << score;
}


void learn(std::string action_string = "")
{
    struct timeval timer_start = utils::start_timer();
    LOG(INFO) << "Learning";
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);
    
    LearnerPtr learner = boost::make_shared<PasulaLearner>(PasulaLearner(config_reader));
    DomainPtr domain_ptr = DomainPtr(new Domain);
    TransitionGroup transitions = read_transitions_from_file(config_reader.get_path("transitions_backup_path"));
    
    if ((config_reader.get_variable<bool>("learning")) && (transitions.size() > 0) ) {
        CINFO("planning") << START_COLOR_BLUE << "Learning" << END_COLOR;
        RuleSetPtr removed_rules_ptr, added_rules_ptr;
        RuleSetPtr rules_ptr;
        if (action_string == "") {
            rules_ptr = learner->learn_action_rules(*rules_ptr, transitions, Symbol(action_string), removed_rules_ptr, added_rules_ptr);
        }
        else {
            rules_ptr = learner->learn(transitions);
        }
        rules_ptr->write_rules_to_file(config_reader.get_path("nid_rules_path"));
        rules_ptr->index(*domain_ptr);
    }
    else {
        CINFO("planning") << "Skipping learning (no transitions or no previous action)";
    }
    LOG(INFO) << "Time taken: " << utils::get_elapsed_time(timer_start);
}


void plan()
{
    struct timeval timer_start = utils::start_timer();

    LOG(INFO) << "Planning";
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);

    RuleSet rules(config_reader.get_path("nid_rules_path"));
    GoalGroup goals(config_reader.get_path("goal_file"));
    PlannerPtr planner = PlannerFactory::get_standard_planner(config_reader);

    State state;
    state.read_from_file(config_reader.get_path("simulator_current_state_path"));

//     float score;
//     SymbolPtr new_res = planner->plan(state, rules, goals, score, true, true);
    bool reached_goal;
    Scenario scenario(rules, state, goals, config_reader);
    std::vector<Symbol> new_res = planner->plan_n_actions(scenario, 8, reached_goal, true, true);
    LOG(INFO) << "Time taken: " << utils::get_elapsed_time(timer_start);
}

void explore()
{
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);

    RuleSet rules(config_reader.get_path("nid_rules_path"));
    PlannerPtr planner = PlannerFactory::get_standard_planner(config_reader);

    State state;
    state.read_from_file(config_reader.get_path("simulator_current_state_path"));
    TransitionGroup transitions = read_transitions_from_file("transitions_full.dat");

    std::vector<Symbol> action_list;
    if (!config_reader.get_variable<bool>("teacher")) {
        LOG(ERROR) << "TODO!!!!";
//         read_action_list("symbols.dat");
    }
    else {
        action_list = rules.get_actions();
    }
    Scenario scenario(rules, state, GoalGroup(), config_reader, action_list);
    Explorer explorer(scenario, boost::make_shared<TransitionGroup>(transitions));
    explorer.is_state_known();
}


SymbolOptional plan_with_subgoals(const RuleSet& rules, const State& state, ExcuseGenerator& excuse_generator, const PredicateGroup& excuse, PlannerPtr planner, const Goal& goal, const ConfigReader& config_reader)
{
    // TODO plan with more than 1 subgoals
    Goal new_goal = excuse_generator.get_subgoal(goal, excuse);
    GoalGroup new_goals;
    new_goals.push_back(new_goal);
    CINFO("planning") << "Planning using subgoals with excuse " << excuse;
    CINFO("planning") << "New goal is " << new_goal;
    Scenario scenario(rules, state, new_goals, config_reader);
    SymbolOptional result = planner->plan(scenario, false);
    return result;
}


void generate_excuses()
{
    std::string config_path = "rexd_config.cfg";
    ConfigReader config_reader;
    config_reader.init(config_path);
    RuleSet rules(config_reader.get_path("nid_rules_path"));
    PlannerPtr planner = PlannerFactory::get_standard_planner(config_reader);

    State state;
    state.read_from_file(config_reader.get_path("simulator_current_state_path"));
    GoalGroup goals = GoalGroup(config_reader.get_path("goal_file"));

    std::vector< Symbol > planning_res;
    RuleSet grounded_rules = rules.get_full_grounded_rules(state);
    grounded_rules.remove_nop_rules();
    LOG(INFO) << "************************* GENERATING EXCUSES *************************";

    Scenario scenario(rules, state, goals, config_reader);
    ExcuseGenerator excuse_generator(scenario);

    PredicateGroup excuse = excuse_generator.get_excuse();
    LOG(INFO)  << "Best excuse is " << excuse;
    
    PredicateGroup symbolic_excuse = excuse.get_symbolic_predicates();
    std::vector<std::string> actions_requiring_excuse = rules.find_rules_with_a_precondition(excuse).get_action_names();

    LOG(INFO) << "Actions requiring excuse: " << actions_requiring_excuse;

    SymbolOptional action_opt;
    if (config_reader.get_variable<bool>("rule_analysis_equivalent_rules")) {
        CINFO("excuses") << "Looking for equivalent rules!";
        TransitionGroup transitions = read_transitions_from_file("transitions_full.dat");
//         float expected_reward;
//         action_opt = excuse_generator.search_for_equivalent_useful_rules(excuse, transitions, boost::make_shared<RuleSet>(rules), expected_reward);
        // TODO
        LOG(ERROR) << "TODO " << "Rule alternatives";
    }

    // predicate subgoals are incompatible with vmin
    if ((config_reader.get_variable<bool>("rule_analysis_subgoals")) && (!config_reader.get_variable<bool>("use_rmax")) ) {
        CINFO("excuses") << "Trying subgoals";
        action_opt = plan_with_subgoals(rules, state, excuse_generator, excuse, planner, goals.get_predicates(), config_reader);
    }
}


void print_help()
{
    std::cout << "Usage: rexd_utils [option] [args]" << std::endl
              << "Options:" << std::endl
              << "\texcuses \t\t- Plans with excuses using current folder files" << std::endl
              << "\tplan \t\t\t- plans using current folder files" << std::endl
              << "\texplore \t\t- explores using current folder files" << std::endl
              << "\trules <rules1> <rules2>\t- compares both rule sets" << std::endl
              << "\tscore <rules1> <rules2>\t- gets score of both rule sets" << std::endl;
}


int main(int argc, char** argv) {
    // generated random seed
    srand((unsigned)time(0));

    if (argc < 2) {
        print_help();
    }
    else {
        std::string aux(argv[1]);
        if (aux == "excuses") {
            utils::init_loggers("generic/logger.conf");
            generate_excuses();
        }
        else if (aux == "plan") {
            utils::init_loggers("generic/logger.conf");
            plan();
        }
        else if (aux == "learn") {
            utils::init_loggers("generic/logger.conf");
            if (argc > 2) {
                learn(argv[2]);
            }
            else {
                learn();
            }
        }
        else if (aux == "explore") {
            utils::init_loggers("generic/logger.conf");
            explore();
        }
        else if (aux == "rules") {
            compare_rules(argv[2], argv[3]);
        }
        else if (aux == "score") {
            compare_rules_score(argv[2], argv[3]);
        }
        else {
            print_help();
        }
    }

    return 0;
}
