/*
 * Various code tests
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

// external
#include <gtest/gtest.h>
#include <boost/assign/std/vector.hpp>
#include <boost/assign/list_inserter.hpp>

// rexd
#include "symbolic/predicates.h"
#include "symbolic/rules.h"
#include "symbolic/state.h"
#include "rexd/rexd_scenario.h"
#include "rexd/pasula_learner.h"
#include <rexd/high_level_planner.h>

// helpers
#include "test_helpers.cpp"
 
TEST(RulesTest, GroundingRules) { 
    PredicateGroup predicates("clear(62), -clear(63), -on(62,71), -onTable()");
    RuleSet symbolic_rules("rules_tests/rules_symbolic_basic.dat");
    RuleSet grounded_rules("rules_tests/rules_grounded_basic.dat");
//     RuleSet grounded_symbolic_rules = symbolic_rules.get_grounded_rules(predicates);
    
    EXPECT_EQ(grounded_rules, symbolic_rules.get_grounded_rules(predicates)) 
        << "Grounding symbolic rules without deictic references failed.";
}
 
TEST(RulesTest, DeicticGroundingRules) {
    PredicateGroup predicates("clear(62), -clear(63), on(71,63), on(72,62), on(72,63), -onTable()");
    RuleSet symbolic_rules("rules_tests/rules_symbolic_deictic.dat");
    RuleSet grounded_rules("rules_tests/rules_grounded_deictic.dat");
//     RuleSet symbolic_rules_grounded = symbolic_rules.get_grounded_rules(predicates);
//     std::sort(symbolic_rules_grounded.begin(), symbolic_rules_grounded.end());
//     std::sort(grounded_rules.begin(), grounded_rules.end());
//     
//     if (symbolic_rules_grounded != grounded_rules) {
//         std::cout << "Rules are different!" << std::endl;
//         std::cout << symbolic_rules_grounded << std::endl;
//         std::cout << "--------------------" << std::endl;
//         std::cout << grounded_rules << std::endl;
//     }
    
    EXPECT_EQ(grounded_rules, symbolic_rules.get_grounded_rules(predicates))
        << "Grounding symbolic rules with deictic references failed.";
}

TEST(DomainTest, GeneratingDomain) {
    ConfigReader config_reader;
    config_reader.init("domain_tests/rexd_config.cfg");
    State state;
    state.read_from_file("domain_tests/domain_state.dat");
    RuleSet rules("domain_tests/domain_rules.dat");
    GoalGroup goals("domain_tests/domain_goals.dat");
    Scenario scenario(rules, state, goals, config_reader);
    std::stringstream ss;
    scenario.domain.write_NID(ss);
    utils::write_string_to_file("generated/symbols.dat", ss.str());
    EXPECT_EQ(compare_two_files("domain_tests/domain_symbols.dat", "generated/symbols.dat"), true)
        << "Domain does not create expected symbols file.";
}

TEST(PPDDLTest, GeneratingPPDDLRules) {
    ConfigReader config_reader;
    config_reader.init("domain_tests/rexd_config.cfg");
    State state;
    state.read_from_file("domain_tests/domain_state.dat");
    RuleSet rules("domain_tests/domain_rules.dat");
    GoalGroup goals("domain_tests/domain_goals.dat");
    RexdScenario rexd_scenario(Scenario(rules, state, goals, config_reader));
    
    rexd_scenario.write_PPDDL_to_file("./domain_tests/template_ppddl", "generated/ppddl_rules.ppddl");
    
    EXPECT_EQ(compare_two_files("domain_tests/ppddl_rules.ppddl", "generated/ppddl_rules.ppddl"), true)
        << "Error generating PPDDL rules file.";
}

TEST(ExploreTest, CheckKnown) 
{
    utils::init_loggers("generic/logger.conf");
    
    using namespace boost::assign;
    typedef std::map<std::string, bool> FolderActionBoolPairs;
    FolderActionBoolPairs test_folders; 
    insert( test_folders )
        ( "empty_known"     , false )
        ( "explored"        , false )
        ( "1action_unknown" , true  )
        ( "almost_explored" , true  )
        ( "almost_explored2", true  )
        ( "almost_explored3", true  );
    
    BOOST_FOREACH(FolderActionBoolPairs::value_type folder_action, test_folders) {
        // load config
        std::string test_folder_full = "exploration_tests/" + folder_action.first + "/";
        ConfigReader config_reader;
        config_reader.init(test_folder_full + "/rexd_config.cfg", test_folder_full);
        
        // load high_level_planner
        std::ofstream logging_file;
        logging_file.open ("/dev/null", std::ios::app);
        HighLevelPlanner high_level_planner(config_reader, &logging_file);
        PredicateGroup excuse;
        std::vector<std::string> actions_requiring_excuse;
        TransitionGroup transitions = read_transitions_from_file(config_reader.get_path("transitions_backup_path"));
        Scenario scenario = load_scenario(config_reader);
        
        // test exploration without forced flag
        SymbolOptional explore_action = high_level_planner.explore(scenario, transitions, false);
        EXPECT_EQ(explore_action.is_initialized(), folder_action.second)
            << "State should be known (no action instead of " << *explore_action << "). Test: " << folder_action.first;
        
        // test exploration with forced flag
        if (scenario.get_rules().empty()) { 
            explore_action = high_level_planner.explore(scenario, transitions, true);
            SymbolOptional ground_truth_action(Symbol("placeshaft()"));
            EXPECT_EQ(explore_action.is_initialized(), false)
                << "State should be known (no action instead of " << *explore_action << "). Test: " << folder_action.first;
        }
        else {
            explore_action = high_level_planner.explore(scenario, transitions, true);
            SymbolOptional ground_truth_action(Symbol("placeshaft()"));
            EXPECT_EQ(explore_action.is_initialized(), true)
                << "A exploration action should have been selected (because forced flag). Test: " << folder_action.first;
        }
    }
}

TEST(ExploreTest, ActionSelection) 
{
    utils::init_loggers("generic/logger.conf");
    
    using namespace boost::assign;
    typedef std::map<std::string, SymbolOptional> FolderActionPairs;
    FolderActionPairs test_folders; 
    insert( test_folders )
        ( "1action_unknown" , SymbolOptional(Symbol("placeshaft()"))       )
        ( "almost_explored" , SymbolOptional(Symbol("screwpendulumhead()")))
        ( "almost_explored2", SymbolOptional(Symbol("placependulum()"))    )
        ( "almost_explored3", SymbolOptional(Symbol("placepeg4()"))        );
    
    BOOST_FOREACH(FolderActionPairs::value_type folder_action, test_folders) {
        // load config
        std::string test_folder_full = "exploration_tests/" + folder_action.first + "/";
        ConfigReader config_reader;
        config_reader.init(test_folder_full + "/rexd_config.cfg", test_folder_full);
        
        // load high_level_planner
        std::ofstream logging_file;
        logging_file.open ("/dev/null", std::ios::app);
        HighLevelPlanner high_level_planner(config_reader, &logging_file);
        PredicateGroup excuse;
        std::vector<std::string> actions_requiring_excuse;
        TransitionGroup transitions = read_transitions_from_file(config_reader.get_path("transitions_backup_path"));
        Scenario scenario = load_scenario(config_reader);
        
        // test exploration without forced flag
        SymbolOptional explore_action = high_level_planner.explore(scenario, transitions, false);
        SymbolOptional ground_truth_action(Symbol("placeshaft()"));
        EXPECT_EQ(explore_action.is_initialized(), true)
                << "A exploration action should have been selected. Test: " << folder_action.first;
        EXPECT_EQ(explore_action, folder_action.second)
            << "Exploring " << *explore_action << " instead of " << *folder_action.second << ". Test: " << folder_action.first;
        
        // test exploration with forced flag
        explore_action = high_level_planner.explore(scenario, transitions, true);
        EXPECT_EQ(explore_action.is_initialized(), true)
                << "A exploration action should have been selected. Test: " << folder_action.first;
        EXPECT_EQ(explore_action, folder_action.second)
            << "Exploring " << *explore_action << " instead of " << *folder_action.second << ". Test: " << folder_action.first;
        
        // test exploration in E3 exploration/exploitation algorithm
        explore_action= high_level_planner.exploit_or_explore(load_scenario(config_reader), transitions, excuse, actions_requiring_excuse);
        EXPECT_EQ(explore_action.is_initialized(), true)
                << "A exploration action should have been selected. Test: " << folder_action.first;
        EXPECT_EQ(explore_action, folder_action.second)
            << "Exploring " << *explore_action << " instead of " << *folder_action.second << ". Test: " << folder_action.first;
    }
}


TEST(LearnerTest, PasulaAllRulesDeterministic) {
    ConfigReader config_reader;
    config_reader.init("domain_tests/rexd_config.cfg");
    TransitionGroup transitions = read_transitions_from_file("domain_tests/transitions.dat");
    RuleSetPtr ground_truth_rules_ptr = boost::make_shared<RuleSet>(RuleSet("domain_tests/learned_nid_rules_all.dat"));
    RuleSetPtr ground_truth_action_rules_ptr = boost::make_shared<RuleSet>(RuleSet("domain_tests/learned_nid_rules_action.dat"));
    LearnerPtr learner_ptr = boost::make_shared<PasulaLearner>(PasulaLearner(config_reader));
    
    RuleSetPtr learned_rules_ptr = learner_ptr->learn_quiet(transitions);
    // set same indexes as ground truth, else files won't be equal
    // rule index cannot be restarted, static variable in constructor
    learned_rules_ptr->set_identifiers(ground_truth_rules_ptr);
    
    RuleSetPtr learned_action_rules_ptr = learner_ptr->learn_action_rules(*ground_truth_rules_ptr, transitions, Symbol("placefaceplatefront()"));
    
    learned_rules_ptr->write_rules_to_file("generated/learned_nid_rules.dat");
    learned_action_rules_ptr->write_rules_to_file("generated/learned_nid_action_rules.dat");
    
    EXPECT_EQ(*learned_rules_ptr, *ground_truth_rules_ptr)
        << "Error learning all rules.";
    EXPECT_EQ(*learned_action_rules_ptr, *ground_truth_action_rules_ptr)
        << "Error learning action rules.";
    EXPECT_EQ(compare_two_files("domain_tests/learned_nid_rules_all.dat", "generated/learned_nid_rules.dat"), true)
        << "Error generating NID rules file.";
    EXPECT_EQ(compare_two_files("domain_tests/learned_nid_rules_action.dat", "generated/learned_nid_action_rules.dat"), true)
        << "Error generating NID rules file.";
}
 
int main(int argc, char **argv) {
    testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}