/*
 * Class to read and store configuration values.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */


#ifndef CONFIG_READER_H
#define CONFIG_READER_H

#include <fstream>
#include <boost/program_options.hpp>

/*!
\class ConfigReader
\brief Class to read the configuration file and command line params.

It uses boost::program_options.\n
Command line params should be trivial to add, but they are not parsed at this moment.

TODO try creating a class ConfigReaderIntellact : public ConfigReader, and move there all IntellAct configs.
*/
class ConfigReader
{

public:
    std::string files_path;

    ConfigReader();
    ConfigReader(const ConfigReader& other);
    virtual ~ConfigReader();

    /** 
     * \brief read the config file from a path, and sets the default file path from which to get relative paths 
     * 
     * \param config_path path of the config file
     * \param files_path_ path to be added to relative paths to get absolute paths
     * \return True if initilization was successful.
     */
    bool init(const std::string& config_path, const std::string& files_path_ = "./");

    /** 
     * \brief Gets the value of a config variable
     * 
     * \param variable variable whose value is to be obtained.
     * \return Value of the variable.
     */
    template<typename T>
    T get_variable(std::string variable) const {
        if (vm.count(variable)) {
            return vm[variable].as<T>();
        }
        else {
            std::cerr << "ConfigReader::ERROR: " << "Variable " << variable << " not found!" << std::endl;
            exit(-1);
        }
    }

    /** 
     * \brief Gets an absolute path from a variable that has a relative path
     * 
     * \param variable variable that has the relative path.
     * \return Value of the absolute path.
     */
    std::string get_path(std::string variable) const {
        if (vm.count(variable)) {
            return (files_path + vm[variable].as<std::string>());
        }
        else {
            std::cerr << "ConfigReader::ERROR: " << "Variable " << variable << " not found!" << std::endl;
            exit(-1);
        }
    }

    /** 
     * \brief Changes a variable value.
     * 
     * HACK? Not sure if I am supposed to do this...
     * 
     * \param variable variable to be changed.
     * \param new_value new value to be set to the variable.
     */
    template<typename T>
    void set_variable(std::string variable, T new_value) {
        boost::program_options::variable_value aux = boost::program_options::variable_value(new_value, false);
        boost::program_options::variables_map::iterator it=vm.find(variable);
        it->second = aux;
    }

private:
    boost::program_options::variables_map vm;
};

#endif // CONFIG_READER_H
