#include "rexd/excuses.h"


#define DEBUG_EXCUSES 0
#define DEBUG_EXCUSE_GENERATION 0
#define DEBUG_EXCUSE_SELECTION 0
#define DEBUG_EXCUSE_RULE_ALT 0
#define DEBUG_EXCUSE_SUBGOALS 0

// Subgoals and excuses search
#define SUBGOALS_MAX_DEPTH 4
#define EXCUSES_CANDIDATES_MAX_DEPTH 8
#define NUM_LOOKAHEAD_ACTIONS_PLANNING 10

// Excuse selection reward penalties
#define EXCUSE_MAX_GOAL_REWARD_PENALTY -8
#define EXCUSE_MAX_REWARD_PENALTY -2.2
#define EXCUSE_MIDDLE_GOAL_REWARD_PENALTY -3
#define EXCUSE_MIDDLE_REWARD_PENALTY -1.2
#define EXCUSE_MIN_GOAL_REWARD_PENALTY -1
#define EXCUSE_MIN_REWARD_PENALTY -0.2

// Special strings for excuse selection
// WARNING the domain shouldn't have this strings to function properly
const std::string ExcuseGenerator::EXCUSE_RULE_PREFIX_ADD = "add";
const std::string ExcuseGenerator::EXCUSE_RULE_PREFIX_NEGATE = "neg";
const char ExcuseGenerator::EXCUSE_RULE_PREFIX_SEPARATOR = '_';


ExcuseGenerator::ExcuseGenerator(const Scenario& scenario_):
    scenario(scenario_),
    excuses_already_generated(false)
{
    int DEBUG = DEBUG_EXCUSES;

    // grounded rules
    grounded_rules = scenario.get_rules().get_full_grounded_rules(scenario.get_state()); // TODO improve
    grounded_rules.remove_nop_rules();
    grounded_rules.add_negated_outcomes_as_preconditions();
    grounded_rules.index(scenario.domain);

    if (DEBUG > 0)
        CINFO("excuses") << "Creating excuse generator!";
    
    if (DEBUG > 2)
        CINFO("excuses") << scenario.domain;
    if (DEBUG > 1)
        CINFO("excuses") << "Domain size is " << scenario.domain.get_predicates_size();

    // causal graph
    causal_graph.init(scenario.domain, grounded_rules);
    if (DEBUG > 1) {
        causal_graph.dump(scenario.domain);
    }

    // goals
    if (scenario.config_reader.get_variable<bool>("final_reward_action")) {
        RuleSet finish_rule(scenario.config_reader.get_path("nid_finish_rule_path"));
        simple_goal = finish_rule.front()->goals.adapt_finish_rule_goal_to_simple_predicate_goal();
    }
    else {
        simple_goal = scenario.get_goals().front();
    }

    if (DEBUG > 0)
        CINFO("excuses") << "Created excuse generator!";
}


PredicateGroup ExcuseGenerator::get_excuse()
{
    if ( (!excuses_already_generated) && scenario.config_reader.get_variable<bool>("rule_analysis") ) {
        CINFO("excuses") << "Generating excuses... " << NOENDL;
        excuses_generated = get_good_excuse();
        CINFO("excuses") << "Best excuse is " << excuses_generated;
        excuses_already_generated = true;
    }
    return excuses_generated;
}


RuleSet ExcuseGenerator::get_rules_to_add_excuses(const PredicateGroup& excuses) const
{
    RuleSet extra_rules;

    for (PredicateGroup::const_iterator pred_it = excuses.begin(); pred_it != excuses.end(); ++pred_it) {
        GroundedRulePtr rule = create_excuse_change_predicate(*pred_it);
        extra_rules.push_back(rule);
    }

    return extra_rules;
}


PredicateList ExcuseGenerator::get_possible_excuses(const Goal& goal)
{
    int DEBUG = DEBUG_EXCUSE_GENERATION;
    PredicateList excuses;
    std::set<int> excuses_set;

    // Get excuses
    for ( PredicateList::const_iterator goal_pred=goal.begin() ; goal_pred != goal.end(); goal_pred++ ) {
        size_t goal_idx = scenario.domain.get_predicate_index(*goal_pred);
        if (DEBUG > 0)
            CINFO("excuses") << "Excuse candidates for goal " << *goal_pred;
        std::set<int> new_excuses = get_possible_excuses(goal_idx);
        excuses_set.insert( new_excuses.begin(), new_excuses.end() );
    }

    std::vector<int> excuses_vector;
    excuses_vector.assign(excuses_set.begin(), excuses_set.end());
    sort(excuses_vector.begin(), excuses_vector.end());

    for (size_t i = 0; i < excuses_vector.size(); i++) {
        excuses.push_back(scenario.domain.get_predicate(excuses_vector[i]));
    }
    if (DEBUG > 0)
        CINFO("excuses") << "Excuse candidates" << excuses_vector;

    return excuses;
}

bool ExcuseGenerator::is_reachable(const Predicate goal_pred) const
{
    if (grounded_rules.can_obtain_predicate(goal_pred) > 0.0) {
        return true;
    }
    else {
        return false; // no rule can obtain it
    }
}


std::set<int> ExcuseGenerator::get_possible_excuses(size_t goal_idx)
{
    std::set<int> relevant_domain, excuses, visited_predicates;
    get_relevant_domain_and_excuses(goal_idx, relevant_domain, excuses, visited_predicates);
    return excuses;
}


std::set<int> ExcuseGenerator::get_relevant_domain(size_t goal_idx)
{
    std::set<int> relevant_domain, excuses, visited_predicates;
    get_relevant_domain_and_excuses(goal_idx, relevant_domain, excuses, visited_predicates);
    return relevant_domain;
}


void ExcuseGenerator::get_relevant_domain_and_excuses(size_t predicate_idx, std::set<int>& relevant_domain, std::set<int>& excuses, std::set<int>& visited_predicates, std::vector<uint> recursive_stack)
{
    int DEBUG = DEBUG_EXCUSE_GENERATION;
    if (recursive_stack.size() > EXCUSES_CANDIDATES_MAX_DEPTH)
        return;

    visited_predicates.insert(predicate_idx);
    relevant_domain.insert(predicate_idx);
    recursive_stack.push_back(predicate_idx);
    Predicate predicate = scenario.domain.get_predicate(predicate_idx);

    if (scenario.get_state().satisfies(predicate)) {
        // TODO maybe shouldn't be considered
        // already satisfied, ignoring....
        if (DEBUG > 0)
            CINFO("excuses") << "debug: " << predicate << " already satisfied.";
    }
    else {
        if (DEBUG > 0)
            CINFO("excuses") << "debug: " << predicate << " (" << predicate_idx << "), param_names.size()";
        std::vector<int> variable_relevant_domain;
        std::vector<int> eff_to_pre = causal_graph.get_eff_to_pre(predicate_idx);
        std::vector<int> eff_to_eff = causal_graph.get_eff_to_eff(predicate_idx);
//         variable_relevant_domain = eff_to_pre;
        // ¿¿ concat with eff_to_eff ???
        for (size_t i = 0; i < eff_to_pre.size(); ++i) {
            Predicate pred_aux = scenario.domain.get_predicate(eff_to_pre[i]);
//                 .get_negated();
//                 variable_relevant_domain.push_back(eff_to_pre[i]);
            if (predicate.param_names.size() >= pred_aux.param_names.size()) {
                variable_relevant_domain.push_back(scenario.domain.get_predicate_index(pred_aux));
            }
        }
        if (DEBUG > 0)
            CINFO("excuses") << "debug: " << "Variable relevant domain size is " << variable_relevant_domain.size();

        if ( !is_reachable(predicate) ) {
            if (DEBUG > 0)
                CINFO("excuses") << "debug: Adding predicate [reachable] " << predicate << " (" << predicate_idx << ")";
            // no rule can obtain it
            excuses.insert(predicate_idx);
        }
        else {
            if (DEBUG > 0)
                CINFO("excuses") << "debug: Adding predicate [else] " << predicate << " (" << predicate_idx << ")";
            // a rule can obtain it
            excuses.insert(predicate_idx);
        }

        for (std::vector<int>::iterator it_p = variable_relevant_domain.begin(); it_p != variable_relevant_domain.end(); ++it_p) {
            if (visited_predicates.count(*it_p) == 0) { // Not visited
                if (DEBUG > 1)
                    CINFO("excuses") << "Recursive for " << *it_p;
                get_relevant_domain_and_excuses(*it_p, relevant_domain, excuses, visited_predicates, recursive_stack);
            }
            else {
                // add recursively dom_rel(f)
//                 CINFO("excuses") << "Adding recursively dom_rel(f)";
//                 excuses.insert(*it_p); // TODO
                std::vector<uint>::iterator old_predicate = recursive_stack.end();
                old_predicate--; // now pointing to predicate_idx
                while ( (old_predicate != recursive_stack.begin()) && (*old_predicate!=predicate_idx) ) {
                    old_predicate--;
                    if (*old_predicate!=predicate_idx) {
                        excuses.insert(*old_predicate);
                        if (DEBUG > 0)
                            CINFO("excuses") << "debug: Adding recursively dom_rel(f), predicate: " << scenario.domain.get_predicate(*old_predicate) << " (" << predicate_idx << ")";
                    }
                }
            }
        }
    }

//     return excuses;
}



PredicateGroup ExcuseGenerator::get_good_excuse()
{
    PredicateGroup excuses;
    PredicateList excuse_candidates = get_possible_excuses(simple_goal);
    if (excuse_candidates.size() > 0) {
        // option 1: add all rules and check which one is better
        if (scenario.config_reader.get_variable<std::string>("planner") == "ippc") {
            excuses = obtain_excuses_planning_with_aux_actions(excuse_candidates);
        }
        // option 2: change state
        else if (scenario.config_reader.get_variable<std::string>("planner") == "prada") {
            excuses = obtain_excuses_planning_with_changed_state(scenario.get_rules(), scenario.get_goals(), excuse_candidates);
        }

        if (excuses.size() == 0) {
            excuses = simple_goal.get_not_satisfied(scenario.get_state());
        }
    }
    return excuses;
}


bool ExcuseGenerator::get_possible_subgoals(size_t goal_idx, size_t excuse_idx, std::set<int>& possible_subgoals, std::set<int> visited_predicates, size_t depth)
{
    int DEBUG = DEBUG_EXCUSE_SUBGOALS;
    if (depth > SUBGOALS_MAX_DEPTH)
        return false;
    bool found_excuse = false;

    visited_predicates.insert(goal_idx);
    Predicate goal_predicate = scenario.domain.get_predicate(goal_idx);

    if (goal_idx == excuse_idx) {
        found_excuse = true;
    }
    else if (scenario.get_state().satisfies(goal_predicate)) {
        // TODO maybe shouldn't be considered
        // already satisfied, ignoring....
//         CINFO("excuses") << goal_predicate << " already satisfied.";
    }
    else {
        if (DEBUG > 0)
            CINFO("excuses") << "debug: " << goal_predicate << " (" << goal_idx << "), param_names.size()=" << goal_predicate.param_names.size();
//         CINFO("excuses") << "debug: " << goal_idx;
        std::vector<int> variable_relevant_domain;
        std::vector<int> eff_to_pre = causal_graph.get_eff_to_pre(goal_idx);
        std::vector<int> eff_to_eff = causal_graph.get_eff_to_eff(goal_idx);
//         variable_relevant_domain = eff_to_pre;
        // ¿¿ concat with eff_to_eff ???

        for (size_t i = 0; i < eff_to_pre.size(); ++i) {
            Predicate pred_aux = scenario.domain.get_predicate(eff_to_pre[i]);
//                 .get_negated();
//                 variable_relevant_domain.push_back(eff_to_pre[i]);
            if (goal_predicate.param_names.size() >= pred_aux.param_names.size()) {
                variable_relevant_domain.push_back(scenario.domain.get_predicate_index(pred_aux));
            }
        }

        std::set< int > subgraphs_with_excuse;
        for (std::vector<int>::iterator it_p = variable_relevant_domain.begin(); it_p != variable_relevant_domain.end(); ++it_p) {
            if (visited_predicates.count(*it_p) == 0) { // Not visited
                if (DEBUG > 1)
                    CINFO("excuses") << "Recursive for " << scenario.domain.get_predicate(*it_p);
                bool found_recursive_excuse = get_possible_subgoals(*it_p, excuse_idx, possible_subgoals, visited_predicates, depth+1);
                if (DEBUG > 1)
                    CINFO("excuses") << "Result recursive for " << scenario.domain.get_predicate(*it_p) << " was " << found_recursive_excuse;
                if (found_recursive_excuse) {
                    found_excuse = true;
                    subgraphs_with_excuse.insert(*it_p); // it depends on excuse, so it is marked to not be added as subgoal
                }
            }
        }
        if (found_excuse) { // add preconditions not related to excuse
            for (std::vector<int>::iterator it_p = variable_relevant_domain.begin(); it_p != variable_relevant_domain.end(); ++it_p) {
                if ((visited_predicates.count(*it_p) == 0) && (subgraphs_with_excuse.count(*it_p) == 0)) { // Doesn't depend on the excuse
                    if (is_reachable(scenario.domain.get_predicate(*it_p))) {
                        possible_subgoals.insert(*it_p);
                        if (DEBUG > 0)
                            CINFO("excuses") << "Adding " << scenario.domain.get_predicate(*it_p) << "(" << *it_p << ")";
                    }
                }
            }
        }
    }

    return found_excuse;
}



int ExcuseGenerator::get_distance_to_goal(const Goal& goal, size_t target_predicate_idx)
{
    std::set<int> predicates;
    int min_dist = std::numeric_limits<int>::max();

    // Get excuses
    for ( PredicateList::const_iterator goal_pred=goal.begin() ; goal_pred != goal.end(); goal_pred++ ) {
        size_t goal_idx = scenario.domain.get_predicate_index(*goal_pred);
        int new_dist = get_distance_to_predicate(target_predicate_idx, goal_idx);
        if (new_dist < min_dist)
            min_dist = new_dist;
    }

    return min_dist;
}

// TODO from predicate to all goals at the same time
int ExcuseGenerator::get_distance_to_predicate(size_t target_predicate_idx, size_t predicate_idx)
{
    int DEBUG = DEBUG_EXCUSE_SUBGOALS;
    int result = 0;
    std::set<size_t> visited_predicates;
    typedef std::pair <size_t,int> PredicateAndDistance;
    std::queue<PredicateAndDistance> predicate_queue;
    PredicateAndDistance predicate_idx_dist (predicate_idx, 0);

    predicate_queue.push(predicate_idx_dist);
    visited_predicates.insert(predicate_idx);

    bool found = false;
    while ( (!found) && (!predicate_queue.empty()) ) {
        PredicateAndDistance current_predicate_idx_dist = predicate_queue.front();
        predicate_queue.pop();
        size_t current_predicate_idx = current_predicate_idx_dist.first;
        int current_predicate_dist = current_predicate_idx_dist.second;
        if (DEBUG > 3)
            CINFO("excuses") << "get_distance_to_predicate: " << "debug: " << scenario.domain.get_predicate(current_predicate_idx) << " (" << current_predicate_dist << ")";

        if (target_predicate_idx == current_predicate_idx) {
            if (DEBUG > 3)
                CINFO("excuses") << "Found with distance " << current_predicate_dist;
            result = current_predicate_dist;
            found = true;
        }
        else {
            std::vector<int> variable_relevant_domain;
            std::vector<int> eff_to_pre = causal_graph.get_eff_to_pre(current_predicate_idx);
            std::vector<int> eff_to_eff = causal_graph.get_eff_to_eff(current_predicate_idx);
            variable_relevant_domain = eff_to_pre;

            for (std::vector<int>::iterator it_p = variable_relevant_domain.begin(); it_p != variable_relevant_domain.end(); ++it_p) {
                if (visited_predicates.count(*it_p) == 0) { // Not visited
                    visited_predicates.insert(*it_p);
                    PredicateAndDistance new_predicate_idx_dist (*it_p, current_predicate_dist + 1);
                    predicate_queue.push(new_predicate_idx_dist);
                }
            }
        }
    }
    if (!found) {
        if (DEBUG > 3)
            CINFO("excuses") << "get_distance_to_predicate: " << "debug: " << " Not found!";
        result = std::numeric_limits<int>::max();
    }

    return result;
}

// auxiliar function for sort_by_distance_to_goal
// TODO make a lambda function with cxx11
bool compare_pred_dist_pair(std::pair<int, int> i, std::pair<int, int> j)
{
    return (i.second<j.second);
}

std::vector<int> ExcuseGenerator::sort_by_distance_to_goal(const Goal& goal, const std::set<int>& possible_subgoals)
{
    std::vector<int> possible_subgoals_vect;
    std::vector< std::pair<int,int> > distances_to_goal;
    for (std::set<int>::const_iterator possible_subgoal_it = possible_subgoals.begin(); possible_subgoal_it != possible_subgoals.end(); ++possible_subgoal_it) {
        std::pair<int, int> dist_to_goal(*possible_subgoal_it, get_distance_to_goal(goal, *possible_subgoal_it));
        distances_to_goal.push_back(dist_to_goal);
    }

    std::sort (distances_to_goal.begin(), distances_to_goal.end(), compare_pred_dist_pair);
    for (std::vector< std::pair<int,int> >::iterator dist_it = distances_to_goal.begin(); dist_it != distances_to_goal.end(); ++dist_it) {
        possible_subgoals_vect.push_back(dist_it->first);
    }
    return possible_subgoals_vect;
}


bool ExcuseGenerator::is_mutex_free(PredicateList orig_predicates, const int new_predicate_idx)
{
    int DEBUG = DEBUG_EXCUSE_GENERATION;
    bool mutex_free = true;
    std::set<int> relevant_domain;
    for ( PredicateList::iterator pred_it=orig_predicates.begin(); pred_it != orig_predicates.end(); ++pred_it) {
        int pred_idx = scenario.domain.get_predicate_index(*pred_it);
        std::set<int> addition_domain = get_relevant_domain(pred_idx);
        relevant_domain.insert( addition_domain.begin(), addition_domain.end() );
    }

    // new predicate
    std::set<int> new_predicate_domain;
    new_predicate_domain = get_relevant_domain(new_predicate_idx);
    std::vector<int> new_predicate_domain_vect;
    new_predicate_domain_vect.assign(new_predicate_domain.begin(), new_predicate_domain.end());

    // check if old domain is compatible new predicate
    for (std::vector<int>::iterator domain_it=new_predicate_domain_vect.begin(); domain_it!=new_predicate_domain_vect.end(); ++domain_it) {
        // check if negated predicate is required by the domain -> mutex!
        Predicate predicate = scenario.domain.get_predicate(*domain_it);
        predicate.negate();
        int negated_pred_idx = scenario.domain.get_predicate_index(predicate);
        if (relevant_domain.count(negated_pred_idx) > 0) {
            mutex_free = false;
            if (DEBUG > 0)
                CINFO("excuses") << "Possible mutex, predicate found in original predicates is " << predicate;
        }
    }

    return mutex_free;
}


/**
 * Subgoals
 */


PredicateGroup ExcuseGenerator::get_subgoal(const Goal& goal, const PredicateGroup& excuse)
{
    int DEBUG = DEBUG_EXCUSE_SUBGOALS;
    PredicateGroup subgoal(goal);

    if (DEBUG > 0)
        CINFO("excuses") << "Getting subgoals...";

    std::set<int> possible_subgoals;
    for ( PredicateGroup::const_iterator excuse_it = excuse.begin(); excuse_it != excuse.end(); ++excuse_it) {
        size_t excuse_idx = scenario.domain.get_predicate_index(*excuse_it);
        for ( PredicateGroup::iterator goal_pred=subgoal.begin() ; goal_pred != subgoal.end(); ) {
            if (DEBUG > 0)
                CINFO("excuses") << START_COLOR_WHITE << "Checking goal" << END_COLOR;
            size_t goal_idx = scenario.domain.get_predicate_index(*goal_pred);
            if (get_possible_subgoals(goal_idx, excuse_idx, possible_subgoals)) {
                if (DEBUG > 2)
                    CINFO("excuses") << "Removing goal " << *goal_pred;
                goal_pred = subgoal.erase(goal_pred);
            }
            else {
                ++ goal_pred;
            }
        }
    }

    // add goal_dependencies if mutex free
    std::vector<int> possible_subgoals_vect;
//     possible_subgoals_vect.assign(possible_subgoals.begin(), possible_subgoals.end());
    possible_subgoals_vect = sort_by_distance_to_goal(goal, possible_subgoals);
    if (DEBUG > 0)
        CINFO("excuses") << "Possible subgoals are " << possible_subgoals_vect;

    for ( std::vector<int>::iterator subgoal_it=possible_subgoals_vect.begin() ; subgoal_it != possible_subgoals_vect.end(); ++subgoal_it) {
        if (DEBUG > 1)
            CINFO("excuses") << "Checking subgoal " << scenario.domain.get_predicate(*subgoal_it);
        if (is_mutex_free(subgoal, *subgoal_it)) {
            // check if required by excuses
            bool not_required_by_excuses = true;
            for ( PredicateGroup::const_iterator excuse_it = excuse.begin(); excuse_it != excuse.end(); ++excuse_it) {
                if (DEBUG > 1)
                    CINFO("excuses") << "Checking subgoal " << scenario.domain.get_predicate(*subgoal_it) << " required by " << *excuse_it;
                std::set<int> aux_vect;
                if (get_possible_subgoals(*subgoal_it, scenario.domain.get_predicate_index(*excuse_it), aux_vect)) { // required
                    not_required_by_excuses = false;
                }
            }
            if (not_required_by_excuses) {
                subgoal.push_back(scenario.domain.get_predicate(*subgoal_it));
                if (DEBUG > 1)
                    CINFO("excuses") << "Adding subgoal " << scenario.domain.get_predicate(*subgoal_it);
            }
        }
    }

    return subgoal;
}

/**
 * Generate rule alternatives
 */

std::vector<RuleSet> ExcuseGenerator::get_rules_that_overcome_excuse(const RuleSet rules, const PredicateGroup excuse, const TransitionList& transitions) const
{
    int DEBUG = DEBUG_EXCUSE_RULE_ALT;
    std::vector<RuleSet> extra_rulesets;

    double orig_score = rules.score(transitions);

    RuleSet rules_requiring_excuse;

    PredicateGroup symbolic_excuse(excuse);
    symbolic_excuse.make_symbolic();
    rules_requiring_excuse = rules.find_rules_with_a_precondition(symbolic_excuse);

//         RuleSet rules_requiring_excuse = rules.find_rules_with_precondition(excuse);
    // RuleSet rules_requiring_excuse = rules.find_rules_with_outcome(excuse);TODO
    // TODO change/remove precondition
    CINFO("excuses") << "Changing " << rules_requiring_excuse.size() << " rules.";
    for ( RuleList::iterator rule_it=rules_requiring_excuse.begin() ; rule_it != rules_requiring_excuse.end(); ++rule_it ) {
        RuleSet rules_copy(rules);
        rules_copy.remove_rule(*rule_it);

        RuleSet new_rules = create_new_rule_candidates_with_excuse(*rule_it, symbolic_excuse, transitions);
        CINFO("excuses") << "Checking " << new_rules.size() << " new rules for action " << (*rule_it)->name << " with score ( " << orig_score << " ): " << NOENDL;
        for ( RuleList::iterator new_rule_it=new_rules.begin() ; new_rule_it != new_rules.end(); ++new_rule_it ) {
            RuleSet new_generated_rules(rules_copy);
            new_generated_rules.push_back(*new_rule_it);
            double new_score = new_generated_rules.score(transitions);
            if (new_score != (-1. * std::numeric_limits<double>::max()) ) {
                if (DEBUG > 1)
                    CINFO("excuses") << "Got new rule with score " << new_score << " (orig score was " << orig_score << ")";
                if (DEBUG > 2)
                    CINFO("excuses") << **new_rule_it;
                if (new_score >= orig_score) {
                    CINFO("excuses") << "*" << NOENDL;
                    if (DEBUG > 0)
                        CINFO("excuses") << "Got new rule \"" << (*new_rule_it)->name
                                         << "\" with score " << new_score << " (orig score was " << orig_score << ")";
                    if (DEBUG > 1)
                        CINFO("excuses") << **new_rule_it;
                    extra_rulesets.push_back(new_generated_rules);
                }
                else {
                    CINFO("excuses") << "." << NOENDL;
                }
            }
        }
        CINFO("excuses");
    }

    return extra_rulesets;
}


/**
 *
 * Obtaining excuses changing the state and replanning
 *
 */

PredicateGroup ExcuseGenerator::obtain_excuses_planning_with_changed_state(const RuleSet& rules, const GoalGroup& goals, const PredicateGroup& excuse_candidates)
{
    std::vector<float> planning_results;
    PlannerPtr new_planner = PlannerFactory::get_standard_planner(scenario.config_reader);
    for ( PredicateList::const_iterator excuse_it=excuse_candidates.begin() ; excuse_it != excuse_candidates.end(); ++excuse_it ) {
        State new_state(scenario.get_state());
        new_state.update_predicate(*excuse_it);
        CINFO("excuses") << "  Using new state with excuse " << *excuse_it;
        float planning_result;
        if (goals.is_goal(new_state)) {
            CINFO("excuses") << "New state is goal.";
            planning_result = 1000.0;
        }
        else {
            Scenario new_scenario(scenario);
            new_scenario.set_state(new_state);
            new_planner->plan(new_scenario, planning_result, false, false);
//                     std::cout << "Planning result is " << planning_result;
        }
//                 if (planning_result > 0.0) { // only add positive results
        planning_results.push_back(planning_result);
//                 }
    }
    PredicateGroup best_excuse = get_best_excuse(excuse_candidates, planning_results);

    return best_excuse;
}

PredicateGroup ExcuseGenerator::get_best_excuse(PredicateGroup excuses, std::vector<float> planning_results) const
{
    PredicateGroup best_excuse;
    size_t max_idx;
    double max_dist = -1;

    for (size_t i = 0; i < excuses.size(); i++) {
        if (planning_results[i] > 0) {
            CINFO("excuses") << "get_best_excuse: " << "debug: " << excuses[i] << "  with result " << planning_results[i];
            double new_dist = 1/planning_results[i];
            if (new_dist > max_dist) {
                CINFO("excuses") << "get_best_excuse: " << "debug: " << "is better! Distance is " << new_dist;
                max_dist = new_dist;
                max_idx = i;
            }
        }
    }

    if (max_dist >= 0) {
        best_excuse.push_back(excuses[max_idx]);
    }

    return best_excuse;
}


/**
 *
 * Obtaining excuses by creating auxiliary actions that add predicates.
 *
 */

PredicateGroup ExcuseGenerator::obtain_excuses_planning_with_aux_actions(const PredicateGroup& excuse_candidates)
{
    int DEBUG = DEBUG_EXCUSE_SELECTION;
    State state_and_excuses(scenario.get_state());
    state_and_excuses.add_predicates(excuse_candidates);
    state_and_excuses.remove_duplicated();
    RuleSet grounded_rules = scenario.get_rules().get_grounded_rules(state_and_excuses);
    RuleSet extra_rules = get_rules_to_add_excuses(excuse_candidates);

    PredicateGroup excuses;
    std::vector<Symbol> excuse_plan;
    if (!scenario.config_reader.get_variable<bool>("final_reward_action")) { // TODO HACK to speed up when use final_reward_action (it will never find excuses with such high rewards)
        excuse_plan = obtain_plan_with_aux_actions(grounded_rules, extra_rules, EXCUSE_MAX_REWARD_PENALTY, EXCUSE_MAX_GOAL_REWARD_PENALTY);
        excuses = get_excuses_from_excuse_actions(excuse_plan);
    }
    if (excuses.empty()) { // retry
        CINFO("excuses") << "(retrying) " << NOENDL;
        excuse_plan = obtain_plan_with_aux_actions(grounded_rules, extra_rules, EXCUSE_MIDDLE_REWARD_PENALTY, EXCUSE_MIDDLE_GOAL_REWARD_PENALTY);
        excuses = get_excuses_from_excuse_actions(excuse_plan);
    }
    if (excuses.empty()) { // retry
        CINFO("excuses") << "(retrying hard)" << NOENDL;
        excuse_plan = obtain_plan_with_aux_actions(grounded_rules, extra_rules, EXCUSE_MIN_REWARD_PENALTY, EXCUSE_MIN_GOAL_REWARD_PENALTY);
        excuses = get_excuses_from_excuse_actions(excuse_plan);
    }
    CINFO("excuses"); // endline after searching excuses and retrying X times

    if (DEBUG > 1)
        CINFO("excuses") << "Excuses obtained were " << excuses;

    return excuses;
}


std::vector<Symbol> ExcuseGenerator::obtain_plan_with_aux_actions(const RuleSet& grounded_rules, const RuleSet& extra_rules, const float min_reward, const float goal_shortcut_reward)
{
    RuleSet new_rules(grounded_rules);
    int DEBUG = DEBUG_EXCUSE_SELECTION;
    for ( RuleSet::const_iterator rule_it=extra_rules.begin() ; rule_it != extra_rules.end(); ++rule_it ) {
        int dist_to_goal = get_distance_to_goal(simple_goal, scenario.domain.get_predicate_index((*rule_it)->outcomes.front().front()));
        if (dist_to_goal == std::numeric_limits<int>::max()) {
        }
        else if (dist_to_goal > 0) {
            (*rule_it)->set_goal(-0.1 + (min_reward/dist_to_goal));
            if (DEBUG > 1)
                CINFO("excuses") << "Reward for " << (*rule_it)->outcomes.front().front() << " is " << (*rule_it)->goals.front();
            new_rules.push_back(*rule_it);
        }
        else {
            (*rule_it)->set_goal(goal_shortcut_reward);
            if (DEBUG > 1)
                CINFO("excuses") << "Reward for " << (*rule_it)->outcomes.front().front() << " is " << (*rule_it)->goals.front();
            new_rules.push_back(*rule_it);
        }
    }

    PlannerPtr new_planner = PlannerFactory::get_standard_planner(scenario.config_reader);
    bool reached_goal;
    Scenario new_scenario(scenario); // TODO with no Vmin!
    new_scenario.set_rules(new_rules);
    //return new_planner->plan_n_actions(scenario, NUM_LOOKAHEAD_ACTIONS_PLANNING, reached_goal, true, true);
    return new_planner->plan_n_actions(new_scenario, scenario.config_reader.get_variable<uint>("planning_horizon"), reached_goal, false, false);
}


RuleSet ExcuseGenerator::create_new_rule_candidates_with_excuse(const RulePtr rule, const PredicateList symbolic_excuses, const TransitionList& transitions) const
{
    int DEBUG = DEBUG_EXCUSE_SELECTION;
    RuleSet result;
    if (DEBUG > 0)
        CINFO("excuses") << "Rule candidates for rule " << rule->name << " with symbolic excuses " << symbolic_excuses;

    if (DEBUG > 3)
        CINFO("excuses") << "Obtaining rule transitions...";
    TransitionList rule_transitions = rule->get_covering_transitions(transitions);

    if (DEBUG > 3)
        CINFO("excuses") << "Obtaining possible preconditions...";
    PredicateGroup new_preconditions;
    for ( TransitionList::const_iterator transition=rule_transitions.begin() ; transition != rule_transitions.end(); ++transition ) {
        new_preconditions.add_predicates(transition->get_prev_state());
    }
    new_preconditions.remove_duplicated();
    new_preconditions.remove(symbolic_excuses);
    if (DEBUG > 2)
        CINFO("excuses") << "Final possible preconditions are " << new_preconditions;

    if (DEBUG > 3)
        CINFO("excuses") << "Generating rules with new preconditions...";

    PredicateGroup new_preconditions_symbolic = new_preconditions.get_symbolic_predicates_permutations(rule->param_names);
    new_preconditions_symbolic.remove_duplicated();

    if (DEBUG > 3)
        CINFO("excuses") << "Adding symbolic predicates " << new_preconditions_symbolic;
    for ( PredicateList::iterator sym_pred_it=new_preconditions_symbolic.begin() ; sym_pred_it != new_preconditions_symbolic.end(); ++sym_pred_it) {
        SymbolicRulePtr new_rule = boost::make_shared<SymbolicRule>(SymbolicRule(*rule));
        new_rule->preconditions.remove(symbolic_excuses);
        if (!new_rule->preconditions.has_predicate(*sym_pred_it)) {
            if (DEBUG > 4)
                CINFO("excuses") << "Adding new rule candidate with precondition " << *sym_pred_it;
            new_rule->preconditions.push_back(*sym_pred_it);
            result.push_back(new_rule);
        }
    }

    return result;
}


PredicateGroup ExcuseGenerator::get_excuses_from_excuse_actions(std::vector<Symbol> excuse_plan)
{
    PredicateGroup result;

    for (std::vector<Symbol>::iterator plan_it = excuse_plan.begin(); plan_it != excuse_plan.end(); ++plan_it) {
        PredicatePtr excuse_ptr = get_symbol_from_excuse_action(*plan_it);
        if (excuse_ptr) {
            result.push_back(*excuse_ptr);
        }
    }

    return result;
}


PredicatePtr ExcuseGenerator::get_symbol_from_excuse_action(Symbol action)
{
    PredicatePtr excuse_ptr;

    if (action.name.find(EXCUSE_RULE_PREFIX_ADD) == 0) {
        std::stringstream ss(action.name);
        std::string item, predicate_name;
        bool positive = true, has_more_params;

        // notation: add_name[_neg]{_param1}
        // get string between EXCUSE_RULE_PREFIX_SEPARATOR
        std::getline(ss, item, EXCUSE_RULE_PREFIX_SEPARATOR);
        std::getline(ss, item, EXCUSE_RULE_PREFIX_SEPARATOR);
        if (item == EXCUSE_RULE_PREFIX_NEGATE) {
            positive = false;
            std::getline(ss, item, EXCUSE_RULE_PREFIX_SEPARATOR);
        }
        predicate_name = item;
        excuse_ptr = PredicatePtr(new Predicate(predicate_name, positive));

        has_more_params = std::getline(ss, item, EXCUSE_RULE_PREFIX_SEPARATOR);
        while(has_more_params) {
            excuse_ptr->param_names.push_back(item);
            has_more_params = std::getline(ss, item, EXCUSE_RULE_PREFIX_SEPARATOR);
        }
    }

    return excuse_ptr;
}


GroundedRulePtr ExcuseGenerator::create_excuse_change_predicate(const Predicate& predicate) const
{
    std::stringstream rule_name;
    rule_name << EXCUSE_RULE_PREFIX_ADD;
    if (!predicate.positive) {
        rule_name << EXCUSE_RULE_PREFIX_SEPARATOR << EXCUSE_RULE_PREFIX_NEGATE;
    }
    rule_name << EXCUSE_RULE_PREFIX_SEPARATOR << predicate.name;
    for (vector_string::const_iterator param_it = predicate.param_names.begin(); param_it != predicate.param_names.end(); ++param_it) {
        rule_name << EXCUSE_RULE_PREFIX_SEPARATOR << *param_it;
    }
    SymbolicRule symbolic_rule = SymbolicRule(Symbol(rule_name.str(), std::vector<std::string>()), PredicateList());
    GroundedRulePtr rule = GroundedRulePtr(new GroundedRule(symbolic_rule, std::vector<std::string>()));

    PredicateList predicates;
    predicates.push_back(predicate);
    rule->add_outcome(predicates, 1.0);

    return rule;
}

