/*
 * Class to find excuses given a scenario, and perform tasks related to excuses.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef INTELLACT_EXCUSES_H
#define INTELLACT_EXCUSES_H

#include <queue>

#include <symbolic/rules.h>
#include <symbolic/domain.h>
#include "symbolic/state.h"

#include <rexd/causal_graph.h>
#include <rexd/planner_factory.h>
#include "rexd/planner.h"
#include "rexd/learner.h"

/**
 * \class ExcuseGenerator
 * \brief Finds excuses (planning failures)
 * 
 * Also generates subgoals and rule alternatives.
 */
class ExcuseGenerator {
public:
    ExcuseGenerator(const Scenario& scenario);

    // get excuse
    PredicateGroup get_excuse();

    // get rule alternatives candidates
    std::vector<RuleSet> get_rules_that_overcome_excuse(const RuleSet rules, const PredicateGroup excuse, const TransitionList& transitions) const;

    // get subgoals
    PredicateGroup get_subgoal(const Goal& goal, const PredicateGroup& excuse);


private:
    CausalGraph causal_graph;
    RuleSet grounded_rules;
    Goal simple_goal;
    Scenario scenario;
    
    bool excuses_already_generated;
    PredicateGroup excuses_generated;

    static const std::string EXCUSE_RULE_PREFIX_ADD,
           EXCUSE_RULE_PREFIX_NEGATE;
    static const char EXCUSE_RULE_PREFIX_SEPARATOR;

    // Get candidates
    PredicateGroup get_good_excuse();
    PredicateList get_possible_excuses(const Goal& goal);
    std::set<int> get_possible_excuses(size_t goal_idx);
    std::set<int> get_relevant_domain(size_t goal_idx);
    void get_relevant_domain_and_excuses(size_t predicate_idx,
                                         std::set<int>& relevant_domain,
                                         std::set<int>& excuses,
                                         std::set<int>& visited_predicates,
                                         std::vector<uint> recursive_stack = std::vector<uint>());

    // utils
    bool is_reachable(const Predicate goal_pred) const;
    bool is_mutex_free(PredicateList orig_predicates, const int new_predicate_idx);

    // Rule variants
    RuleSet create_new_rule_candidates_with_excuse(const RulePtr rule, const PredicateList symbolic_excuses, const TransitionList& transitions) const;

    // Subgoals
    std::vector<int> sort_by_distance_to_goal(const Goal& goal, const std::set<int>& possible_subgoals);
    bool get_possible_subgoals(size_t goal_idx, size_t excuse_idx, std::set<int>& possible_subgoals, std::set<int> visited_predicates = std::set<int>(), size_t depth=1);

    // Distances
    int get_distance_to_goal(const Goal& goal, size_t target_predicate_idx);
    int get_distance_to_predicate(size_t target_predicate_idx, size_t predicate_idx);

    // Changing state
    PredicateGroup obtain_excuses_planning_with_changed_state(const RuleSet& rules, const GoalGroup& goals, const PredicateGroup& excuse_candidates);
    PredicateGroup get_best_excuse(PredicateGroup excuses, std::vector<float> planning_results) const;

    // Temporary action for excuses
    PredicateGroup obtain_excuses_planning_with_aux_actions(const PredicateGroup& excuse_candidates);
    std::vector<Symbol> obtain_plan_with_aux_actions(const RuleSet& grounded_rules, const RuleSet& extra_rules, const float min_reward, const float goal_shortcut_reward);
    RuleSet get_rules_to_add_excuses(const PredicateGroup& excuses) const;
    GroundedRulePtr create_excuse_change_predicate(const Predicate& predicate) const;
    PredicateGroup get_excuses_from_excuse_actions(std::vector<Symbol> excuse_plan);
    PredicatePtr get_symbol_from_excuse_action(Symbol action);
};

#endif
