/*
    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) 2013  dmartinez <email>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "rexd/explorer.h"

#define DEBUG_EXPLORATION 0

Explorer::Explorer(const Scenario& scenario_, TransitionGroupConstPtr transitions_ptr_):
    scenario(scenario_),
    transitions_ptr(transitions_ptr_)
{ }

Explorer::Explorer(const Explorer& other):
    scenario(other.scenario),
    transitions_ptr(other.transitions_ptr)
{ }

Explorer::~Explorer()
{

}


bool Explorer::is_rule_known(RuleConstPtr rule_ptr, const float visited_threshold) const
{
    int DEBUG = DEBUG_EXPLORATION;
    if (DEBUG > 0) {
        if (rule_ptr->dangerous)
            CINFO("explore") << "Not exploring rule " << rule_ptr->name << " as it is dangerous.";
    }
    return ((rule_ptr->dangerous) || (rule_ptr->is_special()) || (rule_ptr->count_covering_transitions_preconditions(*transitions_ptr) >= visited_threshold));
}


RulePtr Explorer::create_explore_rule(RuleConstPtr rule_ptr, const float rmax)
{
    RulePtr resulting_rule = rule_ptr->make_shared();
    Goal explore_goal(resulting_rule->preconditions, rmax);
    resulting_rule->goals.push_back(explore_goal);
    return resulting_rule;
}


RuleSetPtr Explorer::get_rmax_explore_rules(const float visited_threshold, const float rmax) const
{
    int DEBUG = DEBUG_EXPLORATION;
    if (DEBUG > 0)
        CINFO("explore") << "\033[1;35m" << "Start" << "\033[0m";

    RuleSetPtr explore_rules = RuleSetPtr(new RuleSet());

    BOOST_FOREACH(RuleConstPtr rule_ptr, scenario.get_rules()) {
//     for ( RuleList::iterator rule_it = scenario.get_rules().begin() ; rule_it != scenario.get_rules().end(); ++rule_it) {
        if (is_rule_known(rule_ptr, visited_threshold)) {
            explore_rules->push_back(rule_ptr->make_shared());
        }
        else {
            if (DEBUG > 0)
                CINFO("explore") << "Creating explore rule";
            explore_rules->push_back( create_explore_rule(rule_ptr, rmax) );
        }
    }
    return explore_rules;
}


bool Explorer::is_state_known() const
{
    int DEBUG = 0;
    uint known_threshold = scenario.config_reader.get_variable<uint>("exploration_known_threshold");
    bool avoid_dangerous_actions = scenario.config_reader.get_variable<bool>("confirm_dangerous_actions");
    CINFO("explore") << "Checking if state is known... " << NOENDL;
    BOOST_FOREACH(const Symbol& action, scenario.get_action_list()) {
        RuleSet action_rules = scenario.get_rules().get_action_rules(action.name);
        RuleSet covering_rules = action_rules.get_covering_rules(scenario.get_state());
        if (covering_rules.size() > 1) {
            CINFO("explore") << "(There are many covering rules) " << NOENDL;
            for (RuleList::iterator rule_it = covering_rules.begin(); rule_it != covering_rules.end(); ++rule_it) {
                uint num_explorations = (*rule_it)->count_covering_transitions_preconditions(*transitions_ptr);
                if (DEBUG > 1)
                    CINFO("explore") << "Checking rule " << **rule_it << "\nWas explored " << num_explorations << " times.";

                if (num_explorations < known_threshold) {
                    CINFO("explore") << " False ( rule " << action << " explored only " << num_explorations << " times)";
                    return false;
                }
            }
        }
        if (covering_rules.size() == 0) {
            CINFO("explore") << " False (no covering rules for action " << action << ")";
            return false;
        } else if (!(avoid_dangerous_actions && covering_rules[0]->dangerous)) {
            SymbolicRule rule(*covering_rules[0]);
            if (rule.count_covering_transitions_preconditions(*transitions_ptr) < known_threshold) {
                CINFO("explore") << " False ( rule " << action << " explored only " << rule.count_covering_transitions_preconditions(*transitions_ptr) << " times)";
                return false;
            }
        } else { // covered by just one rule, dangerous
            CINFO("explore") << "Dangerous rule " << covering_rules[0]->name << ". Not exploring it.";
        }
    }
    CINFO("explore") << " True.";
    return true;
}


SymbolOptional Explorer::select_action_for_exploration(const bool forced, const uint visited_threshold) const
{
    int DEBUG = DEBUG_EXPLORATION;
    if (DEBUG > 0)
        CINFO("explore") << "\033[1;35m" << "Start" << "\033[0m";
    SymbolOptional selected_action_opt;
    RulePtr best_rule_ptr;
    uint min_coverage = visited_threshold; // if (min_coverage == visited_threshold) then everything is known
    std::vector<Symbol> action_list_shuffle = scenario.get_action_list();
    std::random_shuffle ( action_list_shuffle.begin(), action_list_shuffle.end() );

    for (std::vector< Symbol >::const_iterator action_it = action_list_shuffle.begin(); action_it != action_list_shuffle.end(); ++action_it) {
        RuleSet action_rules = scenario.get_rules().get_action_rules(action_it->name);
        RuleSet covering_rules = action_rules.get_covering_rules(scenario.get_state());
        if (covering_rules.size()>=1) {
            if (DEBUG > 0)
                CINFO("explore") << "Checking covering rules for action " << *action_it;
            for (RuleList::iterator rule_it = covering_rules.begin(); rule_it != covering_rules.end(); ++rule_it) {
                if (!(*rule_it)->dangerous) {
                    uint num_explorations = (*rule_it)->count_covering_transitions_preconditions(*transitions_ptr);
                    if (DEBUG > 1)
                        CINFO("explore") << "Checking rule " << (*rule_it)->name << ": Was explored " << num_explorations << " times.";

                    if (num_explorations < min_coverage) {
                        min_coverage = num_explorations;
                        best_rule_ptr = boost::make_shared<SymbolicRule>(*static_cast<const SymbolicRule*>(&(**rule_it)));
                        CINFO("explore") << "Unexplored rule found for action: " << *action_it << ", explored only " << num_explorations << " times)";
                    }
                }
                else { // don't explore dangerous rules
                    CINFO("explore") << "Dangerous rule " << (*rule_it)->name << ". Not exploring it.";
                }
            }
        }
        else {
            CINFO("explore") << "No covering rules for action " << *action_it << ")";
            selected_action_opt = SymbolOptional(*action_it);
            min_coverage = 0;
            best_rule_ptr = RulePtr();
        }
    }

    SymbolOptional action_to_explore;
    if (forced || (min_coverage < visited_threshold)) {
        if (best_rule_ptr) {
            CINFO("explore") << "Selecting rule " << best_rule_ptr->name << " with min coverage " << min_coverage << NOENDL;
            if (DEBUG > 2)
                CINFO("explore") << "\033[1;35m" << "Best rule is" << *best_rule_ptr << "\033[0m";
            RuleSet grounded_rules = (*static_cast<const SymbolicRule*>(&(*best_rule_ptr))).get_grounded_rules(scenario.get_state());
            std::random_shuffle ( grounded_rules.begin(), grounded_rules.end() );
            action_to_explore = SymbolOptional(*grounded_rules.front());
            CINFO("explore") << " (best grounding: " << *action_to_explore << ")";
        }
        else {
            CINFO("explore") << "\033[1;35m" << "There is no best rule" << "\033[0m";
            if (selected_action_opt) {
                CINFO("explore") << " (random grounding for action " << *selected_action_opt << " )";
                action_to_explore = SymbolOptional(*selected_action_opt);
                std::vector< std::string > params = scenario.get_state().get_variables();
                std::random_shuffle ( params.begin(), params.end() );
                action_to_explore->ground(params);
            }
            else { // no safe action to explore, explore a random grounded action
                RuleSet grounded_rules = scenario.get_rules().get_grounded_rules(scenario.get_state());

                if (grounded_rules.size() >0) {
                    CINFO("explore") << " (random grounding)";
                    std::random_shuffle ( grounded_rules.begin(), grounded_rules.end() );
                    action_to_explore = SymbolOptional(*grounded_rules.front());
                }
                else {
                    if (!scenario.get_rules().empty()) {
                        CERROR("explore") << " TODO (random)";
                    }
                    action_to_explore = SymbolOptional(/* TODO CREATE ACTION */);
                    std::vector< std::string > params = scenario.get_state().get_variables();
//                     action_to_explore->ground(params);
                }
            }
        }
    }
    return action_to_explore;
}

