/*
 * <one line to give the program's name and a brief idea of what it does.>
 * Copyright (C) 2014  <copyright holder> <email>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "rexd_scenario.h"

RexdScenario::RexdScenario(const RexdScenario& rexd_scenario): 
    Scenario(rexd_scenario),
    noop_reward(rexd_scenario.noop_reward),
    planning_horizon(rexd_scenario.planning_horizon),
    use_rmax(rexd_scenario.use_rmax),
    use_vmin(rexd_scenario.use_vmin),
    vmin_reward(rexd_scenario.vmin_reward)
{
    if (use_rmax) {
        set_rmax(rexd_scenario.transitions_ptr);
    }
}

RexdScenario::RexdScenario(const Scenario& scenario): 
    Scenario(scenario),
    use_rmax(false),
    use_vmin(false)
{
    this->noop_reward = config_reader.get_variable<float>("planning_noop_reward");
    this->planning_horizon = config_reader.get_variable<uint>("planning_horizon");
    
//     if (config_reader.get_variable<bool>("final_reward_action")) {
//         RuleSet finish_rule(config_reader.get_path("nid_finish_rule_path"));
//         rules.add_rules(finish_rule);
//         finish_rule.index(domain); // Not needed?
//         // -finished() predicate used by final reward action
//         rules.add_predicate_to_preconditions(Predicate(MAKE_SYMBOL_FINISHED, false));
//     }
}


void RexdScenario::set_rmax(const TransitionGroupPtr transitions_ptr_)
{
    use_rmax = false;
    if (config_reader.get_variable<bool>("use_rmax")) {
        this->use_rmax = true;
        this->rmax_reward = config_reader.get_variable<float>("rmax_reward");
        this->explore_threshold = config_reader.get_variable<uint>("exploration_known_threshold");
        this->transitions_ptr = transitions_ptr_;
    }
}


void RexdScenario::set_vmin(double reward_modification)
{
    use_vmin = false;
    if (config_reader.get_variable<bool>("use_vmin")) {
        if (!config_reader.get_variable<bool>("final_reward_action"))
            LOG(WARNING) << "V-MIN requires final_reward_action to be enabled.";
        this->use_vmin = true;
        this->vmin_reward = config_reader.get_variable<float>("vmin_reward") + reward_modification;
    }
}


void RexdScenario::set_planning_options(const float& noop_reward_, const uint planning_horizon_)
{
    this->noop_reward = noop_reward_;
    this->planning_horizon = planning_horizon_;
}


void RexdScenario::prepare_ppddl_rules_and_state(RuleSetPtr& complete_rules_ptr, StatePtr& state_ppddl_ptr) const
{
    complete_rules_ptr = boost::make_shared<RuleSet>(rules);
    state_ppddl_ptr = boost::make_shared<State>(state);
    
    // RMAX explore actions
    if (use_rmax) {
        Explorer explorer(*this, transitions_ptr);
        complete_rules_ptr = explorer.get_rmax_explore_rules(explore_threshold, rmax_reward);
    }
    
    complete_rules_ptr->remove_nop_rules(); // cannot do below, would remove min reward teacher action from rmax_vmin

    // V-MIN teacher action
    if (use_vmin) {
        // -started() required so teacher can just be called as the first rule
        complete_rules_ptr->add_predicate_to_outcomes(Predicate(MAKE_SYMBOL_STARTED, true));
        state_ppddl_ptr->update_predicate(Predicate(MAKE_SYMBOL_STARTED, false));
        // V-MIN Teacher action
        RulePtr rule_ppddl_teacher = create_rule_ppddl_teacher(vmin_reward);
        complete_rules_ptr->push_back(rule_ppddl_teacher);
    }
    
    complete_rules_ptr = boost::make_shared<RuleSet>(complete_rules_ptr->get_full_grounded_rules(state));
    complete_rules_ptr->PPDDL_grounding();
}


void RexdScenario::create_PPDDL_scenario(RuleSetPtr& rules_ppddl_ptr, StatePtr& state_ppddl_ptr, GoalGroupPtr& goals_ppddl_ptr, DomainPtr& domain_ppddl_ptr) const
{
    prepare_ppddl_rules_and_state(rules_ppddl_ptr, state_ppddl_ptr);

    // get PPDDL grounded goal
    goals_ppddl_ptr = boost::make_shared<GoalGroup>(goals);
    goals_ppddl_ptr->PPDDL_grounding();

    // get PPDDL domain.
    domain_ppddl_ptr = DomainPtr(new Domain);
    rules_ppddl_ptr->index(*domain_ppddl_ptr);
    domain_ppddl_ptr->index_predicates(goals_ppddl_ptr->get_predicates());

    //optional: adding state. Info about the rules may be enough (exception if excuses available)
    state_ppddl_ptr = boost::make_shared<State>(state); // copy
    state_ppddl_ptr->PPDDL_grounding();
    domain_ppddl_ptr->index_predicates(*state_ppddl_ptr);

    // add special rules
    RulePtr rule_ppddl_noop = create_rule_ppddl_noop(domain_ppddl_ptr, goals_ppddl_ptr, noop_reward);
    rules_ppddl_ptr->push_back(rule_ppddl_noop);
}


void RexdScenario::write_PPDDL(const std::string template_file, std::ostream &out) const
{
    RuleSetPtr rules_ppddl_ptr;
    StatePtr state_ppddl_ptr;
    GoalGroupPtr goal_ppddl_ptr;
    DomainPtr domain_ppddl_ptr;
    create_PPDDL_scenario(rules_ppddl_ptr, state_ppddl_ptr, goal_ppddl_ptr, domain_ppddl_ptr);

    // write init and domain
    out << utils::get_file_contents(template_file, "<init>" , "<pre_domain>");
    domain_ppddl_ptr->write_PPDDL(out);

    // write rules
    std::map<std::string, RuleSet> rules_per_action = rules_ppddl_ptr->get_rules_per_action();
    write_actions_PPDDL(out, rules_per_action, *goal_ppddl_ptr);

    // write problem vars and end
    out << utils::get_file_contents(template_file, "<post_rules>" , "<horizon>");
    out << planning_horizon;
    out << utils::get_file_contents(template_file, "<horizon>" , "<end>");
    out << std::endl;
}


void RexdScenario::write_actions_PPDDL(std::ostream& out, std::map<std::string, RuleSet> &rules_per_action, const GoalGroup& goals_ppddl) const
{
    // TODO HACK IMPROVE!!!

    // Several actions
    for (std::map<std::string, RuleSet>::iterator action_it = rules_per_action.begin(); action_it != rules_per_action.end(); ++action_it) {
        GoalGroup rule_goals;
        out << "(:action " << action_it->first << std::endl;
        out << "  :effect (and" << std::endl;

        // Action has several rules
        for (RuleSet::iterator rule_it = action_it->second.begin(); rule_it != action_it->second.end(); ++rule_it) {
            // Action has several outcomes
            if (!(*rule_it)->outcomes.empty()) {
                if (!(*rule_it)->preconditions.empty()) {
                    out << "    (when ";
                    (*rule_it)->preconditions.write_PPDDL(out);
                }
                out << "(probabilistic ";
                for (OutcomeList::const_iterator outcome_it = (*rule_it)->outcomes.begin(); outcome_it != (*rule_it)->outcomes.end(); ++outcome_it) {
                    if (outcome_it->changes_predicates()) {
                        if (outcome_it->is_useful()) {
                            outcome_it->write_PPDDL(out);
                        }
                    }
                }
                out << ")";
                if (!(*rule_it)->preconditions.empty()) {
                    out << ")" << std::endl;
                }
            }

            if ((*rule_it)->goals.size() > 0) { // use rule goals
                rule_goals.insert(rule_goals.end(), (*rule_it)->goals.begin(), (*rule_it)->goals.end());
            }
        }
//         utils::remove_duplicates(rule_goals);
        rule_goals.write_PPDDL(out);
        if ((action_it->first != "finish_action") && (action_it->first != "noop"))
            goals_ppddl.write_PPDDL(out);
        out << "  )" << std::endl;
        out << ")" << std::endl;
    }
}


RulePtr RexdScenario::create_rule_ppddl_noop(const DomainPtr domain_ptr, const GoalGroupPtr goals_ppddl_ptr, const float noop_reward)
{
    RulePtr rule_noop_ptr = RulePtr(new SymbolicRule(Symbol("noop()"), PredicateGroup()));

    // obtain domain positive and negative predicates
    PredicateGroup positive_known_predicates, negated_known_predicates;
    for (std::vector<Predicate>::const_iterator pred_it = domain_ptr->known_predicates.begin(); pred_it != domain_ptr->known_predicates.end(); ++pred_it) {
        Predicate predicate(*pred_it);
        predicate.positive = true;
        positive_known_predicates.push_back(predicate);
        predicate.positive = false;
        negated_known_predicates.push_back(predicate);
    }
    positive_known_predicates.remove_duplicated();
    negated_known_predicates.remove_duplicated();
    rule_noop_ptr->preconditions.add_predicates(positive_known_predicates);
    rule_noop_ptr->preconditions.add_predicates(negated_known_predicates);
    Outcome outcome_pos(positive_known_predicates, 0.5);
    Outcome outcome_neg(negated_known_predicates, 0.5);
    rule_noop_ptr->add_outcome(outcome_pos);
    rule_noop_ptr->add_outcome(outcome_neg);

    rule_noop_ptr->goals = *goals_ppddl_ptr;
    Goal noop_goal(noop_reward); // noop reward is small
//     rule_noop_ptr->goals.push_back(goal_ppddl_ptr);
    rule_noop_ptr->goals.push_back(noop_goal);

    return rule_noop_ptr;
}



RulePtr RexdScenario::create_rule_ppddl_teacher(const float min_reward)
{
    RulePtr rule_teacher_ptr = RulePtr(new SymbolicRule(MAKE_SYMBOL_TEACHER, PredicateGroup("-started() -finished()")));
    rule_teacher_ptr->outcomes.push_back(Outcome("1.0 started() finished()"));
    Goal teacher_goal(PredicateGroup("-started() -finished()"), min_reward);
    rule_teacher_ptr->goals.push_back(teacher_goal);
    return rule_teacher_ptr;
}

void RexdScenario::write_PPDDL_to_file(const std::string template_file, const std::string file_path) const
{
    int DEBUG = 0;
    std::ofstream rules_file(file_path.c_str());
    if (DEBUG > 0)
        std::cout << "RuleSet::write_rules_to_file - " << "Writting rules to file: '" << file_path << "'" << std::endl;
    if (rules_file.is_open()) {
        write_PPDDL(template_file, rules_file);
    } else
        std::cerr << "Couldn't write rules to " << file_path << std::endl;
    rules_file.close();
}


/**
 * RDDL TODO
 */
/*void Rule::get_rddl_outcome_preconditions(std::map<Predicate, std::vector<Outcome> >& rddl_outcomes)
{
        for ( OutcomeList::iterator out_it=outcomes.begin() ; out_it != outcomes.end(); ++out_it ) {
                for (PredicateList::iterator pred_it=out_it->begin(); pred_it!=out_it->end(); ++pred_it) {
                        Outcome new_precondition(preconditions, out_it->probability);
                        new_precondition.push_back(Predicate(Symbol(name,param_names), true));
                        rddl_outcomes[*pred_it].push_back(new_precondition);
                }
        }
}

std::ostream& RuleSet::writeRDDL(std::ostream &out)
{
        // An outcome is a predicate with a probability. Using it to store preconditions and probability
        std::map<Predicate, std::vector<Outcome> > rddl_outcomes;
        for (RuleList::iterator rule_it = begin(); rule_it != end(); ++rule_it) {
                (*rule_it)->get_rddl_outcome_preconditions(rddl_outcomes);
        }


//         for predicate in domain
//                 add if rddl_outcomes[predicate] then probability true
//                 add if rddl_outcomes[-predicate] then probability false
//                 else same value

        out << "(:action ";
//         out << name;
        out << "\n";
        out << "CONTEXT:\n";
//         out << r.preconditions;
        out << "\n";
        out << "OUTCOMES:\n";
//      for ( OutcomeList::iterator it=r.outcomes.begin() ; it != r.outcomes.end(); it++ )
        out << "\n";

//        (pendulumplaced_pn') = (if (pendulumplaced_pn) then (KronDelta true) else (if (placependulum) then (Bernoulli (ACTION_SUCCESS)) else (KronDelta false)));

        return out;

//         for ( size_t i = 0; i < size(); ++i) {
//                 for ( size_t j = 0; j < (*this)[i]->outcomes.size(); ++j) {
//                         Outcome outcome((*this)[i]->outcomes.size());
//                         for (PredicateList::iterator pred_it = outcome.begin(); pred_it != outcome.end(); ++pred_it) {
//                                 out << "(" << pred_it->name << "') = (";
//                                 if ((*this)[i]->preconditions.size() > 0) {
//                                         out << "if (";
//                                         write_PPDDL_predicategroup((*this)[i]->preconditions, out);
//                                         out << ") then ";
//                                 }
//                                 if (outcome.probability == 1.0) {
//                                         out << "(KronDelta true)";
//                                 }
//                                 else {
//                                         out << "(Bernoulli ACTION_SUCCESS"  << i << "_" << j << ")";
//                                 }
//                                 out << ");" << std::endl;
//                         }
//                 }
//         }
}

std::ostream& RuleSet::write_RDDL_success_ratios(std::ostream& out) const
{
        for ( size_t i = 0; i < size(); ++i) {
                for ( size_t j = 0; j < (*this)[i]->outcomes.size(); ++j) {
                        out << "ACTION_SUCCESS" << i << "_" << j << " : {non-fluent, real, default = " << (*this)[i]->outcomes[i].probability << "};" << std::endl;
                }
        }
        return out;
}

std::ostream& RuleSet::write_RDDL_predicategroup(const PredicateGroup& predicates, std::ostream& out) const
{
        out << "(and ";
        for (PredicateList::const_iterator pred_it = predicates.begin(); pred_it != predicates.end(); ++pred_it) {
                if (!pred_it->positive) {
                        out << "(not ";
                }
                out << pred_it->name;
                if (!pred_it->positive) {
                        out << ")";
                }
        }
        out << ")";
        return out;
}*/

