/*
 * Scenario class that has all information needed to create planning models to be used by REX-D.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef REXDSCENARIO_H
#define REXDSCENARIO_H

#include <symbolic/scenario.h>
#include <rexd/explorer.h>

/**
 * \class RexdScenario
 * \brief Extended Scenario to generate PPDDL domains to use with IppcPlanner.
 * 
 * Writes PPDDL domains with all extra rules and predicates needed for R-MAX and V-MIN algorithms.
 */
class RexdScenario : public Scenario
{
public:
    /**
     * \brief Creates RexdScenario. R-MAX and V-MIN are disabled by default.
     */
    RexdScenario(const Scenario& scenario);
    
    /**
     * \brief Creates RexdScenario. R-MAX and V-MIN settings are copied. New object uses pointer to same transitions data.
     */
    RexdScenario(const RexdScenario& rexd_scenario);
    
    /**
     * \brief Initializes V-MIN if enabled in config file.
     */
    void set_vmin(double reward_modification = 0.0);
    
    /**
     * \brief Initializes R-MAX if enabled in config file. Transitions are not copied, just using the pointer to the same data.
     */
    void set_rmax(const TransitionGroupPtr transitions_ptr_);
    
    /**
     * \brief Sets config variables used when planning.
     */
    void set_planning_options(const float& noop_reward_, const uint planning_horizon_);

    void write_PPDDL_to_file(const std::string template_file, const std::string file_path) const;
    
    
protected:
    void prepare_ppddl_rules_and_state(RuleSetPtr& rmax_rules_ptr, StatePtr& state_ppddl_ptr) const;
    
private:
    // planning stuff
    float noop_reward;
    uint planning_horizon;
    
    // rmax, vmin stuff
    bool use_rmax;
    bool use_vmin;
    uint explore_threshold;
    float rmax_reward;
    float vmin_reward;
    TransitionGroupPtr transitions_ptr;
    
    static RulePtr create_rule_ppddl_teacher(const float min_reward);
    
    void create_PPDDL_scenario(RuleSetPtr& rules_ppddl_ptr, StatePtr& state_ppddl_ptr, GoalGroupPtr& goals_ppddl_ptr, DomainPtr& domain_ppddl_ptr) const;

    void write_PPDDL(const std::string template_file, std::ostream &out) const;
    
    void write_actions_PPDDL(std::ostream& out, std::map<std::string, RuleSet> &rules_per_action, const GoalGroup& goals) const;

    static RulePtr create_rule_ppddl_noop(const DomainPtr domain_ptr, const GoalGroupPtr goals_ppddl_ptr, const float noop_reward);
};

#endif // RMAXSCENARIO_H
