#include "rexd/rl_main.h"

_INITIALIZE_EASYLOGGINGPP

ReinforcementLearner::ReinforcementLearner():
    transition_obtained(false),
    accumulated_reward(0),
    current_problem(1),
    current_episode(1),
    current_episode_iteration(0),
    total_number_iterations(0)
{
}

bool ReinforcementLearner::init(const std::string& config_path, const std::string& files_path)
{
    bool init_ok = config_reader.init(config_path, files_path);
    utils::init_loggers(files_path + "generic/logger.conf");
    LOG(INFO) << "Starting new execution.\n\n";

    // generated random seed
    srand((unsigned)time(0));

    // Load rules
    RuleSet rules = RuleSet(config_reader.get_path("nid_rules_path"));
    
    // load goals
    GoalGroup goals = GoalGroup(config_reader.get_path("goal_file"));

    // load action list if needed
    std::vector<Symbol> action_list;
    if (!config_reader.get_variable<bool>("teacher")) { 
        // if no teacher, the list of actions have to be provided to learn
        action_list = read_action_list("symbols.dat");
        LOG(INFO) << "Action list is " << action_list;
    }
    
    scenario_ptr = ScenarioPtr(new Scenario(rules, State(), goals, config_reader, action_list));
    
    transitions = read_transitions_from_file(config_reader.get_path("transitions_backup_path"));
    scenario_ptr->domain.index_transitions(transitions);
    
    // TODO read saved dead end excuses from previous execution
//     dead_end_excuses = PredicateGroup("-broken(71) -broken(72) -broken(73) -broken(74) -broken(75) -broken(76) broken(71) broken(72) broken(73) broken(74) broken(75) broken(76) ");
//     rules_ptr->set_dangerous_rules(dead_end_excuses);

    // load planner and learner
    high_level_planner = boost::make_shared<HighLevelPlanner>(HighLevelPlanner(config_reader, &logging_file));
    learner = boost::make_shared<PasulaLearner>(PasulaLearner(config_reader));

    // load auto teacher if needed
    if (config_reader.get_variable<bool>("auto_teacher")) {
        auto_teacher = AutoTeacher(config_reader);
    }

    visualizer.init(config_reader);

    return init_ok;
}

bool ReinforcementLearner::main_iteration()
{
    if (config_reader.get_variable<bool>("monitoring")) {
        return main_monitoring_iteration();
    }
    else {
        return main_learning_iteration();
    }
}


bool ReinforcementLearner::reset_rexd()
{
    bool result;
    // total_number_iterations--; // special iteration, doesn't count as normal iteration -> anyway for visualization is fine to have a small pause
    logging_file << "Reward: " << accumulated_reward << std::endl;

    result = reset_scenario();

    visualizer.write_episode_log(config_reader.get_variable<std::string>("scenario"), current_problem, current_episode, accumulated_reward, total_number_iterations);
    accumulated_reward = 0;
    current_episode_iteration = 0;
    return result;
}


bool ReinforcementLearner::main_learning_iteration()
{
    LOG(INFO) << "\n" << "********************************************************"; // 2 newlines
    // new iteration
    total_number_iterations++;
    current_episode_iteration++;

    // get state
    show_info_request_state();
    previous_state = scenario_ptr->get_state();
    scenario_ptr->set_state(read_state());

    // get reward
    last_reward = read_reward();
    accumulated_reward += last_reward;

    // check infinite loops
    if (current_episode_iteration > 100) { // If this episode > 100 iterations, suppose things have gone wrong
        transition_obtained = false;
        logging_file << "Dead." << std::endl;
        return reset_rexd();
    }

    // check goal
    if (scenario_ptr->is_in_goal()) {
        if (transition_obtained) { // learn last action when arriving to the goal
            learning();
        }
        LOG(INFO) << START_COLOR_GREEN << "IS GOAL! Reward: " << accumulated_reward << END_COLOR;
        transition_obtained = false;
        show_info_is_goal();
        return reset_rexd();
    }
    else { // normal behaviour (learn and plan)
        learning();
        planning();

        return true;
    }
}


// WARNING This method is a infinite loop
bool ReinforcementLearner::main_monitoring_iteration()
{
    LOG(INFO) << "\n" << "********************************************************"; // 2 newlines

    // get state
    show_info_request_state();
    previous_state = scenario_ptr->get_state();
    scenario_ptr->set_state(read_state());
    read_reward();

    // check goal
    if (scenario_ptr->is_in_goal()) {
        LOG(INFO) << START_COLOR_GREEN << "IS GOAL!" << END_COLOR;
        transition_obtained = false;
        show_info_is_goal();

        // wait for action recognition
        read_action();

        return true;
    }
    else { // normal behaviour (learn and plan)
        monitoring();

        return true;
    }
}


void ReinforcementLearner::update_transitions()
{
    int DEBUG = 0;
    if (transition_obtained) {
        State current_state = scenario_ptr->get_state();
        current_state.remove_special_predicates();
        previous_state.remove_special_predicates();
        Transition transition_full(previous_state, previous_action, current_state, last_reward);
        transitions.push_back(transition_full);
        if (DEBUG > 0)
            LOG(INFO) << "Writing transitions";
        transitions.write_transitions_to_file(config_reader.get_path("transitions_backup_path"));
    }
}

void ReinforcementLearner::learning()
{
    update_transitions();

    struct timeval timer_start = utils::start_timer();
    if ((config_reader.get_variable<bool>("learning")) && (transition_obtained) && (transitions.size() > 0) && (previous_action.name.size()>1) ) {
        CINFO("planning") << START_COLOR_BLUE << "Learning" << END_COLOR;
        show_info_start_learning();
        RuleSetPtr new_rules_ptr, removed_rules_ptr, added_rules_ptr;
        new_rules_ptr = learner->learn_action_rules(scenario_ptr->get_rules(), transitions, previous_action, removed_rules_ptr, added_rules_ptr);
        new_rules_ptr->set_dangerous_rules(dead_end_excuses);
        new_rules_ptr->write_rules_to_file(config_reader.get_path("nid_rules_path"));
        scenario_ptr->set_rules(*new_rules_ptr);

        show_info_end_learning(*removed_rules_ptr, *added_rules_ptr, transition_context);
        visualizer.write_learning_log(*removed_rules_ptr, *added_rules_ptr, transition_context, total_number_iterations);
    }
    else {
        CINFO("planning") << "Skipping learning (no transitions or no previous action)";
    }
    logging_file << "Time learning: " << utils::get_elapsed_time(timer_start) << std::endl;
}


void ReinforcementLearner::planning()
{
    show_info_start_planning();
    struct timeval timer_start = utils::start_timer();
    transition_obtained = true; // default
    expected_reward = 0; // default
    PredicateGroup excuse;
    std::vector< std::string > actions_requiring_excuse;
    CINFO("planning") << START_COLOR_BLUE << "Planning" << END_COLOR;

    // explore and plan
    SymbolOptional planning_res = high_level_planner->exploit_or_explore(*scenario_ptr, transitions, excuse, actions_requiring_excuse);
    logging_file << "Time planning: " << utils::get_elapsed_time(timer_start) << std::endl;
    show_info_end_planning(planning_res, boost::lexical_cast<std::string>(excuse));

    // If solution found execute it. Else teacher / explore
    if (planning_res) {
        execute_action(*planning_res);
    }
    else {
        if (config_reader.get_variable<bool>("teacher")) {
            teacher(excuse, actions_requiring_excuse);
        }
        else {
            planning_res = high_level_planner->explore(*scenario_ptr, transitions, true);
            execute_action(*planning_res);
        }
    }
    if ( planning_res && planning_res->is_special() ) {
        transition_obtained = false;
    }

    visualizer.write_status_log(accumulated_reward, expected_reward, total_number_iterations);
}

void ReinforcementLearner::teacher(const PredicateGroup& excuse, std::vector<std::string> actions_requiring_excuse)
{
    LOG(INFO) << START_COLOR_RED << "Teacher needed." << END_COLOR;
    LOG(INFO) << "Excuse is \"" << excuse << "\" requiring actions " << actions_requiring_excuse;
    show_info_teacher_request(boost::lexical_cast<std::string>(excuse), actions_requiring_excuse);
    transition_context = "Teacher";

    // obtain action from auto teacher or human teacher
    SymbolOptional action_opt;
    if (config_reader.get_variable<bool>("auto_teacher")) {
        action_opt = auto_teacher.get_action(scenario_ptr->get_state(), scenario_ptr->get_action_list(), actions_requiring_excuse);
        if ((!action_opt) || (*action_opt == MAKE_SYMBOL_HELP))
            action_opt = read_action();
    }
    else {
        action_opt = read_action();
    }

    // Special cases (retry, dead...)
    if ((!action_opt) || (*action_opt == MAKE_SYMBOL_RETRY) || (*action_opt == MAKE_SYMBOL_UNKNOWN)) {
        transition_obtained = false;
    }
    // if autoteacher request finish action (and has not been requested by rexd planner)
    //   then finish action reward < Vmin -> dead_end!
    else if ((*action_opt == MAKE_SYMBOL_DEAD) || (*action_opt == MAKE_SYMBOL_FINISH)) {
        transition_obtained = false;
        LOG(INFO) << START_COLOR_RED << "Dead end! Aborting iteration..." << END_COLOR;
        if (!excuse.empty()) {
            LOG(INFO) << "Adding " << excuse << " as dangerous problem.";
            logging_file << "Dangerous: " << excuse << std::endl;
            dead_end_excuses.add_predicates(excuse);
        }
        logging_file << "Dead end." << std::endl;
        if (config_reader.get_variable<bool>("final_reward_action")) {
            if (config_reader.get_variable<bool>("simulator")) {
                execute_action(MAKE_SYMBOL_FINISH, false); // execute as if it isn't teacher
            }
        }
        else {
            reset_rexd();
        }
    }
    else { // Execution
        scenario_ptr->add_action_to_action_list(*action_opt);
        previous_action = *action_opt;

        if (config_reader.get_variable<bool>("simulator")) {
            execute_action(*action_opt, true);
        }
        logging_file << "Teacher: " << *action_opt << std::endl;
    }
}

void ReinforcementLearner::monitoring()
{
    LOG(INFO) << "Planning";
    show_info_start_planning();

    TransitionGroup transitions;
    PredicateGroup excuse;
    std::vector<std::string> actions_requiring_excuse;
    SymbolOptional planning_res = high_level_planner->exploit_or_explore(*scenario_ptr, transitions, excuse, actions_requiring_excuse);
    std::string excuse_str = boost::lexical_cast<std::string>(excuse);

    if (!planning_res) {
        show_info_end_planning(SymbolOptional(), excuse_str);
        LOG(INFO) << "No action";
    }
    else {
        int actions_until_finish = high_level_planner->get_planner_ptr()->estimate_distance_to_finish(*scenario_ptr);
        LOG(INFO) << "Recomended action " << *planning_res << " (actions to finish : " << actions_until_finish << ")";
        show_info_end_planning(planning_res, excuse_str, actions_until_finish);
    }

    // Wait for manipulation recognition
    read_action();
}


std::vector<Symbol> ReinforcementLearner::read_action_list(std::string file_path)
{
    std::vector<Symbol> action_list;
    std::ifstream symbols_file(file_path.c_str());
    if (symbols_file.is_open()) {
        std::string line;
        getline (symbols_file,line);
        while ( symbols_file.good() ) {
            if (line.find("action")!=std::string::npos) {
                std::stringstream ss(line);
                std::string item,item2;
                std::getline(ss, item, ' ');

                std::stringstream params_ss, action_ss;
                std::getline(ss, item2, ' ');
                for (int i = 0; i < (atoi(item2.c_str())-1); ++i) { // add all params but last one
                    params_ss << "x,";
                }
                if (atoi(item2.c_str()) > 0) // last one without ,
                    params_ss << "y";
                action_ss << item << "(" << params_ss.str() << ")";
//                                 CLOG << "Adding action " << action.str() << std::endl;
                action_list.push_back(Symbol(action_ss.str()));
            }
            std::getline (symbols_file,line);
        }
    }
    return action_list;
}



