#include "symbolic/domain.h"


/**
 * OPERATORS
 *
 */

std::ostream& operator<<(std::ostream &out, const Domain& dom)
{
    out << "Domain::" << "<<" << " - " << "Domain is :\n";
    for (size_t i = 0; i < dom.known_predicates.size(); i++)
        out << "Domain::" << "<<" << " - " << "Predicate " << i << ": " << dom.known_predicates[i] << "\n";

    return out;
}

/**
 * Domain
 *
 */

Domain::Domain()
{ }

template<class T>
int Domain::index_atom(const T& atom, std::vector<T>& known_atoms)
{
    int idx;
    bool found = false;
    for (size_t i = 0; i < known_atoms.size(); i++) {
        if (atom == known_atoms[i]) {
            idx = i;
            found = true;
        }
    }
    if (!found) {
        idx = known_atoms.size();
        known_atoms.push_back(atom);
    }
    return idx;
}

template<class T>
int Domain::get_atom_index(const T& atom, const std::vector<T>& known_atoms) const
{
    for (size_t i = 0; i < known_atoms.size(); i++) {
        if (atom == known_atoms[i]) {
            return i;
        }
    }
    return -1;
}

void Domain::index_predicates(const PredicateGroup& predicates)
{
    for (PredicateList::const_iterator predicate = predicates.begin(); predicate != predicates.end(); ++predicate) {
        index_predicate(*predicate);
    }
}

void Domain::index_goals(const GoalGroup& goals)
{
    for (GoalGroup::const_iterator goal_it = goals.begin(); goal_it != goals.end(); ++goal_it) {
        index_predicates(*goal_it);
    }
}

void Domain::index_transitions(const TransitionGroup& transitions)
{
    for (TransitionGroup::const_iterator trans_it = transitions.begin(); trans_it != transitions.end(); ++trans_it) {
        index_predicates(trans_it->get_prev_state());
        index_predicates(trans_it->get_post_state());
        index_action(trans_it->action);
    }
}

int Domain::index_predicate(const Predicate& predicate)
{
    Predicate sym_predicate(predicate);
    sym_predicate.make_symbolic();
    int id = index_atom<Predicate>(predicate, known_predicates);

    // add objects if GroundedPredicate
    if (!predicate.is_symbolic) {
        for (std::vector<std::string>::const_iterator param_it = predicate.param_names.begin(); param_it != predicate.param_names.end(); ++param_it) {
            index_atom<std::string>(*param_it, known_objects);
        }
    }
    return id;
}

int Domain::index_action(const Symbol& action)
{
    Symbol generic_action(action);
    generic_action.make_symbolic();
    return index_atom<Symbol>(generic_action, known_actions);
}

size_t Domain::get_predicate_index(const Predicate& predicate)
{
    Predicate sym_predicate(predicate);
    sym_predicate.make_symbolic();
    int id = get_atom_index(predicate, known_predicates);
    if (id < 0) { // not found
        id = index_predicate(predicate);
    }
    return id;
}

size_t Domain::get_predicate_index_const(const Predicate& predicate) const
{
    Predicate sym_predicate(predicate);
    sym_predicate.make_symbolic();
    int id = get_atom_index(predicate, known_predicates);
    return id;
}

size_t Domain::get_action_index(const Symbol& action)
{
    Symbol generic_action(action);
    int id = get_atom_index(generic_action, known_actions);
    if (id < 0) { // not found
        id = index_action(action);
    }
    return id;
}

size_t Domain::get_object_index(const std::string& object)
{
    return get_atom_index(object, known_objects);
}

void Domain::write_PPDDL(std::ostream& out) const
{
    std::set<Symbol> predicate_symbol_set; // to write domain we don't need sign(+/-). Predicate is just a symbol with a sign(+/-)
    BOOST_FOREACH(const Predicate & predicate, known_predicates) {
        predicate_symbol_set.insert(predicate);
    }

    out << "(:predicates " << std::endl;
    BOOST_FOREACH(const Symbol & symbol, predicate_symbol_set) {
        symbol.write_PPDDL(out);
        out << std::endl;
    }
    out << ")" << std::endl;
}

void Domain::write_NID(std::ostream& out) const
{
    std::set<Symbol> predicate_symbol_set, action_symbol_set; // to write domain we don't need sign(+/-). Predicate is just a symbol with a sign(+/-)
    BOOST_FOREACH(const Predicate & predicate, known_predicates) {
        predicate_symbol_set.insert(predicate);
    }

    BOOST_FOREACH(const Symbol & action, known_actions) {
        action_symbol_set.insert(action);
    }

    out << "# predicates" << std::endl;
    BOOST_FOREACH(const Symbol & primitive, predicate_symbol_set) {
        out << primitive.name << " " << primitive.param_names.size() << " primitive binary" << std::endl;
    }

    out << "# actions" << std::endl;
    BOOST_FOREACH(const Symbol & action, action_symbol_set) {
        out << action.name << " " << action.param_names.size() << " action binary" << std::endl;
    }
}





