/*
 * Class that represents the predicates and rules domain. It stores the different existing objects and symbols.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_DOMAIN_H
#define IRI_RULE_LEARNER_DOMAIN_H 1

#include <vector>
#include <fstream>
#include <cmath>

#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>

#include <symbolic/symbols.h>
#include <symbolic/predicates.h>
#include <symbolic/transitions.h>
// #include <symbolic/rules.h>

/*!
\class Domain
\brief Class indexing all domain symbols, actions and objects.

At this moment it just indexes symbols, actions and objects. This indexing is just used to generate excuses, but it would be nice to use it in other parts also. If it was used for planning and learning, symbol arguments wouldn't have to be just one character or a number to be compatible with libprada implementation.
*/
class Domain
{
public:
    std::vector< Predicate > known_predicates;
    std::vector< Symbol > known_actions;
    std::vector< std::string > known_objects;

    Domain();

    void clear() {
        known_predicates.clear();
        known_actions.clear();
        known_objects.clear();
    }

    int index_predicate(const Predicate& predicate);
    int index_action(const Symbol& action);
    void index_predicates(const PredicateGroup& predicates);
    void index_goals(const GoalGroup& goals);
    void index_transitions(const TransitionGroup& transitions);

    size_t get_object_index(const std::string& object);
    size_t get_action_index(const Symbol& action);
    size_t get_predicate_index(const Predicate& predicate);
    size_t get_predicate_index_const(const Predicate& predicate) const;

    size_t get_predicates_size() {
        return known_predicates.size();
    }

    Predicate get_predicate(size_t index) {
        return known_predicates[index];
    }
    Symbol get_action(size_t index) {
        return known_actions[index];
    }
    std::string get_object(size_t index) {
        return known_objects[index];
    }

    void write_PPDDL(std::ostream& out) const;
    void write_NID(std::ostream& out) const;

    friend std::ostream& operator<<(std::ostream &out, const Domain& dom);

private:
    template<class T>
    int index_atom(const T& atom, std::vector<T>& known_atoms);
    template<class T>
    int get_atom_index(const T& atom, const std::vector<T>& known_atoms) const;
};

typedef boost::shared_ptr<Domain> DomainPtr;


#endif
