/*
 * Clases representing predicates and goals.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_PREDICATES_H
#define IRI_RULE_LEARNER_PREDICATES_H 1


#include <sys/types.h>
#include <vector>
#include <set>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>

#include <boost/lexical_cast.hpp>

#include <symbolic/symbols.h>

/*!
\class Predicate
\brief A Predicate is a Symbol with a sign.
*/
class Predicate : public Symbol
{
public:
    bool positive;

    Predicate(std::string name, std::vector<std::string> param_names, bool positive_) :
        Symbol(name, param_names),
        positive(positive_)
    { }

    Predicate(Symbol symbol, bool positive_):
        Symbol(symbol),
        positive(positive_)
    { }

    /**
     * \brief obtains predicate value as a string
     * 
     * WARNING: At this moment only bool values are supported
     */
    std::string get_value() const {
        if (positive)
            return "true";
        else
            return "false";
    }

    void negate() {
        positive = !positive;
    }

    Predicate get_negated() const {
        return Predicate(*this, !positive);
    }

    /**
     * \brief gets the predicate params that are symbolic
     * 
     * Symbolic params are the ones defined in ::DEFAULT_PARAM_NAMES : "X", "Y", "Z", "V", "W", "Q".
     */
    std::vector<std::string> get_symbolic_params() const;
    
    /**
     * \brief gets the predicates with all the possible parameter permutations given a list possible parameters
     * 
     * \param symbolic_params params from which permutations of parameters are created. For every permutation a new predicate is created. Default parameters are added if there are not enough parameters to fill the predicate parameters.
     * \return predicates having all the permutations as parameters
     */
    std::vector<Predicate> get_symbolic_predicates_permutations(const std::vector<std::string> symbolic_params) const;

    void write_PPDDL(std::ostream &out) const;

    /**
     * \brief compares two predicates as if they were symbolic, so parameter names are not compared
     */
    bool symbolic_compare(const Predicate& p) const;
    
    friend bool operator==(const Predicate &p1, const Predicate &p2);
    friend bool operator<(const Predicate& l, const Predicate& r);
    friend std::ostream& operator<<(std::ostream &out, const Predicate& p);
};

typedef std::vector<Predicate> PredicateList;
typedef boost::shared_ptr<Predicate> PredicatePtr;


/*!
\class PredicateGroup
\brief A vector of Predicate.
*/
class PredicateGroup : public PredicateList
{
public:
    PredicateGroup() { }
    PredicateGroup(const PredicateGroup& pred_group);
    PredicateGroup(const PredicateList& predicates);
    PredicateGroup(std::string predicates_);

    /**
     * \brief checks if current_predicates satisfy all predicates in PredicateGroup
     * 
     * \return True if PredicateGroup is satisfied by current_predicates
     */
    bool is_satisfied(const PredicateGroup current_predicates) const;

    /**
     * \brief checks if PredicateGroup satisfies the Predicate predicate
     * 
     * \return True if predicate is satisfied by PredicateGroup
     */
    bool satisfies(const Predicate predicate) const;

    inline void add_predicates(const PredicateGroup new_predicates) {
        insert(end(), new_predicates.begin(), new_predicates.end());
    }
    
    /**
     * \brief Removes current predicates and adds the new ones.
     */
    inline void replace_predicates(const PredicateGroup new_predicates) {
        clear();
        add_predicates(new_predicates);
    }

    /**
     * \brief Count number of predicates satisfied by current_predicates
     */
    int count_satisfied(PredicateGroup current_predicates) const;

    PredicateGroup get_not_satisfied(const PredicateGroup predicates) const;

    int number_parameters() const;
    
    bool has_predicate(const Predicate& predicate) const;

    void change_variable(const std::string& orig_var, const std::string& new_var);
    
    /**
     * \brief Adds new_predicates and removes their negated counterparts if exists
     */
    void update_predicates(const PredicateGroup& new_predicates);
    
    /**
     * \brief Adds new_predicate and removes its negated counterpart if exists
     */
    void update_predicate(const Predicate& new_predicate);

    /*!
     * \brief Makes the predicates symbolic
     *
     * Changes the params to be the default symbolic params: X, Y, Z, V, W, Q.
     */
    void make_symbolic();
    std::vector<std::string> get_variables() const;
    std::vector<int> get_numeric_variables() const;
    PredicateGroup get_predicates_with_var(const std::string variable) const;
    PredicateGroup get_predicates_with_vars(const std::vector< std::string > variables) const;

    // Comparisons and differences
    /**
     * \brief Add differences with the compared group that are positive
     */
    void add_positive_differences(PredicateGroup compared_group);

    /**
     * \brief Get differences with compared PredicateGroup
     *
     *  Adds predicates in compared_group that are not present in current predicates.
     * Predicates not present in compared_group will be added negated.
     */
    PredicateGroup get_differences(const PredicateGroup compared_group) const;

    /**
     * \brief gets the predicates with all the possible parameter permutations given a list possible parameters
     * 
     * \param param_names params from which permutations of parameters are created. For every permutation a new predicate is created. Default parameters are added if there are not enough parameters to fill the predicate parameters.
     * \return predicates having all the permutations as parameters
     */
    PredicateGroup get_symbolic_predicates_permutations(const std::vector< std::string >& param_names) const;
    
    /**
     * \brief gets the symbolic versions of the predicates
     */
    PredicateGroup get_symbolic_predicates() const;
    
    /**
     * \brief gets the predicate params that are symbolic
     * 
     * Symbolic params are the ones defined in ::DEFAULT_PARAM_NAMES : "X", "Y", "Z", "V", "W", "Q".
     */
    std::vector<std::string> get_symbolic_params() const;

    void remove(const Predicate predicate);
    void remove(const PredicateGroup predicates_to_remove);
    void remove_duplicated();
    /** \brief removes duplicated predicates which are equal by a symbolic comparison (see Predicate::symbolic_compare ) */
    void remove_duplicated_symbolic();
    void remove_variables_not_present(std::vector<std::string> variables);

    PredicateGroup get_negatet_predicates() const;
    void remove_negated_predicates();

    void PPDDL_grounding();
    void write_PPDDL(std::ostream &out) const;
    void write_negated_PPDDL(std::ostream &out) const;

    void write_to_file(const std::string file_path) const;

    /**
     * Operators
     */
    friend std::ostream& operator<<(std::ostream &out, const PredicateGroup& p);
    friend bool operator==(const PredicateGroup &p1, const PredicateGroup &p2);
    friend bool operator<(const PredicateGroup& l, const PredicateGroup& r);
};


/*!
\class Outcome
\brief A PredicateGroup with an associated probability

The predicates define the changes made to the state when the outcome is applied. The probability defines the chances that this outcome is selected in a rule.
*/
class Outcome : public PredicateGroup
{
public:
    float probability;
    bool noisy;

    /**
     * Constructors
     */
    Outcome(std::string line);

    Outcome(float probability_, bool noisy_ = false) :
        PredicateGroup(),
        probability(probability_),
        noisy(noisy_)
    { }

    Outcome(PredicateGroup predicates_, float probability_) :
        PredicateGroup(predicates_),
        probability(probability_),
        noisy(false)
    { }

    /**
     * Helpers
     */
    double probability_obtain_predicate(const Predicate predicate) const;
    /** \brief true if changes any predicate */
    bool changes_predicates() const;
    /** \brief true if changes any predicate and has a probability > 0.0 */
    bool is_useful() const;
    /** \brief Changes predicates by updating them with its PredicateGroup */
    bool apply(PredicateGroup& predicates) const;

    void write_PPDDL(std::ostream &out) const;

    /**
     * Operators
     */
    friend std::ostream& operator<<(std::ostream &out, const Outcome& o);
    friend bool operator==(const Outcome &o1, const Outcome &o2);
    friend bool operator<(const Outcome& l, const Outcome& r);
};

typedef std::vector<Outcome> OutcomeList;
bool operator==(const OutcomeList &o1, const OutcomeList &o2);

/*!
\class OutcomeGroup
\brief A vector of outcomes
*/
class OutcomeGroup : public OutcomeList
{
public:
    /** \brief true if changes any predicate */
    bool changes_predicates() const;

    friend bool operator<(const OutcomeGroup& l, const OutcomeGroup& r);  // WARNING does not compare probabilities
};


/*!
\class Goal
\brief A vector of predicates with an associated reward.

The reward is obtained when the set of predicates are satisfied.
*/
class Goal : public PredicateGroup
{
public:
    float reward;

    /**
     * Constructors
     */
    Goal(float reward_ = 1.0) :
        PredicateGroup(),
        reward(reward_)
    { }

    Goal(PredicateGroup predicates_, float reward_ = 1.0) :
        PredicateGroup(predicates_),
        reward(reward_)
    { }

    Goal(const std::string& goal_string);

    /** \brief true state satisfies the PredicateGroup */
    bool is_goal(const PredicateGroup state) const {
        return is_satisfied(state);
    };
    
    /** \brief returns the goal reward  if predicates satisfies the PredicateGroup */
    float obtain_reward(const PredicateGroup& predicates) const;

    void write_PPDDL(std::ostream &out) const;

    /**
     * Operators
     */
    friend std::ostream& operator<<(std::ostream &out, const Goal& g);

    // TODO goal operators
//         friend bool operator==(const Goal &o1, const Goal &o2);
};


/*!
\class GoalGroup
\brief A vector of goals
*/
class GoalGroup : public std::vector<Goal>
{
public:
    GoalGroup() {}
    GoalGroup(std::string file_path);

    void PPDDL_grounding();

    bool is_goal(const PredicateGroup& state) const;
    float obtain_reward(const PredicateGroup& predicates) const;

    /** \brief Adapts a goal defined by a finish function to a goal defined by a set of predicates giving a positive reward when satisfied
     *
     * DEPRECATED
     * Just used for backward compatibility.
     * Gets a set of goals which are the rewards obtained by a finish function.\n
     *  - Ex: [+1.5 onTable(1), -1.0 free(2)]
     * \return A set of predicates which give positive rewards
     */
    Goal adapt_finish_rule_goal_to_simple_predicate_goal();

    PredicateGroup get_predicates() const;

    void write_PPDDL(std::ostream &out) const;

    friend std::ostream& operator<<(std::ostream &out, const GoalGroup& g);
};

typedef boost::shared_ptr<Goal> GoalPtr;
typedef boost::shared_ptr<const Goal> GoalConstPtr;

typedef boost::shared_ptr<GoalGroup> GoalGroupPtr;
typedef boost::shared_ptr<const GoalGroup> GoalGroupConstPtr;


// get data from string
Predicate get_predicate_from_string(std::string item);
PredicateGroup read_predicates_from_file(const std::string file_path);

#endif
