#include "symbolic/rules.h"

#define PASULA

using namespace std;

/**
 * OPERATORS
 *
 */

std::ostream& operator<<(std::ostream &out, const Rule& r)
{
    out << "ACTION: # " << r.identifier << "\n";
    out << *static_cast<const Symbol*>(&r);
    out << "\n";
    out << "CONTEXT:\n";
    out << r.preconditions;
    out << "\n";
    out << "OUTCOMES:\n";
    for (OutcomeGroup::const_iterator out_it = r.outcomes.begin() ; out_it != r.outcomes.end(); ++out_it) {
        out << *out_it << std::endl;
    }
    out << std::endl;

    return out;
}


std::ostream& operator<<(std::ostream &out, const RuleSet& r)
{
    for (RuleList::const_iterator rule_it = r.begin() ; rule_it != r.end(); ++rule_it) {
        out << **rule_it;
    }

    // if no rules, write default rule
    if (r.empty()) {
        out << "ACTION:" << std::endl << "  default()" << std::endl
            << "CONTEXT:" << std::endl << "  --" << std::endl
            << "OUTCOMES:" << std::endl << "  0.5 <no-change>" << std::endl << "  0.5 <noise>" << std::endl;
    }

    return out;
}

bool operator==(const Rule &r1, const Rule &r2)
{
    bool result;
    result = (*static_cast<const Symbol*>(&r1) == r2);
    result = (result && (r1.preconditions == r2.preconditions));
    result = (result && (r1.outcomes == r2.outcomes));
    return result;
}

bool operator<(const Rule& l, const Rule& r)
{
    if (l.name < r.name) {
        return true;
    } else if (r.name < l.name) {
        return false;
    } else if (l.outcomes < r.outcomes)
        return true;
    else if (r.outcomes < l.outcomes)
        return false;
    else if (l.preconditions < r.preconditions)
        return true;
    else if (r.preconditions < l.preconditions)
        return false;
    else
        return false;
}

bool operator>(const Rule& l, const Rule& r)
{
    return r < l;
}

bool operator<=(const Rule& l, const Rule& r)
{
    return ((l < r) || (!(l > r)));
}

bool operator>=(const Rule& l, const Rule& r)
{
    return ((l > r) || (!(l < r)));
}

bool operator==(const RuleSet &r1, const RuleSet &r2)
{
    if (r1.size() != r2.size()) {
        return false;
    }
    // make copies to sort them
    RuleSet r1_copy(r1), r2_copy(r2);
    std::sort(r1_copy.begin(), r1_copy.end());
    std::sort(r2_copy.begin(), r2_copy.end());
    RuleSet::iterator r1_it, r2_it;
    r2_it = r2_copy.begin();
    for (r1_it = r1_copy.begin(); r1_it != r1_copy.end(); ++r1_it) {
        if (**r1_it != **r2_it) {
            return false;
        }
        r2_it++;
    }
    return true;
}

bool operator!=(const RuleSet &r1, const RuleSet &r2)
{
    return !(r1 == r2);
}

// bool operator<(const RuleSet& l, const RuleSet& r );
// bool operator>(const RuleSet& l, const RuleSet& r );
// bool operator<=(const RuleSet& l, const RuleSet& r );
// bool operator>=(const RuleSet& l, const RuleSet& r );

/**
 * Rule
 *
 */

Rule::Rule(const Rule& rule) :
    Symbol(rule),
    preconditions(rule.preconditions),
    outcomes(rule.outcomes),
    goals(rule.goals),
    dangerous(rule.dangerous),
    identifier(rule.identifier)
{
}

Rule::Rule(const Symbol &symbol_, const PredicateList &preconditions_) :
    Symbol(symbol_),
    preconditions(preconditions_),
    dangerous(false)
{
    static ulong current_identifier = 0;
    identifier = current_identifier++;
}

void Rule::add_outcome(const PredicateList predicates, const float probability)
{
    outcomes.push_back(Outcome(predicates, probability));
}

void Rule::add_outcome(const Outcome outcome)
{
    outcomes.push_back(outcome);
}

double Rule::probability_obtain_predicate(const Predicate predicate) const
{
    double result = 0.0;
    for (OutcomeList::const_iterator outcome = this->outcomes.begin(); outcome != this->outcomes.end(); outcome++) {
        result += outcome->probability_obtain_predicate(predicate);
    }
    return result;
}

void Rule::sanitize_probabities()
{
    // Adds or reduces probability of the empty outcome to get a sum of probabilities = 1.0
    float sum_probs = 0.0;
    OutcomeList::iterator empty_outcome = this->outcomes.end();
    for (OutcomeList::iterator it = this->outcomes.begin(); it != this->outcomes.end(); it++) {
        sum_probs += it->probability;
        if (!it->is_useful()) {
            empty_outcome = it;
        }
    }
    if (sum_probs < 1.0) { // add missing probability to empty outcome
        if (empty_outcome == this->outcomes.end()) {
            outcomes.push_back(Outcome(1.0 - sum_probs));
        } else {
            empty_outcome->probability += 1.0 - sum_probs;
        }
    } else { // remove exceding probability from first outcome
        this->outcomes[0].probability -= (sum_probs - 1.0);
    }
}

bool Rule::has_useful_outcome() const
{
    bool result = false;
    for (OutcomeList::const_iterator outcome = this->outcomes.begin(); outcome != this->outcomes.end(); outcome++) {
        if (outcome->changes_predicates()) {
            result = true;
        }
    }
    return result;
}

TransitionList Rule::get_covering_transitions(const TransitionList transitions) const
{
    TransitionList result;
    for (TransitionList::const_iterator transition_it = transitions.begin() ; transition_it != transitions.end(); ++transition_it) {
        if (covers_transition_preconditions(*transition_it)) {
            result.push_back(*transition_it);
        }
    }
    return result;
}


uint Rule::count_covering_transitions_preconditions(const TransitionList& transitions) const
{
    int num = 0;
    for (TransitionList::const_iterator transition_it = transitions.begin() ; transition_it != transitions.end(); ++transition_it) {
        if (covers_transition_preconditions(*transition_it)) {
            ++num;
        }
    }
    return num;
}

int Rule::get_number_literals() const
{
    int result = 0;
    result += preconditions.size();
    for (OutcomeList::const_iterator outcome_it = this->outcomes.begin(); outcome_it != this->outcomes.end(); outcome_it++) {
        result += outcome_it->size();
    }
    return result;
}

bool Rule::has_outcome_predicate(const Predicate& predicate) const
{
    bool result = false;
    for (OutcomeList::const_iterator outcome_it = this->outcomes.begin(); outcome_it != this->outcomes.end(); outcome_it++) {
        if (outcome_it->has_predicate(predicate) && (outcome_it->probability > 0.0)) {
            result = true;
        }
    }
    return result;
}

void Rule::add_negated_outcomes_as_preconditions()
{
    for (OutcomeList::const_iterator outcome_it = this->outcomes.begin(); outcome_it != this->outcomes.end(); outcome_it++) {
        for (PredicateList::const_iterator predicate = outcome_it->begin(); predicate != outcome_it->end(); ++predicate) {
            preconditions.push_back(predicate->get_negated());
        }
    }
    preconditions.remove_duplicated();
}

void Rule::add_predicate_to_preconditions(const Predicate& new_predicate)
{
    preconditions.update_predicate(new_predicate);
}

void Rule::add_predicate_to_outcomes(const Predicate& new_predicate)
{
    BOOST_FOREACH(Outcome & outcome, outcomes) {
        outcome.update_predicate(new_predicate);
    }
}

void Rule::set_reward(const float reward)
{
    if (reward != 0.0) {
        Goal goal(PredicateGroup(), reward);
        goals.push_back(goal);
    }
}




/**
 * SymbolicRule
 *
 */

SymbolicRule::SymbolicRule(const Rule& rule):
    Rule(rule)
{
    is_symbolic = true;
}

SymbolicRule::SymbolicRule(const SymbolicRule& rule):
    Rule(rule)
{
    is_symbolic = true;
}

SymbolicRule::SymbolicRule(const Symbol &symbol_, const PredicateList &preconditions_):
    Rule(symbol_, preconditions_)
{
    is_symbolic = true;
}

inline bool SymbolicRule::covers_transition_preconditions(const Transition& transition) const
{
    if (transition.action.name == name) {
        GroundedRule grounded_rule(*this, transition.action.param_names);
        return grounded_rule.covers_transition_preconditions(transition);
    } else
        return false;
}

std::vector<int> SymbolicRule::experiences_per_outcome(const Transition transition) const
{
    if (transition.action.name == name) {
        GroundedRule grounded_rule(*this, transition.action.param_names);
        return grounded_rule.experiences_per_outcome(transition);
    } else {
        std::cerr << "WARNING: SymbolicRule::experiences_per_outcome should be called only if action name and preconditions are satisfied." << std::endl;
        return std::vector<int>();
    }
}

bool SymbolicRule::satisfiesPreconditions(const PredicateGroup predicate_group) const
{
    bool prec_satisfied = false;
    RuleList grounded_rules = get_grounded_rules(predicate_group);

    RuleList::iterator grule_it = grounded_rules.begin();
    while (!prec_satisfied && (grule_it != grounded_rules.end())) {
        if ((*grule_it)->satisfiesPreconditions(predicate_group)) {
            prec_satisfied = true;
        }
        ++grule_it;
    }
    return prec_satisfied;
}

RuleList SymbolicRule::get_grounded_rules(const PredicateGroup predicate_group, const std::vector< std::vector <int> >& variables_per_param, bool remove_invalid) const
{
    int DEBUG = 0;
    RuleList rules;

    // try all combinations and mantain only correct ones
    std::vector< std::vector< string > > combinations_of_params;
    for (size_t i = 0; i < variables_per_param.size(); ++i) { // for each parameter ...(X, Y, ...) || i=0 -> X, i=1 -> Y, ...
        std::vector< std::vector< string > > new_combinations_of_params;
        for (size_t j = 0; j < variables_per_param[i].size(); ++j) {
            std::vector< string > new_params;
            std::string current_value = boost::lexical_cast<std::string>(variables_per_param[i][j]);
            if (DEBUG > 2)
                LOG(INFO) << "i(" << i << ") is " << variables_per_param[i] << " and i,j(" << i << "," << j << ") is " << current_value;

            if (i == 0) { // if it is first parameter (X)
                new_params.push_back(current_value);
                new_combinations_of_params.push_back(new_params);
            } else { // 2°+ parameters (Y, Z, ...), add new parameter to all previous combinations
                for (size_t k = 0; k < combinations_of_params.size(); ++k) {
                    // don't add the same value as argument twice
                    if (find(combinations_of_params[k].begin(), combinations_of_params[k].end(), current_value) == combinations_of_params[k].end()) {
                        // add when value was not added previously
                        new_params = combinations_of_params[k];
                        new_params.push_back(current_value);
                        new_combinations_of_params.push_back(new_params);
                    }
                }
            }
        }
        combinations_of_params = new_combinations_of_params;
    }

    if (DEBUG > 1)
        LOG(INFO) << "Combinations of params are " << combinations_of_params;

    for (size_t i = 0; i < combinations_of_params.size(); ++i) {
        GroundedRulePtr new_rule = boost::make_shared<GroundedRule>(*this, combinations_of_params[i]);
//         new_rule->ground_deictic_references(predicate_group);
        RuleSet new_rules = new_rule->get_deictic_grounded_rules(predicate_group, true);

        if (DEBUG > 1)
            LOG(INFO) << "Grounded rules are: " << new_rules;

        for (RuleList::iterator rule_it = new_rules.begin(); rule_it != new_rules.end(); ++rule_it) {
            if (!remove_invalid) {
                rules.push_back(*rule_it);
            } else if ((*rule_it)->satisfiesPreconditions(predicate_group)) {
                rules.push_back(*rule_it);
            }
        }
    }

    if (DEBUG > 0)
        LOG(INFO) << "Number of grounded rules is " << rules.size();

    return rules;
}

RuleList SymbolicRule::get_grounded_rules(const PredicateGroup predicate_group, const bool full) const
{
    int DEBUG = 0;
    if (DEBUG > 0)
        LOG(INFO) << "Grounding for rule " << name << " with state: " << predicate_group;

    // if there are no params -> no need for grounding
    if (param_names.size() == 0) {
        RuleList rule_list;
        std::vector<std::string> params;
        rule_list.push_back(boost::make_shared<GroundedRule>(GroundedRule(*this, params)));
        return rule_list;
    }

    std::vector<int> variables = predicate_group.get_numeric_variables();
    std::map< std::string, std::vector< std::vector<int> > > map_variables_candidates;
    std::map< std::string, std::vector<int> > map_variables;

    // Add to map_variables_candidates the possible arguments that can be passed to each parameter in preconditions predicates
    for (PredicateList::const_iterator precondition_it = preconditions.begin(); precondition_it != preconditions.end(); ++precondition_it) {
        for (size_t i = 0; i < precondition_it->param_names.size(); ++i) {
            if (full) {
                map_variables_candidates[precondition_it->param_names[i]].push_back(variables);
            } else if (precondition_it->positive) {
                std::vector<int> possible_variables;
                for (PredicateList::const_iterator predicate_it = predicate_group.begin(); predicate_it != predicate_group.end(); ++predicate_it) {
                    if ((precondition_it->name == predicate_it->name) && (precondition_it->positive == predicate_it->positive)) {
                        possible_variables.push_back(atoi(predicate_it->param_names[i].c_str()));
                    }
                }
                map_variables_candidates[precondition_it->param_names[i]].push_back(possible_variables);
            } else { // TODO have better strategy than just adding all
                map_variables_candidates[precondition_it->param_names[i]].push_back(variables);
            }
        }
    }
    if (full) {
        for (vector_string::const_iterator param_it = param_names.begin(); param_it != param_names.end(); ++param_it) {
            map_variables_candidates[*param_it].push_back(variables);
        }
    }

    for (std::map< std::string, std::vector< std::vector<int> > >::iterator it = map_variables_candidates.begin(); it != map_variables_candidates.end(); ++it) {
        std::vector< std::vector<int> > candidates = it->second;
        if (candidates.size() == 0) {
            map_variables[it->first] = variables;
        } else {
            std::vector< std::vector<int> >::iterator candidate_it = candidates.begin();
            std::vector<int> current_candidates = *candidate_it;
            ++candidate_it ;
            for (; candidate_it != candidates.end(); ++candidate_it) {
                current_candidates = vector_intersection(current_candidates, *candidate_it);
//                                 and(current_candidates, *candidate_it);
            }
            map_variables[it->first] = current_candidates;
            if (current_candidates.size() == 0) {
                return RuleList();
            }
        }
    }

    RuleList grounded_rules;
    vector< vector<int> > variables_per_param;
    // TODO add deictic references
    for (std::vector<std::string>::const_iterator param_it = param_names.begin(); param_it != param_names.end(); ++param_it) {
        RuleList new_rules;
        vector<int> param_variables;
        if (map_variables[*param_it].size() > 0) {
            param_variables = map_variables[*param_it];
        } else {
            param_variables = variables;
        }

        variables_per_param.push_back(param_variables);
    }

    return get_grounded_rules(predicate_group, variables_per_param, !full);
}



/**
 * GroundedRule
 *
 */

GroundedRule::GroundedRule(const GroundedRule& grounded_rule):
    Rule(grounded_rule),
    symbolic_params(grounded_rule.symbolic_params),
    symbolic_preconditions(grounded_rule.symbolic_preconditions),
    symbolic_outcomes(grounded_rule.symbolic_outcomes),
    symbolic_goals(grounded_rule.symbolic_goals)
{
    is_symbolic = false;
}

GroundedRule::GroundedRule(const Rule& rule, std::vector< std::string > grounded_params_, std::vector<std::string> grounded_deictic_):
    Rule(rule),
    symbolic_params(rule.param_names)
{
    param_names = grounded_params_;
    if (param_names.size() == symbolic_params.size()) {
        ground_preconditions(grounded_params_, grounded_deictic_);
        ground_outcomes(grounded_params_, grounded_deictic_);
        ground_goals(grounded_params_, grounded_deictic_);
    } else {
        std::cerr <<  "\033[1;31m" << "GroundedRule: ERROR - wrong number of params (" << symbolic_params.size() << "!=" << param_names.size()
                  << ") for action \"" << name << "\"" << "\033[0m" << std::endl;
        std::cerr << "Symbolic params: ";
        utils::print_std_vector_string(symbolic_params, std::cerr);
        std::cerr << " --- Grounded params: ";
        utils::print_std_vector_string(param_names, std::cerr);
        std::cerr << std::endl;
    }
    is_symbolic = false;
}

GroundedRule::GroundedRule(const Rule& rule, std::vector< std::string > grounded_params_, PredicateGroup state):
// TODO remove copy-paste after migrating to C++11:   GroundedRule(rule, grounded_params_)
    Rule(rule),
    symbolic_params(rule.param_names)
{
    param_names = grounded_params_;
    if (param_names.size() == symbolic_params.size()) {
        ground_preconditions(grounded_params_);
        ground_outcomes(grounded_params_);
        ground_goals(grounded_params_);
    } else {
        std::cerr <<  "\033[1;31m" << "GroundedRule: ERROR - wrong number of params (" << symbolic_params.size() << "!=" << param_names.size()
                  << ") for action \"" << name << "\"" << "\033[0m" << std::endl;
        std::cerr << "Symbolic params: ";
        utils::print_std_vector_string(symbolic_params, std::cerr);
        std::cerr << " --- Grounded params: ";
        utils::print_std_vector_string(param_names, std::cerr);
        std::cerr << std::endl;
    }
    is_symbolic = false;
    ground_deictic_references(state);
}

void GroundedRule::ground_deictic_references(PredicateGroup state)
{
    RuleList candidates = get_deictic_grounded_rules(state);
    if (candidates.size() > 0) {
        RulePtr good_rule = candidates.front();
        preconditions = good_rule->preconditions;
        outcomes = good_rule->outcomes;
    }

}

RuleList GroundedRule::get_deictic_grounded_rules(const PredicateGroup predicate_group, const bool full) const
{
    PredicateGroup deictic_predicates = preconditions.get_symbolic_predicates();
    deictic_predicates.remove_duplicated();
    std::vector<std::string> deictic_params = deictic_predicates.get_symbolic_params();
    int DEBUG = 0;
    if (DEBUG > 0)
        std::cout << "GroundedRule: " << "Deictic grounding for rule " << name << std::endl << " state: " << predicate_group << std::endl;

    // if there are no params -> no need for grounding
    if (deictic_params.size() == 0) {
        RuleList rule_list;
        rule_list.push_back(boost::make_shared<GroundedRule>(GroundedRule(*this)));
        return rule_list;
    }

    std::vector<int> variables = predicate_group.get_numeric_variables();
    std::map< std::string, std::vector< std::vector<int> > > map_variables_candidates;
    std::map< std::string, std::vector<int> > map_variables;

    // obtain map variable -> vector< vector<candidates> >
    for (PredicateList::const_iterator precondition_it = preconditions.begin(); precondition_it != preconditions.end(); ++precondition_it) {
        for (size_t i = 0; i < precondition_it->param_names.size(); ++i) {
            if (full) {
                map_variables_candidates[precondition_it->param_names[i]].push_back(variables);
            } else if (precondition_it->positive) {
                std::vector<int> possible_variables;
                for (PredicateList::const_iterator predicate_it = predicate_group.begin(); predicate_it != predicate_group.end(); ++predicate_it) {
                    if ((precondition_it->name == predicate_it->name) && (precondition_it->positive == predicate_it->positive)) {
                        possible_variables.push_back(atoi(predicate_it->param_names[i].c_str()));
                    }
                }
                map_variables_candidates[precondition_it->param_names[i]].push_back(possible_variables);
            } else { // TODO have better strategy than just adding all
                map_variables_candidates[precondition_it->param_names[i]].push_back(variables);
            }
        }
    }

    // obtain map variables
    for (std::map< std::string, std::vector< std::vector<int> > >::iterator it = map_variables_candidates.begin(); it != map_variables_candidates.end(); ++it) {
        std::vector< std::vector<int> > candidates = it->second;
        if (candidates.size() == 0) {
            map_variables[it->first] = variables;
        } else {
            std::vector< std::vector<int> >::iterator candidate_it = candidates.begin();
            std::vector<int> current_candidates = *candidate_it;
            ++candidate_it ;
            for (; candidate_it != candidates.end(); ++candidate_it) {
                current_candidates = vector_intersection(current_candidates, *candidate_it);
//                                 and(current_candidates, *candidate_it);
            }
            map_variables[it->first] = current_candidates;
            if (current_candidates.size() == 0) {
                return RuleList();
            }
        }
    }

    // obtain variables per param
    RuleList grounded_rules;
    vector< vector<int> > variables_per_param;
    for (std::vector<std::string>::const_iterator param_it = deictic_params.begin(); param_it != deictic_params.end(); ++param_it) {
        RuleList new_rules;
        vector<int> param_variables;
        if (map_variables[*param_it].size() > 0) {
            param_variables = map_variables[*param_it];
        } else {
            param_variables = variables;
        }

        variables_per_param.push_back(param_variables);
    }

    return get_deictic_grounded_rules(predicate_group, variables_per_param, !full);
}

RuleList GroundedRule::get_deictic_grounded_rules(const PredicateGroup predicate_group, const std::vector< std::vector <int> >& variables_per_param, bool remove_invalid) const
{
    int DEBUG = 0;
    RuleList rules;

    // try all combinations and mantain only correct ones
    std::vector< std::vector< string > > combinations_of_params;
    for (size_t i = 0; i < variables_per_param.size(); ++i) {
        std::vector< std::vector< string > > new_combinations_of_params;
        for (size_t j = 0; j < variables_per_param[i].size(); ++j) {
            std::vector< string > new_params;
            std::stringstream ss;
            ss << (variables_per_param[i][j]);
            if (i == 0) {
                new_params.push_back(ss.str());
                new_combinations_of_params.push_back(new_params);
            } else {
                for (size_t k = 0; k < combinations_of_params.size(); ++k) {
                    new_params = combinations_of_params[k];
                    new_params.push_back(ss.str());
                    new_combinations_of_params.push_back(new_params);
                }
            }
        }
        combinations_of_params = new_combinations_of_params;
    }

    if (DEBUG > 1)
        std::cout << "SymbolicRule: " << "Number of combinations of params is " << combinations_of_params.size() << std::endl;

    for (size_t i = 0; i < combinations_of_params.size(); ++i) {
        GroundedRulePtr new_rule = boost::make_shared<GroundedRule>(*this, param_names, combinations_of_params[i]);

        if (DEBUG > 1)
            std::cout << "SymbolicRule: " << std::endl << "Rule: " << *new_rule << std::endl;

        if (!remove_invalid) {
            rules.push_back(new_rule);
        } else if (new_rule->satisfiesPreconditions(predicate_group)) {
            rules.push_back(new_rule);
        }
    }

    if (DEBUG > 0)
        std::cout << "SymbolicRule: " << "Number of grounded rules is " << rules.size() << std::endl;

    return rules;
}

void GroundedRule::ground_predicate_group(PredicateGroup& predicates, const std::vector< std::string >& symbolic_params_, const std::vector< std::string >& grounded_params_)
{
    std::vector<std::string>::const_iterator grounded_param_it;
    grounded_param_it = grounded_params_.begin();
    for (std::vector<std::string>::const_iterator param_it = symbolic_params_.begin(); param_it != symbolic_params_.end(); ++param_it) {
        predicates.change_variable(*param_it, *grounded_param_it);
        ++grounded_param_it;
    }
}

void GroundedRule::ground_preconditions(std::vector< std::string > grounded_params_, std::vector<std::string> grounded_deictic_params_)
{
    symbolic_preconditions = preconditions;
    ground_predicate_group(preconditions, symbolic_params, grounded_params_);
    if (grounded_deictic_params_.size() > 0) {
        std::vector<std::string> deictic_params = preconditions.get_symbolic_params();
        ground_predicate_group(preconditions, deictic_params, grounded_deictic_params_);
    }
}

void GroundedRule::ground_outcomes(std::vector< std::string > grounded_params_, std::vector<std::string> grounded_deictic_params_)
{
    symbolic_outcomes = outcomes;
    for (OutcomeList::iterator outcome_it = outcomes.begin(); outcome_it != outcomes.end(); ++outcome_it) {
        ground_predicate_group(*outcome_it, symbolic_params, grounded_params_);
    }

    if (grounded_deictic_params_.size() > 0) {
        for (OutcomeList::iterator outcome_it = outcomes.begin(); outcome_it != outcomes.end(); ++outcome_it) {
            std::vector<std::string> deictic_params = outcome_it->get_symbolic_params();
            ground_predicate_group(*outcome_it, deictic_params, grounded_deictic_params_);
        }
    }
}

void GroundedRule::ground_goals(std::vector< std::string > grounded_params_, std::vector<std::string> grounded_deictic_params_)
{
    symbolic_goals = goals;
    for (GoalGroup::iterator goal_it = goals.begin(); goal_it != goals.end(); ++goal_it) {
        ground_predicate_group(*goal_it, symbolic_params, grounded_params_);
    }

    if (grounded_deictic_params_.size() > 0) {
        for (GoalGroup::iterator goal_it = goals.begin(); goal_it != goals.end(); ++goal_it) {
            std::vector<std::string> deictic_params = goal_it->get_symbolic_params();
            ground_predicate_group(*goal_it, deictic_params, grounded_deictic_params_);
        }
    }
}

inline bool GroundedRule::covers_transition_preconditions(const Transition& transition) const
{
    return ((name == transition.action.name) && satisfiesPreconditions(transition.get_prev_state()));
}

std::vector<int> GroundedRule::experiences_per_outcome(const Transition transition) const
{
    // supposes that preconditions are satisfied
    std::vector<int> result;

    for (OutcomeList::const_iterator outcome_it = this->outcomes.begin(); outcome_it != this->outcomes.end(); outcome_it++) {
        if (*outcome_it == transition.get_changes()) {
            result.push_back(1);
        } else {
            result.push_back(0);
        }
    }

    return result;
}


bool GroundedRule::satisfiesPreconditions(const PredicateGroup predicates) const
{
    if (preconditions.is_satisfied(predicates))
        return true;
    else
        return false;
}



/**
 * RuleSet
 *
 */

RuleSet::RuleSet()
{ }

RuleSet::RuleSet(const RuleSet& rules_)
{
    add_rules(rules_);
}

RuleSet::RuleSet(const RuleList& rules_)
{
    add_rules(rules_);
}

RuleSet::RuleSet(std::string file_path)
{
    read_rules_from_file(file_path);
}

void RuleSet::add_rules(const RuleList& new_rules)
{
    for (RuleList::const_iterator rule_it = new_rules.begin(); rule_it != new_rules.end(); ++rule_it) {
        if ((*rule_it)->is_grounded()) {
            push_back(boost::make_shared<GroundedRule>(*static_cast<const GroundedRule*>(&(**rule_it))));
        } else {
            push_back(boost::make_shared<SymbolicRule>(*static_cast<const SymbolicRule*>(&(**rule_it))));
        }
    }
}

void RuleSet::read_rules_from_file(std::string file_path)
{
    std::string line;
    std::ifstream rules_file(file_path.c_str());

    // Clear previous rules
    clear();

    if (rules_file.is_open()) {
        std::string line_content;
        std::string line_content2;
        string line_title;

        getline(rules_file, line_title);
        while (rules_file.good()) {
            // ACTION:
            // read until ACTION: found
            while (line_title.find("ACTION:") == std::string::npos) {
                getline(rules_file, line_title);
            }
            
            boost::optional<uint> rule_id;
            // Get action id from ACTION: #X title
            if (line_title.find("#") != std::string::npos) { // if has id
                std::string ss_item;
                std::stringstream line_title_ss(line_title);
                std::getline(line_title_ss, ss_item, '#');
                std::getline(line_title_ss, ss_item, '#');
                try {
                    ss_item = utils::trim(ss_item);
                    rule_id = boost::lexical_cast<uint>(ss_item);
                } catch (boost::bad_lexical_cast &) {
                    // ignore
                    LOG(WARNING) << "Error when loading rule. Is there a # in a \"Action:\" line?";
                    LOG(WARNING) << "Rule line is " << line_title;
                }
            }

            // Action name
            getline(rules_file, line_content);

            // CONTEXT:
            getline(rules_file, line_title);

            // Preconditions
            getline(rules_file, line_content2);

            SymbolicRule rule = SymbolicRule(Symbol(utils::trim(line_content)),
                                             PredicateGroup(line_content2));
            if (rule_id) {
                rule.identifier = *rule_id;
            }

            // OUTCOMES:
            getline(rules_file, line_title);

            // Outcome i
            getline(rules_file, line_content);
            while ((line_content != "") && (line_content != "REWARD:")) {
                rule.add_outcome(Outcome(line_content));
                getline(rules_file, line_content);
                line_content = utils::trim(line_content);
            }

            // check if rule has rewards
            line_title = line_content;
            if (line_title == "REWARD:") {
                getline(rules_file, line_content);
                line_content = utils::trim(line_content);
                while (line_content != "") {
                    // Add rule preconditions to rule goals
                    Goal goal(line_content);
                    goal.add_predicates(rule.preconditions);
                    rule.goals.push_back(goal);
                    // Get next line
                    getline(rules_file, line_content);
                    line_content = utils::trim(line_content);
                }
            }

            // Create rule
            if ((rule.name != DUMMY_ACTION_NAME) && (rule.name != DEFAULT_ACTION_NAME)) {
                push_back(boost::make_shared<SymbolicRule>(rule));
            }

            // read lines until not empty line
            getline(rules_file, line_title);
            line_title = utils::remove_spaces(line_title);
            while ((not rules_file.eof()) && rules_file.good() && (line_title.compare("") == 0)) {
                getline(rules_file, line_title);
            }
        }
    }

    sanitize_probabities();

//         std::cout << "Read rules are " << *this << std::endl;
    rules_file.close();
}

void RuleSet::write_rules_to_file(std::string file_path) const
{
    int DEBUG = 0;
    std::ofstream rules_file(file_path.c_str());
    if (DEBUG > 0)
        std::cout << "RuleSet::write_rules_to_file - " << "Writting rules to file: '" << file_path << "'" << std::endl;
    if (rules_file.is_open()) {
        rules_file << *this;
    } else
        std::cerr << "Couldn't write rules to " << file_path << std::endl;
    rules_file.close();
}


void RuleSet::sanitize_probabities()
{
    for (RuleList::iterator rule = begin() ; rule != end(); ++rule) {
        (*rule)->sanitize_probabities();
    }
}


double RuleSet::can_obtain_predicate(const Predicate predicate) const
{
    double result = 0.0;
    for (RuleList::const_iterator rule = begin() ; rule != end(); ++rule) {
        result += (*rule)->probability_obtain_predicate(predicate);
    }
    return result;
}


bool RuleSet::remove_rules(const RuleSet& rules_to_remove)
{
    bool result = true;

    for (RuleList::const_iterator rule_remove_it = rules_to_remove.begin() ; rule_remove_it != rules_to_remove.end(); ++rule_remove_it) {
        if (!remove_rule(*rule_remove_it)) {
            result = false;
        }
    }
    return result;
}


bool RuleSet::remove_rule(const RulePtr rule_to_remove)
{
    bool result = false;
    for (RuleList::iterator rule_it = begin() ; rule_it != end();) {
        if ((*rule_to_remove) == (**rule_it)) {
            result = true;
            rule_it = erase(rule_it);
        } else {
            ++rule_it;
        }
    }
    return result;
}


RuleSet RuleSet::get_action_rules(const std::string action_name) const
{
    RuleSet action_rules;

    for (RuleList::const_iterator rule = begin() ; rule != end(); ++rule) {
        if ((*rule)->name == action_name) {
            action_rules.push_back(*rule);
        }
    }
    return action_rules;
}


RuleSet RuleSet::get_grounded_action_rules(const Symbol& grounded_action, const PredicateGroup& state) const
{
    RuleSet grounded_action_rules;
    for (RuleSet::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        GroundedRulePtr grounded_rule_ptr;
        if (((*rule_it)->is_grounded()) && ((**rule_it) == grounded_action)) {  // if grounded and actions are equal
            grounded_rule_ptr = boost::make_shared<GroundedRule>(*static_cast<const GroundedRule*>(&(**rule_it)));
        } else if ((!(*rule_it)->is_grounded()) && ((*rule_it)->name == grounded_action.name)) { // if symbolic and actions names are equal (not arguments)
            grounded_rule_ptr = boost::make_shared<GroundedRule>(GroundedRule(**rule_it, grounded_action.param_names, state)); // state is passed to ground deictic references
        }
        if (grounded_rule_ptr) {
            grounded_action_rules.push_back(grounded_rule_ptr);
        }
    }
    return grounded_action_rules;
}


RuleSet RuleSet::remove_action_rules(std::string action)
{
    RuleSet removed_rules;

    for (RuleList::iterator rule = begin() ; rule != end();) {
        if ((*rule)->name == action) {
            removed_rules.push_back(*rule);
            rule = erase(rule);
        } else {
            ++rule;
        }
    }
    return removed_rules;
}

void RuleSet::remove_nop_rules()
{
    for (RuleList::iterator rule = begin() ; rule != end();) {
        if ((*rule)->has_useful_outcome()) {
            ++rule;
        } else {
            rule = erase(rule);
        }
    }
}


void Rule::analyze_differences(const Rule& compared_rule, int& preconditions_diff_n, int& outcomes_diff_n,  double& probabilities_diff) const
{
    // preconditions
    PredicateGroup preconditions_diff = preconditions.get_differences(compared_rule.preconditions);
    preconditions_diff_n = preconditions_diff.size();

    // outcomes
    outcomes_diff_n = outcomes.size() - compared_rule.outcomes.size();

    // probabilities
    probabilities_diff = outcomes.front().probability - compared_rule.outcomes.front().probability;
}


void RuleSet::analyze_differences(RuleSet compared_ruleset, int& rule_number_diff, int& preconditions_diff, int& outcomes_num_diff, double& first_outcome_prob_diff)
{
    preconditions_diff = 0;
    outcomes_num_diff = 0;
    first_outcome_prob_diff = 0;
    rule_number_diff = - size();
    for (RuleList::iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        preconditions_diff -= (*rule_it)->preconditions.size();
        outcomes_num_diff -= (*rule_it)->outcomes.size();
        first_outcome_prob_diff -= (*rule_it)->outcomes[0].probability;
    }

    rule_number_diff += compared_ruleset.size();
    for (RuleList::iterator rule_it = compared_ruleset.begin() ; rule_it != compared_ruleset.end(); ++rule_it) {
        preconditions_diff += (*rule_it)->preconditions.size();
        outcomes_num_diff += (*rule_it)->outcomes.size();
        first_outcome_prob_diff += (*rule_it)->outcomes[0].probability;
    }
}


void RuleSet::analyze_differences(RuleSet compared_ruleset, RuleSet& added_rules, RuleSet& removed_rules,
                                  RuleSet& original_changed_rules, RuleSet& compared_changed_rules,
                                  std::vector<int>& preconditions_diff, std::vector<int>& outcomes_diff,  std::vector<double>& probabilities_diff
                                 ) const
{
    RuleSet original_rules(*this);
    std::sort(original_rules.rbegin(), original_rules.rend());  // reverse iterators! -> sorting bigger -> smaller
    std::sort(compared_ruleset.rbegin(), compared_ruleset.rend());  // reverse iterators! -> sorting bigger -> smaller

    RuleSet::iterator rule_it = original_rules.begin();
    RuleSet::iterator comp_rule_it = compared_ruleset.begin();
    while ((rule_it != original_rules.end()) && (comp_rule_it != compared_ruleset.end())) {
        if (**rule_it == **comp_rule_it) {
            preconditions_diff.push_back(0);
            outcomes_diff.push_back(0);
            probabilities_diff.push_back(0.0);
            original_changed_rules.push_back(*rule_it);
            compared_changed_rules.push_back(*comp_rule_it);
            ++rule_it;
            ++comp_rule_it;
        } else if (**rule_it > **comp_rule_it) {
            if (((rule_it + 1) != original_rules.end()) && (**(rule_it + 1) >= **comp_rule_it)) {   // exists rule_it+1 && rule_it+1 >= comp_rule_it
                removed_rules.push_back(*rule_it);
                ++rule_it;
            } else { // compare both
                int rule_preconditions_diff, rule_outcomes_diff;
                double rule_probabilities_diff;
                (*rule_it)->analyze_differences(**comp_rule_it, rule_preconditions_diff, rule_outcomes_diff, rule_probabilities_diff);
                preconditions_diff.push_back(rule_preconditions_diff);
                outcomes_diff.push_back(rule_outcomes_diff);
                probabilities_diff.push_back(rule_probabilities_diff);
                original_changed_rules.push_back(*rule_it);
                compared_changed_rules.push_back(*comp_rule_it);
                ++rule_it;
                ++comp_rule_it;
            }
        } else { // if (**comp_rule_it > **rule_it) {
            if (((comp_rule_it + 1) != compared_ruleset.end()) && (**(comp_rule_it + 1) >= **rule_it)) {     // exists comp_rule_it+1 && comp_rule_it+1 >= rule_it
                added_rules.push_back(*comp_rule_it);
                ++comp_rule_it;
            } else { // compare both
                int rule_preconditions_diff, rule_outcomes_diff;
                double rule_probabilities_diff;
                (*rule_it)->analyze_differences(**comp_rule_it, rule_preconditions_diff, rule_outcomes_diff, rule_probabilities_diff);
                preconditions_diff.push_back(rule_preconditions_diff);
                outcomes_diff.push_back(rule_outcomes_diff);
                probabilities_diff.push_back(rule_probabilities_diff);
                original_changed_rules.push_back(*rule_it);
                compared_changed_rules.push_back(*comp_rule_it);
                ++rule_it;
                ++comp_rule_it;
            }
        }
    }

    // add remaining rules
    while (rule_it != original_rules.end()) {
        removed_rules.push_back(*rule_it);
        ++rule_it;
    }
    while (comp_rule_it != compared_ruleset.end()) {
        added_rules.push_back(*comp_rule_it);
        ++comp_rule_it;
    }
}

void RuleSet::set_identifiers(RuleSetPtr old_rules_ptr)
{
    // Reasons to reverse sort:
    //  - Important rules are those with outcomes
    //  - If rules with outcomes, the identifier will asign better identifiers to rules with outcomes
    //  - Actually, this method could be improved a lot using distances instead of < relations, but not worth the time
    //  as identfiers are used just for gource logging.
    std::sort(rbegin(), rend());  // reverse iterators! -> sorting bigger -> smaller
    std::sort(old_rules_ptr->rbegin(), old_rules_ptr->rend());  // reverse iterators! -> sorting bigger -> smaller

    RuleSet::iterator rule_it = begin();
    RuleSet::iterator old_rule_it = old_rules_ptr->begin();
    while ((rule_it != end()) && (old_rule_it != old_rules_ptr->end())) {
        if (**rule_it == **old_rule_it) {
            (*rule_it)->identifier = (*old_rule_it)->identifier;
            ++rule_it;
            ++old_rule_it;
        } else if (**rule_it > **old_rule_it) {
            if (((rule_it + 1) != end()) && (**(rule_it + 1) >= **old_rule_it)) {     // exists rule_it+1 && rule_it+1 >= old_rule_it
                ++rule_it;
            } else { // just similar
                (*rule_it)->identifier = (*old_rule_it)->identifier;
                ++rule_it;
                ++old_rule_it;
            }
        } else { // if (**old_rule_it > **rule_it) {
            if (((old_rule_it + 1) != old_rules_ptr->end()) && (**(old_rule_it + 1) >= **rule_it)) {     // exists old_rule_it+1 && old_rule_it+1 >= rule_it
                ++old_rule_it;
            } else { // just similar
                (*rule_it)->identifier = (*old_rule_it)->identifier;
                ++rule_it;
                ++old_rule_it;
            }
        }
    }
}

RuleSet RuleSet::find_rules_with_a_precondition(const Predicate& precondition) const
{
    RuleSet result;
    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        if ((*rule_it)->preconditions.has_predicate(precondition)) {
            result.push_back(*rule_it);
        }
    }
    return result;
}

RuleSet RuleSet::find_rules_with_a_precondition(const PredicateGroup& possible_preconditions) const
{
    RuleSet result;
    for (PredicateGroup::const_iterator pred_it = possible_preconditions.begin() ; pred_it != possible_preconditions.end(); ++pred_it) {
        result.add_rules(find_rules_with_a_precondition(*pred_it));
    }
    return result;
}

RuleSet RuleSet::find_rules_with_grounded_precondition(const Predicate precondition, const PredicateGroup current_state) const
{
    RuleSet result;
    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        bool found = false;
        RuleSet grounded_rules = (*rule_it)->get_grounded_rules(current_state);
        for (RuleList::const_iterator grule_it = grounded_rules.begin() ; grule_it != grounded_rules.end(); ++grule_it) {
            if ((*grule_it)->preconditions.has_predicate(precondition)) {
                found = true;
            }
        }
        if (found) {
            result.push_back(*rule_it);
        }
    }

    return result;
}

std::vector<Symbol> RuleSet::get_actions() const
{
    std::vector<Symbol> result;
    std::set<Symbol> action_symbols;
    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        action_symbols.insert(**rule_it);
    }
    // std::set to std::vector
    result.assign(action_symbols.begin(), action_symbols.end());
    return result;
}

std::vector<std::string> RuleSet::get_action_names() const
{
    std::vector<std::string> result;
    std::set<std::string> action_names;
    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        action_names.insert((*rule_it)->name);
    }
    // std::set to std::vector
    result.assign(action_names.begin(), action_names.end());
    return result;
}


void RuleSet::add_default_rule_if_empty()
{
    if (size() == 0) {
        // add default rule so it doesn't crash
        SymbolicRule rule(Symbol("default"), PredicateList());
        rule.add_outcome(Outcome(0.5, false));
        rule.add_outcome(Outcome(0.5, true));
        push_back(boost::make_shared<SymbolicRule>(rule));
    }
}

double RuleSet::score(const TransitionList transitions) const
{
    double cutting_threshold = -1. * std::numeric_limits<double>::max();
    return score(transitions, cutting_threshold);
}

double RuleSet::score(const TransitionList transitions, const double cutting_threshold) const
{
    std::vector<float> experience_weights(transitions.size());
    std::fill(experience_weights.begin(), experience_weights.end(), 1);
    return score(transitions, cutting_threshold, experience_weights);
}

double RuleSet::score(const TransitionList transitions, const double cutting_threshold, const std::vector<float> experience_weights) const
{
    uint DEBUG = 0;
    if (DEBUG > 0) {
        std::cout << "SCORE [start]" << std::endl;
    }
    if (DEBUG > 1) {
        std::cout << "Cutting threshold is " << cutting_threshold << std::endl;
    }

    // learning parameters
    double alpha = 0.5;
    double p_min = 1e-9;
    double p_min_noisyDefaultRule = 1e-11;

    // (1) Penalty
    double penalty = 0.0;
    penalty += get_number_literals();
//     FOR1D_(rulesC.rules, i) {
//         penalty += rulesC.rules.elem(i)->numberLiterals();
//     }
    penalty *= alpha;
    if (DEBUG > 0) {
        std::cout << "alpha: " << alpha << std::endl;
        std::cout << "penalty: " << penalty << std::endl;
    }

    // (2) Log-Likelihood
    double loglikelihood = 0.0, exLik;
    if (DEBUG > 0) {
        std::cout << "Calculating likelihoods..." << std::endl;
    }

    uint i = 0;
    for (TransitionList::const_iterator trans_it = transitions.begin() ; trans_it != transitions.end(); ++trans_it) {
//     FOR1D(experiences, i) {
        if (DEBUG > 1) {
            std::cout << "+++ ex " << i << ": ";
//             experiences(i)->action->write(cout);
            std::cout << trans_it->action << std::endl;
            std::cout << std::endl;
        }

        RuleSet covering_rules = get_covering_rules(*trans_it);

//         const uintA& covering_rules = rulesC.nonDefaultRules_per_experience(i);
        if (DEBUG > 1) {
//             PRINT(covering_rules);
            std::cout << "Covering rules are " << covering_rules.size() << ": " << std::endl;
            std::cout << covering_rules << std::endl;
        }

        // only one non-default rule covers (good case)
        if (covering_rules.size() == 1) {
            RulePtr rule_ptr = covering_rules.front();
//             Rule* rule = rulesC.rules.elem(covering_rules(0));
            if (DEBUG > 2) {
//                 std::cout << "Use rule #" << covering_rules(0) << "  " << rulesC.experiences_per_rule(covering_rules(0)) << std::endl;
                std::cout << "Use rule #" << *(covering_rules.front()) << std::endl;
//                 rule->write(cout);
            }

//             uint o = 0;
            std::vector<int> exp_per_out = rule_ptr->experiences_per_outcome(*trans_it);
//             MT::Array< uintA >& exs_per_out = rulesC.experiences_per_ruleOutcome(covering_rules(0));
            if (DEBUG > 2) {
//                 PRINT(exs_per_out);
                std::cout << "Experiences per rule outcome " << std::endl;
                /*std::cout << exp_per_out <<*/
                for (size_t i = 0; i < exp_per_out.size(); i++)
                    std::cout << exp_per_out[i] << ", ";
                std::cout << std::endl;
            }
            exLik = 0.0;
            for (size_t j = 0; j < exp_per_out.size(); j++) {
//             for (std::vector<int>::iterator exp_it=exp_per_outcome.begin(); exp_it!=exp_per_out.end(); ++exp_it) {
//             FOR1D(exs_per_out, o) {
                // Non-noise outcome
                if ((exp_per_out[j] == 1) && (!rule_ptr->outcomes[j].noisy)) {
//                 if (o<exs_per_out.N-1  &&  exs_per_out(o).findValue(i) >= 0) {
                    exLik += rule_ptr->outcomes[j].probability;
                    if (DEBUG > 2) {
                        std::cout << "after o=" << j << ": " << exLik << std::endl;
//                         printf("after o=%i:  %10.5f\n", j, exLik);
                    }
                }
                // Noise-outcome:  always covers
//                 if (o == exs_per_out.N-1) {
                if (rule_ptr->outcomes[j].noisy) {
                    exLik += p_min * rule_ptr->outcomes[j].probability;
                    if (DEBUG > 2) {
                        std::cout << "after o=" << j << ": " << exLik << std::endl;
//                         printf("after o=%i:  %10.15f\n", j, exLik);
                    }
                }
            }
            if (exLik <= 0.) {
                if (DEBUG > 0) {
                    std::cerr << "ERROR: " << "bad referencing  exLik=" << exLik << std::endl;
                }
                std::cerr << "!";
                exLik += p_min_noisyDefaultRule * 0.01;
            }
//             CHECK(exLik>0., "bad referencing  exLik="<<exLik);
        }
        // only default covers or more than one non-default covers
        // --> apply default rule
        else {
            /*
            ACTION:
            default()
            CONTEXT:
            --
            OUTCOMES:
            0.5 <no-change>
            0.5 <noise>
             */
            if (DEBUG > 2) {
                std::cout << " Using default rule." << std::endl;
            }
            if (trans_it->nothing_changed())
                exLik = 0.5 * (100.0 * p_min_noisyDefaultRule);   // Avoid modeling persistence by means of noise default rule.
            else
                exLik = 0.5 * p_min_noisyDefaultRule;

//             Rule* default_rule = rulesC.rules.elem(0);
//             if (rulesC.experiences_per_ruleOutcome(0)(0).findValue(i) >= 0)
//                 exLik = default_rule->probs(0) * (100.0 * __p_min_noisyDefaultRule);   // Avoid modeling persistence by means of noise default rule.
//             else
//                 exLik = default_rule->probs(1) * __p_min_noisyDefaultRule;
        }

        loglikelihood += experience_weights[i] * log(exLik); // weight the experiences differently
        if (DEBUG > 1) {
            std::cout << " --> lik=" << exLik << std::endl;
        }
        if (DEBUG > 0) {
            std::cout << "log likelihoood is " << loglikelihood << std::endl;
        }


        if (loglikelihood - penalty  <  cutting_threshold) {
            if (DEBUG > 0) {
                std::cout << "Difference between loglikelihood and penalty is " << loglikelihood - penalty << std::endl;
//                 PRINT(loglikelihood - penalty);
                std::cout << "Cutting threshold is " << cutting_threshold << std::endl;
//                 PRINT(cutting_threshold);
                std::cout << "Score becomes too small... giving up." << std::endl;
                std::cout << "SCORE [end]" << std::endl;
            }
//             return TL::TL_DOUBLE_MIN;
            return (-1. * std::numeric_limits<double>::max());
        }
        i++;
    }

    double score = loglikelihood - penalty;
    if (DEBUG > 0) {
        std::cout << "score is " << score << std::endl;
        std::cout << "SCORE [end]" << std::endl;
    }

    return score;
}

int RuleSet::get_number_literals() const
{
    int result = 0;
    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        result += (*rule_it)->get_number_literals();
    }
    return result;
}

RuleSet RuleSet::get_covering_rules(const Transition transition) const
{
    RuleSet result;
    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        if ((*rule_it)->covers_transition_preconditions(transition)) {
            result.push_back(*rule_it);
        }
    }

    return result;
}

RuleSet RuleSet::get_covering_rules(const PredicateGroup predicates) const
{
    RuleSet result;
    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        if ((*rule_it)->satisfiesPreconditions(predicates)) {
            result.push_back(*rule_it);
        }
    }

    return result;
}

RuleSet RuleSet::get_grounded_rules(const PredicateGroup& predicates) const
{
    RuleSet grounded_rules;

    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        RuleList new_grounded_rules = (*rule_it)->get_grounded_rules(predicates);
        grounded_rules.add_rules(new_grounded_rules);
    }

    return grounded_rules;
}

RuleSet RuleSet::get_full_grounded_rules(const PredicateGroup& predicates) const
{
    RuleSet grounded_rules;

    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        RuleList new_grounded_rules = (*rule_it)->get_grounded_rules(predicates, true);
        grounded_rules.add_rules(new_grounded_rules);
    }

    return grounded_rules;
}

void RuleSet::set_dangerous_rules(const PredicateList& dead_end_excuses)
{
    for (RuleList::iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        for (PredicateList::const_iterator dead_end_excuse_it = dead_end_excuses.begin() ; dead_end_excuse_it != dead_end_excuses.end(); ++dead_end_excuse_it) {
            if ((*rule_it)->has_outcome_predicate(dead_end_excuse_it->get_negated())) {
                (*rule_it)->dangerous = true;
            }
        }
    }
}

bool RuleSet::is_dangerous(const Symbol action, const PredicateGroup& state) const
{
    RuleSet action_rules = get_grounded_action_rules(action, state);
    RuleSet covering_rules = action_rules.get_covering_rules(state);
    for (RuleList::iterator rule_it = covering_rules.begin() ; rule_it != covering_rules.end(); ++rule_it) {
        if ((*rule_it)->dangerous) {
            return true;
        }
    }
    return false;
}

void RuleSet::add_negated_outcomes_as_preconditions()
{
    for (RuleList::iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        (*rule_it)->add_negated_outcomes_as_preconditions();
    }
}

void RuleSet::add_predicate_to_preconditions(const Predicate& new_predicate)
{
    for (RuleList::iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        (*rule_it)->add_predicate_to_preconditions(new_predicate);
    }
}

void RuleSet::add_predicate_to_outcomes(const Predicate& new_predicate)
{
    BOOST_FOREACH(RulePtr rule, *this) {
        rule->add_predicate_to_outcomes(new_predicate);
    }
}

RuleSet RuleSet::get_different_rules(const RuleSet& comp_ruleset) const
{
    RuleSet differences;
    // supposes that rulesets are ordered
    RuleSet::const_iterator rule_it = begin(), comp_rule_it = comp_ruleset.begin();
    for (; (rule_it != end()) && (comp_rule_it != comp_ruleset.end());) {
        if (*rule_it < *comp_rule_it) {
            differences.push_back((*rule_it)->make_shared());
            ++rule_it;
        } else if (*comp_rule_it < *rule_it) {
            differences.push_back((*comp_rule_it)->make_shared());
            ++comp_rule_it;
        } else { //  (*rule_it == *comp_rule_it)
            ++rule_it;
            ++comp_rule_it;
        }
    }
    // add remaining (if any) in RuleSet
    for (; rule_it != end(); ++rule_it) {
        differences.push_back((*rule_it)->make_shared());
    }
    // add remaining (if any) in comp_ruleset
    for (; comp_rule_it != comp_ruleset.end(); ++comp_rule_it) {
        differences.push_back((*comp_rule_it)->make_shared());
    }
    return differences;
}

void RuleSet::add_goal(Goal goal)
{
    for (RuleSet::iterator rule_it = begin(); rule_it != end(); ++rule_it)  {
        (*rule_it)->add_goal(goal);
    }
}


/**
 * Domain indexing
 */

void RuleSet::index(Domain& domain) const
{
    for (RuleList::const_iterator rule_it = begin(); rule_it != end(); ++rule_it) {
        (*rule_it)->index(domain);
    }
}

void Rule::index(Domain& domain) const
{
    domain.index_action(*this);
    domain.index_predicates(preconditions);
    for (OutcomeList::const_iterator out_it = outcomes.begin() ; out_it != outcomes.end(); out_it++) {
        domain.index_predicates(*out_it);
    }
}


/**
 * PPDDL
 */

void RuleSet::PPDDL_grounding()
{
    for (RuleList::iterator rule_it = begin(); rule_it != end(); ++rule_it) {
        (*rule_it)->PPDDL_grounding();
    }
}

void Rule::PPDDL_grounding()
{
    preconditions.PPDDL_grounding();
    for (GoalGroup::iterator goal_it = goals.begin(); goal_it != goals.end(); ++goal_it) {
        goal_it->PPDDL_grounding();
    }
    for (OutcomeList::iterator out_it = outcomes.begin() ; out_it != outcomes.end(); out_it++) {
        out_it->PPDDL_grounding();
    }
    (static_cast<Symbol*>(this))->PPDDL_grounding();
}

bool order_rules_by_name(const RulePtr i, const RulePtr j)
{
    return (i->name < j->name);
}

std::map<std::string, RuleSet> RuleSet::get_rules_per_action() const
{
    std::map<std::string, RuleSet> result;
    if (empty()) {
        return result;
    }
    RuleSet ordered_rules(*this);
    std::sort(ordered_rules.begin(), ordered_rules.end(), order_rules_by_name);

    std::string current_action = ordered_rules[0]->name;
    RuleSet current_ruleset;
    for (RuleList::const_iterator rule_it = ordered_rules.begin(); rule_it != ordered_rules.end(); ++rule_it) {
        if ((*rule_it)->name == current_action) {
            current_ruleset.push_back(*rule_it);
        } else {
            result[current_action] = current_ruleset;
            current_action = (*rule_it)->name;
            current_ruleset = RuleSet();
            current_ruleset.push_back(*rule_it);
        }
    }
    result[current_action] = current_ruleset;
    return result;
}



