/*
 * Clases reprenting symbolic and grounded rules.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_RULES_H
#define IRI_RULE_LEARNER_RULES_H 1

#include <limits>
#include <vector>
#include <fstream>
#include <cmath>
#include <map>

#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>

#include <symbolic/symbols.h>
#include <symbolic/predicates.h>
#include <symbolic/domain.h>
#include <symbolic/transitions.h>


/**
 * \class Rule
 * \brief Define basic rule functionality, used in both symbolic and grounded rules.
 * 
 * Note that the code to ground rules is a mess. Should be replaced with prolog to have a cleaner and better implementation.
 */
class Rule : public Symbol
{
public:
    PredicateGroup preconditions;
    OutcomeGroup outcomes;
    GoalGroup goals;
    bool dangerous;
    ulong identifier;

    /*
     * Constructors
     */
    Rule(const Rule& rule);
    Rule(const Symbol &symbol_, const PredicateList &preconditions_);

    /*!
     * \brief Copy the rule to the heap and return a smart pointer.
     * 
     * The changes of the returned rule are not mirrored back to this one.
     * 
     * \return shared pointer to the copy of the rule
     */
    virtual boost::shared_ptr<Rule> make_shared() const =0;
    
    void set_reward(const float reward);

    /**
     * \brief Obtain the probability of getting a predicate with the rule, assuming that the preconditions are satisfied.
     */
    double probability_obtain_predicate(const Predicate predicate) const;

    void add_outcome(const PredicateList predicates, const float init_probability);
    void add_outcome(const Outcome outcome);
    
    /**
     * \brief adds negated outcome predicates to the preconditions predicates.
     */
    void add_negated_outcomes_as_preconditions();
    
    void add_predicate_to_preconditions(const Predicate& new_predicate);
    
    /**
     * \brief adds the predicate to every outcome.
     */
    void add_predicate_to_outcomes(const Predicate& new_predicate);

    /**
     * \brief Sanitizes probabilities to make them sum 1.0.
     */
    void sanitize_probabities();

    /***** GET INFO FROM RULES *****/
    /**
     * \brief Get number of literals in preconditions and outcomes.
     */
    int get_number_literals() const;

    /**
     * \brief Check if at least one outcome changes predicates and is not noisy.
     */
    bool has_useful_outcome() const;

    /**
     * \brief Check if the rule can obtain the predicate in any of its outcomes.
     */
    bool has_outcome_predicate(const Predicate& predicate) const;

    /*
     * Goals
     */
    void add_goal(Goal goal) {
        goals.push_back(goal);
    }
    void set_goal(float new_reward) {
        Goal goal(new_reward);
        goals.clear();
        add_goal(goal);
    };

    /*
     * Transitions
     */
    /**
     * \brief counts the number of transitions initial states and actions that this rule covers.
     * 
     * A transition is covered when it has the same action name and the preconditions are satisfied by the transition's initial state.
     */
    uint count_covering_transitions_preconditions(const TransitionList& transitions) const;
    
    /**
     * \brief gets transitions covered by the rule.
     */
    TransitionList get_covering_transitions(const TransitionList transitions) const;

    /**
     * \brief Analyze rule differences in number of precondition predicates, outcomes and probabilities.
     * 
     * \param compared_rule rule to be compared.
     * \retval preconditions_diff_n number of different preconditions predicates.
     * \retval outcomes_diff_n number of different outcomes.
     * \retval probabilities_diff difference in the first outcome probability.
     */
    void analyze_differences(const Rule& compared_rule, int& preconditions_diff_n, int& outcomes_diff_n,  double& probabilities_diff) const;

    /**
     * \brief counts the number of transitions initial states and actions that this rule covers.
     * 
     * A transition is covered when it has the same action name and the preconditions are satisfied by the transition's initial state.
     */
    virtual bool covers_transition_preconditions(const Transition& transition) const = 0;
    
    /**
     * \brief checks if rule is grounded.
     */
    virtual bool is_grounded() const = 0;
    
    /**
     * \brief check if prev_state satisfies rule preconditions.
     */
    virtual bool satisfiesPreconditions(PredicateGroup prev_state) const = 0;

    /**
     * \brief Gets a vector where element i indicates the number of transitions that outcome[i] covers
     */
    virtual std::vector<int> experiences_per_outcome(const Transition transition) const = 0;

    /**
     * \brief obtains the grounded rule for the rule given a current state.
     * 
     * By default only grounded rules valid for the current state are obtained.\n
     * The objects used to ground are the ones present in the state.
     * 
     * \param state current state where the rule will be grounded.
     * \param full if true, all grounded rules are obtained, even if their preconditions don't satisfy the state.
     */
    virtual std::vector<boost::shared_ptr<Rule> > get_grounded_rules(const PredicateGroup state, const bool full = false) const = 0;

    // PPDDL
    /**
     * \brief Does a grounding to be used with PPDDL IPPC planners.
     */
    void PPDDL_grounding();

    // domain
    /**
     * \brief Indexes the rule in the domain.
     */
    void index(Domain& domain) const;

    friend std::ostream& operator<<(std::ostream &out, const Rule& r);
    friend bool operator==(const Rule &r1, const Rule &r2);

    /**
     * \brief Compares two rules.
     * 
     * Comparison priority:
     * - 1º - name
     * - 2º - outcomes.size()
     * - 3º - outcomes predicates
     * - 4º - outcomes.probability
     * - 5º - preconditions
     */
    friend bool operator<(const Rule& l, const Rule& r);
    friend bool operator>(const Rule& l, const Rule& r);
    friend bool operator<=(const Rule& l, const Rule& r);
    friend bool operator>=(const Rule& l, const Rule& r);
};

typedef boost::shared_ptr<Rule> RulePtr;
typedef std::vector< RulePtr > RuleList;


/**
 * \class SymbolicRule
 * \brief Symbolic rule implementation. All parameters are considered to be symbolic.
 */
class SymbolicRule : public Rule
{
public:
    SymbolicRule(const Rule& rule);
    SymbolicRule(const SymbolicRule& rule);
    SymbolicRule(const Symbol &symbol_, const PredicateList &preconditions_);
    
    virtual RulePtr make_shared() const { return RulePtr(new SymbolicRule(*this)); }

    virtual inline bool covers_transition_preconditions(const Transition& transition) const;
    virtual std::vector<int> experiences_per_outcome(const Transition transition) const;
    virtual inline bool is_grounded() const {
        return false;
    };

    virtual bool satisfiesPreconditions(const PredicateGroup predicate_group) const;
    virtual RuleList get_grounded_rules(const PredicateGroup predicate_group, const bool full = false) const;

private:
    RuleList get_grounded_rules(const PredicateGroup predicate_group, const std::vector< std::vector <int> >& variables_per_param, bool remove_invalid = true) const;
};


/**
 * \class GroundedRule
 * \brief Grounded rule implementation. All parameters are considered to be grounded.
 */
class GroundedRule : public Rule
{
public:
    /**
     * \brief Gets the set of possible rules by grounded all deictic references based on the predicate_group.
     * 
     * By default, only candidates satisfied by the predicate_group are returned.
     * 
     * \param predicate_group predicates that have to be satisfied by the rules. They also provide the list of objects available.
     * \param full returns all possible grounded rule sets, even if the predicate_group doesn't satisfy them.
     * \return list of rules with all deictic references grounded.
     */
    // TODO ugly hack, should use prolog
    RuleList get_deictic_grounded_rules(const PredicateGroup predicate_group, const bool full = false) const;

    GroundedRule(const Rule& rule, std::vector< std::string > grounded_params_, std::vector<std::string> grounded_deictic_params_  = std::vector<std::string>());
    GroundedRule(const Rule& rule, std::vector< std::string > grounded_params_, PredicateGroup state);
    GroundedRule(const GroundedRule& grounded_rule);
    
    virtual RulePtr make_shared() const { return RulePtr(new GroundedRule(*this)); }

    virtual RuleList get_grounded_rules(const PredicateGroup predicate_group, const bool full = false) const {
        RuleList result;
        boost::shared_ptr<GroundedRule> rule_ptr = boost::make_shared<GroundedRule>(*this);
        result.push_back(rule_ptr);
        return result;
    };

    // redefined virtual methods
    virtual inline bool covers_transition_preconditions(const Transition& transition) const;
    virtual inline bool is_grounded() const {
        return true;
    };
    virtual std::vector<int> experiences_per_outcome(const Transition transition) const;

    virtual bool satisfiesPreconditions(const PredicateGroup predicates) const;
    
private:
    std::vector< std::string > symbolic_params;
    PredicateGroup symbolic_preconditions;
    OutcomeList symbolic_outcomes;
    GoalGroup symbolic_goals;

    void ground_outcomes(std::vector< std::string > grounded_params_, std::vector<std::string> grounded_deictic_params_ = std::vector<std::string>());
    void ground_preconditions(std::vector< std::string > grounded_params_, std::vector<std::string> grounded_deictic_params_  = std::vector<std::string>());
    void ground_goals(std::vector< std::string > grounded_params_, std::vector<std::string> grounded_deictic_params_  = std::vector<std::string>());
    void ground_deictic_references(PredicateGroup state);

    void ground_predicate_group(PredicateGroup& predicates, const std::vector< std::string >& symbolic_params_, const std::vector< std::string >& grounded_params_);

    // TODO ugly hack (should migrate to use prolog?)
    RuleList get_deictic_grounded_rules(const PredicateGroup predicate_group, const std::vector< std::vector <int> >& variables_per_param, bool remove_invalid = true) const;
};



typedef boost::shared_ptr<SymbolicRule> SymbolicRulePtr;
typedef boost::shared_ptr<GroundedRule> GroundedRulePtr;

typedef boost::shared_ptr<const Rule> RuleConstPtr;
typedef boost::shared_ptr<const SymbolicRule> SymbolicRuleConstPtr;
typedef boost::shared_ptr<const GroundedRule> GroundedRuleConstPtr;

/**
 * \class RuleSet
 * \brief A vector of RulePtr with extended functionality.
 */
class RuleSet : public RuleList
{
public:
    // constructors
    RuleSet();
    RuleSet(const RuleList& rules_);
    RuleSet(const RuleSet& rules_);
    RuleSet(std::string file_path);

    // i/o with files
    void read_rules_from_file(std::string file_path);
    void write_rules_to_file(std::string file_path) const;

    /**
     * \brief Sanitizes rule probabilities so they sum 1.0.
     */
    void sanitize_probabities();
    
    /**
     * \brief Gets the action names represented by the rules.
     */
    std::vector<std::string> get_action_names() const;
    
    /**
     * \brief Gets the actions represented by the rules.
     */
    std::vector<Symbol> get_actions() const;

    // rules info
    /**
     * \brief Gets the summed probability that the rules can obtain the predicate considering that all the preconditions are satisfied.
     */
    double can_obtain_predicate(const Predicate predicate) const;

    // add and remove rules
    /**
     * \brief If the rule set is empty, a default rule that changes nothing is added.
     */
    void add_default_rule_if_empty();
    
    void add_rules(const RuleList& new_rules);
    inline void set_rules(const RuleList& new_rules) {
        clear();
        add_rules(new_rules);
    }
    RuleSet remove_action_rules(std::string action);
    bool remove_rules(const RuleSet& rules_to_remove);
    bool remove_rule(const RulePtr rule_to_remove);
    
    /**
     * \brief Removes rules that cannot change nothing (no useful outcomes).
     */
    void remove_nop_rules();

    // Actions stuff
    /**
     * \brief Gets a map that for each action has the vector of rules that represent it.
     */
    std::map<std::string, RuleSet> get_rules_per_action() const;
    
    /**
     * \brief Gets the rules whose action name is the same as the action parameter.
     * 
     * Uses a symbolic action.
     */
    RuleSet get_action_rules(const std::string action_name) const;
    
    /**
     * \brief Get grounded action rules. The state is used just for deictic references.
     * 
     * The state is used just for deictic references.
     * 
     * \param grounded_action a grounded action.
     * \param state state from which to obtain objects to generate the deictic references.
     * \return grounded rules for the grounded action and state
     */
    RuleSet get_grounded_action_rules(const Symbol& grounded_action, const PredicateGroup& state) const;
    
    /**
     * \brief Get grounded rules satisfied by predicates.
     */
    RuleSet get_grounded_rules(const PredicateGroup& predicates) const;
    
    /**
     * \brief Get grounded rules using predicates' objects. They don't have to be satisfied by predicates.
     */
    RuleSet get_full_grounded_rules(const PredicateGroup& predicates) const;

    // get info
    /**
     * \brief Get number of literals, considering all preconditions and outcomes.
     */
    int get_number_literals() const;
    
    /**
     * \brief Get rules that cover the transition preconditions and have the same action name.
     */
    RuleSet get_covering_rules(const Transition transition) const;
    
    /**
     * \brief Get rules whose preconditions are satisfied by the predicates.
     */
    RuleSet get_covering_rules(const PredicateGroup predicates) const;
    
    /**
     * \brief Check if rule is dangerous.
     * 
     * A rule is dangerous when a dead end was obtained, and the negated excuse predicates where in the outcomes of the rule.
     */
    bool is_dangerous(const Symbol action, const PredicateGroup& state) const;
    
    /**
     * \brief Gets the rule that are different between the two rule sets.
     */
    RuleSet get_different_rules(const RuleSet& comp_ruleset) const;

    // set info
    /**
     * \brief Sets the dangerous flag in all rules that can obtain the negated dead_end_excuses.
     */
    void set_dangerous_rules(const PredicateList& dead_end_excuses);
    
    /**
     * \brief Add all negated outcomes predicates to the preconditions.
     */
    void add_negated_outcomes_as_preconditions();
    
    void add_predicate_to_preconditions(const Predicate& new_predicate);
    void add_predicate_to_outcomes(const Predicate& new_predicate);
    void add_goal(Goal goal);

    // rule analysis
    /**
     * \brief Set unique identifiers to rules.
     * 
     * Equal rules maintain the same identifier. Else, the identifiers are passed to the most similar rule if applicable.
     * New different rules get a new identifier.
     * This is just used for visualization purposes.\n
     * TODO documentation.
     */
    void set_identifiers(boost::shared_ptr<RuleSet> new_rules_ptr);
    
    /**
     * \brief Get info about the learning step to be shown to the teacher.
     */
    void analyze_differences(RuleSet compared_ruleset, int& rule_number_diff, int& preconditions_diff, int& outcomes_num_diff, double& first_outcome_prob_diff);
    
    /**
     * \brief Get info about the learning step to be shown to the teacher.
     */
    void analyze_differences(RuleSet compared_ruleset, RuleSet& added_rules, RuleSet& removed_rules,
                             RuleSet& original_changed_rules, RuleSet& compared_changed_rules,
                             std::vector<int>& preconditions_diff, std::vector<int>& outcomes_diff,  std::vector<double>& probabilities_diff
                            ) const;
                            
    RuleSet find_rules_with_a_precondition(const Predicate& precondition) const;
    RuleSet find_rules_with_a_precondition(const PredicateGroup& possible_preconditions) const;
    RuleSet find_rules_with_grounded_precondition(const Predicate precondition, const PredicateGroup current_state) const;

    /**
     * \brief RuleSet score metric (Pasula et al JAIR 07)
     *
     * S(R) = sum(rules) sum(transitions) log(P(s'|s,a,r)) - ( alpha * PEN(r) )
     */
    double score(const TransitionList transitions) const;
    
    /**
     * \brief RuleSet score metric (Pasula et al JAIR 07)
     *
     * S(R) = sum(rules) sum(transitions) log(P(s'|s,a,r)) - ( alpha * PEN(r) )
     */
    double score(const TransitionList transitions, const double cutting_threshold) const;
    
    /**
     * \brief RuleSet score metric (Pasula et al JAIR 07)
     *
     * S(R) = sum(rules) sum(transitions) log(P(s'|s,a,r)) - ( alpha * PEN(r) )
     */
    double score(const TransitionList transitions, const double cutting_threshold, const std::vector<float> experience_weights) const;

    // domain
    /**
     * \brief Indexes the rules in the domain.
     * 
     * As domain depends on rules, this can't be included inside of domain.\n
     * TODO check why!!!
     */
    void index(Domain& domain) const;

    /**
     * \brief Does a grounding to be used with PPDDL IPPC planners.
     */
    void PPDDL_grounding();

    // operators
    friend bool operator==(const RuleSet &r1, const RuleSet &r2);
    friend bool operator!=(const RuleSet &r1, const RuleSet &r2);
//     friend bool operator<(const RuleSet& l, const RuleSet& r );
//     friend bool operator>(const RuleSet& l, const RuleSet& r );
//     friend bool operator<=(const RuleSet& l, const RuleSet& r );
//     friend bool operator>=(const RuleSet& l, const RuleSet& r );
    friend std::ostream& operator<<(std::ostream &out, const RuleSet& r);
};

typedef boost::shared_ptr<RuleSet> RuleSetPtr;
typedef boost::shared_ptr<const RuleSet> RuleSetConstPtr;

// typedef std::vector<GraspPhasePtr> GraspPhaseSequence;

#endif
