/*
    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) 2013  dmartinez <email>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "symbolic/scenario.h"


Scenario::Scenario(const RuleSet& rules_, const State& state_, const GoalGroup& goals_, 
                   const ConfigReader& config_reader_, 
                   const std::vector<Symbol>& action_list_):
    config_reader(config_reader_),
    rules(rules_),
    state(state_),
    goals(goals_),
    action_list(action_list_)
{
    domain.index_goals(goals);
    domain.index_predicates(state);
    rules.index(domain);
    
    if (action_list.empty()) {
        action_list = rules.get_actions();
    }

    if (config_reader.get_variable<bool>("final_reward_action")) {
        RuleSet finish_rule(config_reader.get_path("nid_finish_rule_path"));
        rules.add_rules(finish_rule);
        finish_rule.index(domain);
        // -finished() predicate used by final reward action
        rules.add_predicate_to_preconditions(Predicate(MAKE_SYMBOL_FINISHED, false));
    }
}


Scenario::Scenario(const Scenario& scenario):
    domain(scenario.domain),
    config_reader(scenario.config_reader),
    rules(scenario.rules),
    state(scenario.state),
    goals(scenario.goals),
    action_list(scenario.action_list)
{
}


void Scenario::add_Rules_to_action_list(const RuleSet& rules)
{
    std::vector<Symbol> actions = rules.get_actions();
    BOOST_FOREACH(const Symbol& action, actions) {
        add_action_to_action_list(action);
    }
}


void Scenario::add_action_to_action_list(Symbol action_copy)
{
    action_copy.make_symbolic();
    bool isPresent = (std::find(action_list.begin(), action_list.end(), action_copy) != action_list.end());
    if (!isPresent) {
        action_list.push_back(action_copy);
    }
//     std::random_shuffle ( action_list.begin(), action_list.end() );
}

void Scenario::add_actions_to_action_list(std::vector<Symbol> action_list)
{
    BOOST_FOREACH(const Symbol& action, action_list) {
        add_action_to_action_list(action);
    }
}


// needs refactoring
// It supposes that the first outcome is the successful one
bool Scenario::apply(const GroundedRule& grounded_rule)
{
    bool aux_success;
    float aux_reward;
    return apply(grounded_rule, aux_success, aux_reward);
}


bool Scenario::apply(const GroundedRule& grounded_rule, bool& success, float& reward, const int forced_outcome)
{
    bool result = false;
    if ((!grounded_rule.preconditions.is_satisfied(state)) || (!grounded_rule.outcomes.changes_predicates())) {
        reward = grounded_rule.goals.obtain_reward(state);
        result = false;
    } else if (forced_outcome >= 0) {
        reward = grounded_rule.goals.obtain_reward(state);
        grounded_rule.outcomes[forced_outcome].apply(state);
        success = true;
        result = true;
    } else {
        reward = grounded_rule.goals.obtain_reward(state); // do it before applying outcome

        double rand_01;
        rand_01 = ((double) rand() / (RAND_MAX));

        double accumulated_prob = 0.0;

        OutcomeList::const_iterator outcome_it = grounded_rule.outcomes.begin();
        size_t out_idx = 1;
        accumulated_prob += outcome_it->probability;
        while (accumulated_prob < rand_01) {
            ++ outcome_it;
            ++out_idx;
            accumulated_prob += outcome_it->probability;
        }
        outcome_it->apply(state);
        if (out_idx == 1) // first outcome is the successful one
            success = true;
        else
            success = false;

        result = true;
    }
    return result;
}


bool Scenario::apply_action(Symbol action)
{
    bool success;
    float reward;
    return apply_action(action, success, reward);
}

bool Scenario::apply_action(Symbol action, bool& success, float& reward, const int forced_outcome)
{
    success = false;
    bool rules_applied;
    int number_rules_applied = 0;

    RuleSet grounded_action_rules = rules.get_grounded_action_rules(action, state);
    BOOST_FOREACH(RulePtr rule_ptr, grounded_action_rules) {
        GroundedRulePtr grounded_rule_ptr = boost::make_shared<GroundedRule>(*static_cast<const GroundedRule*>(&(*rule_ptr)));
        if (apply(*grounded_rule_ptr, success, reward, forced_outcome)) {
            number_rules_applied++;
            rules_applied = true;
        }
    }

    if (number_rules_applied > 1) {
        std::cerr << "RuleSet: Rules bad defined, more than one rule was applied with action " << action << "." << std::endl;
    }

    return rules_applied;
}

