/*
 * Class having the information needed to represent an scenario (rules, state, goals, configuration, ...)
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef SCENARIO_H
#define SCENARIO_H

#include <symbolic/domain.h>
#include <symbolic/rules.h>
#include <symbolic/state.h>
#include <rexd/config_reader.h>

/*!
\class Scenario
\brief Class having all the information to define the current scenario status.

Class having all the information to define the current scenario status.\n
Includes the domain, the configuration options, the current state, rules, goals and action list.
*/
class Scenario
{
public:
    Domain domain;
    ConfigReader config_reader;

    Scenario(const Scenario& scenario);
    
    /*!
    \brief Creates an scenario.
    
    If final reward is set in config, the finish rule is read from "nid_finish_rule_path" and added to the rule set.
    This special rule is not learned, so it is read separately from the rest.
    */
    Scenario(const RuleSet& rules_, const State& state_, const GoalGroup& goals_, 
             const ConfigReader& config_reader_, 
             const std::vector<Symbol>& action_list_ = std::vector<Symbol>());

    /*!
    \brief Applies a grounded action in the current scenario.
    
    Updates the state to reflect the action changes.

    \param action action to be applied.
    \return True if action was applied, False otherwise.
    */
    bool apply_action(Symbol action);
    
    /*!
    \brief Applies a grounded action in the current scenario.
    
    Updates the state to reflect the action changes.

    \param action action to be applied.
    \retval success True if the first outcome was applied, False otherwise.
    \retval reward reward obtained by applying the action.
    \param forced_outcome index of the outcome to be applied. If -1, outcomes probabilites are used to decide which one to use.
    \return True if action was applied, False otherwise.
    */
    bool apply_action(Symbol action, bool& success, float& reward, const int forced_outcome = -1);

    /*!
    \brief Applies a grounded rule in the current scenario.
    
    Updates the state to reflect the grounded rule changes.

    \param grounded_rule grounded rule to be applied.
    \return True if rule was applied, False otherwise.
    */
    bool apply(const GroundedRule& grounded_rule);

    /*!
    \brief Applies a grounded rule in the current scenario.
    
    Updates the state to reflect the grounded rule changes.

    \param grounded_rule grounded rule to be applied.
    \retval success True if the first outcome was applied, False otherwise.
    \retval reward reward obtained by applying the rule.
    \param force_best_outcome index of the outcome to be applied. If -1, outcomes probabilites are used to decide which one to use.
    \return True if rule was applied, False otherwise.
    */
    bool apply(const GroundedRule& grounded_rule, bool& success, float& reward, const int force_best_outcome = -1);
    
    /*!
    \brief Checks if scenario is in goal state.

    \return True if scenario is in goal state.
    */
    bool is_in_goal() const {
        return goals.is_goal(state);
    };

    // gets
    inline RuleSet get_rules() const {
        return rules;
    };
    inline State get_state() const {
        return state;
    };
    inline GoalGroup get_goals() const {
        return goals;
    };
    inline std::vector<Symbol> get_action_list() const {
        return action_list;
    }

    // sets
    void set_rules(const RuleSet& rules_) {
        this->rules = rules_;
        if (config_reader.get_variable<bool>("final_reward_action")) { // add -finished precondition
            this->rules.add_predicate_to_preconditions(Predicate(MAKE_SYMBOL_FINISHED, false));
        }
        rules.index(domain);
        add_Rules_to_action_list(rules);
    };
    void set_state(const State& state_) {
        this->state = state_;
        domain.index_predicates(state);
    };
    void set_goals(const GoalGroup& goals_) {
        this->goals = goals_;
        domain.index_goals(goals);
    };
    
    void add_Rules_to_action_list(const RuleSet& rules);
    void add_action_to_action_list(Symbol action_copy);
    void add_actions_to_action_list(std::vector<Symbol> action_list);

    
protected:
    RuleSet rules;
    State state;
    GoalGroup goals;
    std::vector<Symbol> action_list;
    
    
private:
    
};

typedef boost::shared_ptr<Scenario> ScenarioPtr; //!< Shared pointer to scenario

#endif // SCENARIO_H


