/*
 * Class representing a logic symbol with parameters.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_SYMBOLS_H
#define IRI_RULE_LEARNER_SYMBOLS_H 1

#include <algorithm>
#include <vector>
#include <set>
#include <fstream>
#include <sstream>

#include <boost/optional.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>

#include "symbolic/utils.h"

// actions
#define MAKE_SYMBOL_DEAD     Symbol("dead()")
#define MAKE_SYMBOL_RETRY    Symbol("retry()")
#define MAKE_SYMBOL_HELP     Symbol("help()")
#define MAKE_SYMBOL_UNKNOWN  Symbol("unknownManipulation()")
#define MAKE_SYMBOL_TEACHER  Symbol("teacher()")
#define MAKE_SYMBOL_FINISH   Symbol("finish_action()")

// predicates
#define MAKE_SYMBOL_FINISHED Symbol("finished()")
#define MAKE_SYMBOL_STARTED  Symbol("started()")

// action_ptr
#define MAKE_SYMBOL_DEAD_PTR    boost::make_shared<Symbol>(MAKE_SYMBOL_DEAD)
#define MAKE_SYMBOL_RETRY_PTR   boost::make_shared<Symbol>(MAKE_SYMBOL_RETRY)
#define MAKE_SYMBOL_HELP_PTR    boost::make_shared<Symbol>(MAKE_SYMBOL_HELP)
#define MAKE_SYMBOL_UNKNOWN_PTR boost::make_shared<Symbol>(MAKE_SYMBOL_UNKNOWN)
#define MAKE_SYMBOL_TEACHER_PTR boost::make_shared<Symbol>(MAKE_SYMBOL_TEACHER)

static const std::string DEFAULT_PARAM_NAMES[] = {"X", "Y", "Z", "V", "W", "Q"}; //!< Default param names that are considered symbolic.
static const uint NUMBER_DEFAULT_PARAM_NAMES = 6;
static const std::string DUMMY_ACTION_NAME = "dummyAction";
static const std::string DEFAULT_ACTION_NAME = "default";
static std::vector<std::string> DEFAULT_PARAM_NAMES_VECTOR(DEFAULT_PARAM_NAMES, DEFAULT_PARAM_NAMES + NUMBER_DEFAULT_PARAM_NAMES);

/*!
\class Symbol
\brief Class representing a symbol. Symbols consist of a name and a vector of parameters.
*/
class Symbol
{
public:
    std::string name; //!< Name
    std::vector<std::string> param_names; //!< description Vector of parameters
    bool is_symbolic; //!< true if it is symbolic, false if grounded

    Symbol() {
        autodetect_symbolic();
    }
    
    /*!
    \brief Copy constructor
    */
    Symbol(const Symbol &symbol);

    /*!
    \brief constructor

    Parses a string that is assumed to have a symbol.
    Examples of the syntax supported:
     "clear(X Y)", "clear(X,Y)", "clear()".

    \param line string having a valid representation of a symbol
    */
    Symbol(const std::string &line);
    /*! 
    \brief Constructor

    Constructs the symbol.

    \param name_ name
    \param param_names_ vector of param names
    */
    Symbol(const std::string &name_, const std::vector<std::string> &param_names_);

    /*!
    \brief Checks if symbol is symbolic, and updates its is_symbolic member

    If one of the default param names (X, Y, Z, ...) is found,
    is_symbolic is set to true.
    */
    void autodetect_symbolic();
    
    /*!
    \brief Check if symbol is a special symbol
    
    Checks if symbol is a special symbol: dead(), retry(), help(), unknown(), finish(), teacher().

    \return true if it is special
    */
    bool is_special() const;
    
    /*!
    \brief Check if a parameter with the given name exists

    \return true if there is a parameter with the given name
    */
    bool has_param_with_name(const std::string &param) const;
    
    /*!
    \brief Check if parameter with all the given name exist

    \return true if all of them exist
    */
    bool has_param_with_all_names(const std::vector<std::string> &params) const;
    
    /*!
    \brief Check if symbol is initialized

    Check if symbol is initialized, that is that its name is not empty.
    
    \return true if initialized, false if empty
    */
    bool is_valid() const;
    
    /*!
    \brief Gets the index of a param
    
    \param param name of the param to look for its index
    \return true if it is special
    */
    int get_param_name_idx(const std::string &param) const;

    /*!
    \brief Makes the symbol symbolic
    
    Changes the params to be the default symbolic params: X, Y, Z, V, W, Q.
    */
    void make_symbolic();
    
    /*!
    \brief Grounds the symbol with the given parameters
    */
    void ground(std::vector<std::string> variables);

    /*!
    \brief Makes parameter-less symbol appending parameters to the name
    */
    void PPDDL_grounding();
    
    /*!
    \brief Writes PPDDL grounded predicates without parameters
    */
    void write_PPDDL(std::ostream &out) const;

    /*!
    \brief Convert the symbol to string
    
    \return string representing the symbol
    */
    const std::string to_string() const {
        return boost::lexical_cast<std::string>(*this);
    }
    
    /*!
    \brief Compares if not grounded versions of the symbol are equal
    
    Compares the name and the number of parameters, but not the parameters content
    
    \return true if not grounded symbols are equal
    */
    bool symbolic_compare(const Symbol &s) const;
    
    /*! \brief Operator == */
    friend bool operator== (const Symbol &s1, const Symbol &s2);
    /*! \brief Operator != */
    friend bool operator!= (const Symbol &s1, const Symbol &s2);
    /*! \brief Operator <= */
    friend bool operator<(const Symbol& l, const Symbol& r);
    /*! \brief Operator \<\< */
    friend std::ostream& operator<<(std::ostream &out, const Symbol &s);
    
private:
    static std::string get_symbol_name_from_string(const std::string &line);
    static std::vector< std::string > get_symbol_param_names_from_string(const std::string &line);
};

/*! Optional symbol, it is false if it doesn't exist */
typedef boost::optional<Symbol> SymbolOptional;
/*! Shared pointer to symbol */
typedef boost::shared_ptr<const Symbol> SymbolPtr;
/*! Shared pointer to const symbol */
typedef boost::shared_ptr<const Symbol> SymbolConstPtr;

// get data from string
// Symbol get_symbol_from_string(std::string line);


#endif
