#include <symbolic/transitions.h>


std::ostream& operator<<(std::ostream &out, const TransitionGroup &t)
{
    for (TransitionGroup::const_iterator transition = t.begin() ; transition != t.end(); ++transition) {
        Transition transition_copy(*transition);
        transition_copy.next_state.add_positive_differences(transition_copy.prev_state);
        out << transition_copy;
    }
    return out;
}

std::ostream& operator<<(std::ostream &out, const Transition &t)
{
    out << t.prev_state << std::endl;
    out << t.action << std::endl;
    out << t.next_state << std::endl;
    out << t.reward << std::endl;
    return out;
}

TransitionGroup read_transitions_from_file(const std::string& file_path)
{
    TransitionGroup transitions;
    std::ifstream transitions_file(file_path.c_str());
    if (transitions_file.is_open()) {
        std::string line_prev, line_action, line_next, line_reward;

        while (transitions_file.good()) {
            getline(transitions_file, line_prev);
            getline(transitions_file, line_action);
            getline(transitions_file, line_next);
            getline(transitions_file, line_reward);

            // avoid adding twice last transition
            if (transitions_file.good()) {
                try {
                    float transition_reward = boost::lexical_cast<float>(line_reward);
                    Transition transition(line_prev,
                                          Symbol(line_action),
                                          line_next,
                                          transition_reward);
//                                           boost::lexical_cast<float>(line_reward) );
                    transitions.push_back(transition);
                } catch (boost::bad_lexical_cast &) {
                    LOG(WARNING) << "Error reading transition " << transitions.size() + 1;
                }
            }
        }
    } else {
        LOG(ERROR) << "Couldn't read transitions from " << file_path;
    }
    transitions_file.close();
    return transitions;
}

/**
 * Transition
 */

void Transition::remove_unused_predicates()
{
    prev_state.remove_variables_not_present(action.param_names);
    next_state.remove_variables_not_present(action.param_names);
}

void Transition::remove_negated_predicates()
{
    prev_state.remove_negated_predicates();
    next_state.remove_negated_predicates();
}

bool Transition::nothing_changed() const
{
    return prev_state == next_state;
}

PredicateGroup Transition::get_changes() const
{
    return prev_state.get_differences(next_state);
}


/**
 * TransitionGroup
 */

void TransitionGroup::write_transitions_to_file(const std::string& file_path) const
{
    std::ofstream transitions_file(file_path.c_str());
    if (transitions_file.is_open()) {
        transitions_file << *this;
    } else
        std::cerr << "Couldn't write transitions to " << file_path << std::endl;
    transitions_file.close();
}

void TransitionGroup::remove_unused_predicates()
{
    for (TransitionGroup::iterator transition = this->begin() ; transition != this->end(); ++transition) {
        transition->remove_unused_predicates();
    }
}

void TransitionGroup::remove_negated_predicates()
{
    for (TransitionGroup::iterator transition = this->begin() ; transition != this->end(); ++transition) {
        transition->remove_negated_predicates();
    }
}

TransitionGroup TransitionGroup::get_action_list_transitions(std::vector<std::string> actions) const
{
    TransitionGroup important_transitions;

    // remove duplicated actions to don't duplicate transitions
    std::sort(actions.begin(), actions.end());
    std::vector< std::string >::iterator it = std::unique(actions.begin(), actions.end());
    actions.resize(std::distance(actions.begin(), it));

    for (std::vector<std::string>::iterator action = actions.begin(); action != actions.end(); ++action) {
        TransitionGroup aux_transitions = get_action_transitions(*action);
        important_transitions.insert(important_transitions.end(), aux_transitions.begin(), aux_transitions.end());
    }

    return important_transitions;
}

TransitionGroup TransitionGroup::get_action_transitions(const std::string& action_name) const
{
    TransitionGroup important_transitions;
    for (TransitionGroup::const_iterator transition = this->begin() ; transition != this->end(); ++transition) {
        if (transition->action.name == action_name) {
            important_transitions.push_back(Transition(*transition));
        }
    }

    return important_transitions;
}

float TransitionGroup::get_action_reward(const std::string& action_name) const
{
    float reward = 0;
    bool found = false;
    TransitionGroup::const_iterator transition_it = this->begin();
    while (!found && (transition_it != this->end())) {
        if (transition_it->action.name == action_name) {
            reward = transition_it->reward;
            found = true;
        }
        ++transition_it;
    }
    return reward;
}

