/*
 * Clases representing transitions (previous experiences).
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_TRANSITIONS_H
#define IRI_RULE_LEARNER_TRANSITIONS_H 1

#include <vector>
#include <fstream>
#include <symbolic/symbols.h>
#include <symbolic/predicates.h>


/*!
\class Transition
\brief Class representing previous experiences of the form: state_t, action_t, state_t+1, obtained reward
*/
class Transition
{
public:
    PredicateGroup prev_state, next_state;
    Symbol action;
    float reward;

    Transition(const Transition& transition) :
        prev_state(transition.prev_state),
        next_state(transition.next_state),
        action(transition.action),
        reward(transition.reward)
    { }

    Transition(PredicateGroup prev_state_, Symbol action_, PredicateGroup next_state_, float reward_) :
        prev_state(prev_state_),
        next_state(next_state_),
        action(action_),
        reward(reward_)
    { }

    // get info
    PredicateGroup get_prev_state() const {
        return prev_state;
    }
    PredicateGroup get_post_state() const {
        return next_state;
    }
    PredicateGroup get_changes() const;


    bool nothing_changed() const;

    /**
     * \brief Removes predicates whose parameters are not in the action parameters
     */
    void remove_unused_predicates();
    
    /**
     * \brief As we are using closed-world assumption, negated predicates don't have to be specified directly
     */
    void remove_negated_predicates();

    friend std::ostream& operator<<(std::ostream &out, const Transition &s);
};

typedef std::vector<Transition> TransitionList;


/*!
\class TransitionGroup
\brief Vector of transitions.
*/
class TransitionGroup : public TransitionList
{
public:
    void remove_unused_predicates();
    void remove_negated_predicates();
    
    void write_transitions_to_file(const std::string& file_path) const;

    /**
     * \brief Get action reward by looking at its transitions. TODO: not working.
     * 
     * \param action_name name of the action whose reward is to be obtained.
     * \return action reward
     */
    float get_action_reward(const std::string& action_name) const;

    /**
     * \brief Get transitions that include any of the actions in the given list.
     * 
     * \param actions list of actions whose transitions are returned
     * \return list of transitions with the given actions
     */
    TransitionGroup get_action_list_transitions(std::vector<std::string> actions) const;
    
    /**
     * \brief Get transitions that include the given action.
     * 
     * \param action_name name of the action whose transitions are returned
     * \return list of transitions with the given action name
     */
    TransitionGroup get_action_transitions(const std::string& action_name) const;

    friend std::ostream& operator<<(std::ostream &out, const TransitionGroup &s);
};

typedef boost::shared_ptr<TransitionGroup> TransitionGroupPtr;
typedef boost::shared_ptr<const TransitionGroup> TransitionGroupConstPtr;

// other


// write/read to file
TransitionGroup read_transitions_from_file(const std::string& file_path);

#endif
