/*
 * Useful and general functions used in the code.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_UTILS_H
#define IRI_RULE_LEARNER_UTILS_H 1

#include <algorithm>
#include <vector>
#include <set>
#include <fstream>
#include <sstream>
#include <errno.h>
#include <iostream>
#include <iomanip>
#include <boost/shared_ptr.hpp>
#include <boost/foreach.hpp>
#include "boost/filesystem/path.hpp"
#include "boost/filesystem/operations.hpp"
#include <boost/functional/hash.hpp>

#include <symbolic/easylogging++.h>

#define START_COLOR_RED    "\033[1;31m"
#define START_COLOR_GREEN  "\033[1;32m"
#define START_COLOR_YELLOW "\033[1;33m"
#define START_COLOR_BLUE   "\033[1;34m"
#define START_COLOR_PURPLE "\033[1;35m"
#define START_COLOR_CYAN   "\033[1;36m"
#define START_COLOR_WHITE  "\033[1;37m"
#define END_COLOR          "\033[0m"

namespace utils
{
    // utils
    bool is_number(const std::string& s);
    bool compare_vector_strings(std::vector<std::string> v1, std::vector<std::string> v2);
    bool compare_ordered_vector_strings(std::vector<std::string> v1, std::vector<std::string> v2);
    char choose_variable_separator(std::string str);
    std::string trim(const std::string& str, const std::string& whitespace = " \t");
    std::string remove_spaces(std::string str);
    std::string get_file_contents(const std::string template_file);
    std::string get_file_contents(const std::string template_file, const std::string init_keyword , const std::string end_keyword);
    void write_string_to_file(const std::string& file_path, const std::string& out_str, bool quiet = true);
    
    /**
     * \brief Converts string to char*
     * 
     * The difference with the c_str() method is that the returned char* is not const.
     */
    char* string_to_char(const std::string str);
    
    std::string get_parsed_time_from_seconds(const ulong total_seconds);
    uint get_socket_from_current_dir_hash();
    void wait_for_enter_key();

    void print_std_vector_string(std::vector<std::string> strings, std::ostream &out);

    void init_loggers(std::string config_path);

    template <typename T>
    void remove_duplicates(std::vector<T>& vec)
    {
        std::sort(vec.begin(), vec.end());
        vec.erase(std::unique(vec.begin(), vec.end()), vec.end());
    }

    template<class T>
    bool belongs_to_vector(const T& obj, const std::vector<T>& v)
    {
        return (std::find(v.begin(), v.end(), obj) != v.end());
    }

    // TODO HACK DEPRECATED
    // To be removed once boost > 1.48 is the default with new boost::timer
    // #if (BOOST_VERSION < 104800)
    struct timeval start_timer();
    double get_elapsed_time(struct timeval& timer_start);

} // namespace utils


typedef std::vector<std::string> vector_string;

namespace std
{
template<class T>
ostream & operator<<(ostream &stream, const vector<T> &vec)
{
    stream << "[";
    for (size_t i = 0; i < vec.size(); ++i) {
        if (i != 0)
            stream << ", ";
        stream << vec[i];
    }
    stream << "]";
    return stream;
}
}

namespace std
{
template<class T>
vector<T> vector_intersection(vector<T> v1, vector<T> v2)
{
//         int first[] = {5,10,15,20,25};
//         int second[] = {50,40,30,20,10};
    std::vector<T> v(v1.size() + v2.size());                      // 0  0  0  0  0  0  0  0  0  0
    typename std::vector<T>::iterator it;

    std::sort(v1.begin(), v1.end());     //  5 10 15 20 25
    std::sort(v2.begin(), v2.end());   // 10 20 30 40 50

    it = std::set_intersection(v1.begin(), v1.end(), v2.begin(), v2.end(), v.begin());
    // 10 20 0  0  0  0  0  0  0  0
    v.resize(it - v.begin());                    // 10 20
    return v;
}
}

template<class T>
bool operator<(const boost::shared_ptr<T>& l, const boost::shared_ptr<T>& r)
{
    return (*l < *r);
}


#endif
