/*
 * <one line to give the program's name and a brief idea of what it does.>
 * Copyright (C) 2014  <copyright holder> <email>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "high_level_planner.h"


HighLevelPlanner::HighLevelPlanner(const ConfigReader& config_reader, ofstream* logging_file_ptr_):
    subgoals_disabled(false)
{
    logging_file_ptr = logging_file_ptr_;
    planner = PlannerFactory::get_standard_planner(config_reader);
}


SymbolOptional HighLevelPlanner::exploit_or_explore(const Scenario& scenario, 
                                                    const TransitionGroup& transitions, 
                                                    PredicateGroup& excuse,
                                                    std::vector<std::string>& actions_requiring_excuse)
{
    SymbolOptional planning_res;
    ExcuseGenerator excuse_generator = ExcuseGenerator(scenario);
    expected_reward = 0.0;
    transition_context = "None";
    already_confirmed_as_safe = false; // TODO remove
    
    if (!scenario.config_reader.get_variable<bool>("use_rmax")) {
        planning_res = explore(scenario, transitions, false);
    }
    if (!planning_res) {
        planning_res = standard_plan(scenario, transitions);
        check_teacher_demonstration_requested(planning_res);
    }
    if (!planning_res) {
        planning_res = planning_with_alternative_rules(scenario, transitions, excuse_generator);
        check_teacher_demonstration_requested(planning_res);
    }
    if (!planning_res) {
        planning_res = planning_with_subgoals(scenario, excuse_generator);
        check_teacher_demonstration_requested(planning_res);
    }
    
    if (planning_res) {
        confirm_dangerous_actions(scenario, planning_res);
    }
    else {
        excuse = excuse_generator.get_excuse();
        actions_requiring_excuse = get_actions_requiring_excuse(scenario, excuse_generator);
    }
    
    assert(transition_context != "None");
    return planning_res;
}


SymbolOptional HighLevelPlanner::explore(const Scenario& scenario, const TransitionGroup& transitions, bool forced)
{
    int DEBUG = 0;
    SymbolOptional planning_res;

    if (scenario.config_reader.get_variable<bool>("exploration")) {
        Explorer explorer(scenario, boost::make_shared<TransitionGroup>(transitions));
        // TODO separate dangerous_actions and confirm_dangerous_actions
        if ( forced || (!explorer.is_state_known()) ) {
            planning_res = explorer.select_action_for_exploration(forced, scenario.config_reader.get_variable<uint>("exploration_known_threshold"));
            if (planning_res) {
                transition_context = "Exploring";
                (*logging_file_ptr) << "Explore: " << *planning_res << std::endl;
                if (DEBUG > 0)
                    LOG(INFO) << "Selected exploring action " << boost::lexical_cast<std::string>(planning_res->to_string());
            }
        }
        else {
            if (DEBUG > 0)
                LOG(INFO) << "State is already known.";
        }
    }

    return planning_res;
}


SymbolOptional HighLevelPlanner::standard_plan(const Scenario& scenario, const TransitionGroup& transitions)
{
    SymbolOptional planning_res;
    transition_context = "Exploiting";
    RexdScenario rexd_scenario(scenario);
    if (rexd_scenario.config_reader.get_variable<bool>("use_rmax")) {
        rexd_scenario.set_rmax(boost::make_shared<TransitionGroup>(transitions));
        rexd_scenario.set_vmin();
        planning_res = planner->plan(rexd_scenario, expected_reward, true, true);
        if (scenario.config_reader.get_variable<bool>("use_rmax") 
        && (expected_reward > scenario.config_reader.get_variable<float>("rmax_reward"))
        && (*planning_res != MAKE_SYMBOL_TEACHER) ) {
            transition_context = "Exploring";
            avoid_exploration_of_dangerous_rules(scenario, planning_res);
        }
    }
    else {
        planning_res = planner->plan(rexd_scenario, expected_reward, true, true);
    }
    
    return planning_res;
}


// TODO
void HighLevelPlanner::avoid_exploration_of_dangerous_rules(const Scenario& scenario, SymbolOptional& planning_res)
{
    assert(planning_res);
    // if dangerous -> main planning step without exploration
    if (scenario.config_reader.get_variable<bool>("confirm_dangerous_actions")
        && scenario.get_rules().is_dangerous(*planning_res, scenario.get_state()) ) {
        float prob_success = planner->get_expected_action_success_probability(scenario, *planning_res); // TODO this is called twice if prob = 0.0
        if (prob_success < 1.0) {
            *logging_file_ptr << "Ignoring exploring " << *planning_res << " and retry without rmax." << std::endl;
            CINFO("planning") << "Not exploring " << *planning_res << " as it is dangerous. Replanning without RMAX.";
            RexdScenario rexd_scenario(scenario);
            rexd_scenario.set_vmin();
            planning_res = planner->plan(scenario, expected_reward, true, true);
            transition_context = "Exploiting";
        }
        else {
            already_confirmed_as_safe = true;
        }
    }
}


// TODO
void HighLevelPlanner::confirm_dangerous_actions(const Scenario& scenario, SymbolOptional& planning_res)
{
    assert(planning_res);
    // Check for dangerous actions
    if (scenario.config_reader.get_variable<bool>("confirm_dangerous_actions")) {
        if ((!already_confirmed_as_safe) && (scenario.get_rules().is_dangerous(*planning_res, scenario.get_state()))) {
            float prob_success = planner->get_expected_action_success_probability(scenario, *planning_res);
            if (prob_success < 1.0) {
                LOG(INFO) << "Dangerous action which may lead to a dead end with probability " << (1.0 - prob_success) << " detected: " << *planning_res;
                *logging_file_ptr << "Confirmation request: " << *planning_res << std::endl;
                transition_context = "ExploitingDangerous";
//  TODO TODO TODO               planning_res = ask_confirmation_dangerous_action(state, planning_res, prob_success); // transition_context set inside if teacher changes action
            }
        }
    }
}

void HighLevelPlanner::check_teacher_demonstration_requested(SymbolOptional& planning_res)
{
    if ((planning_res) && (*planning_res == MAKE_SYMBOL_TEACHER)) {
        planning_res = SymbolOptional();
        transition_context = "Teacher";
    }
}


SymbolOptional HighLevelPlanner::planning_with_alternative_rules(const Scenario& scenario, const TransitionGroup& transitions, ExcuseGenerator& excuse_generator)
{
    if (!scenario.config_reader.get_variable<bool>("rule_analysis_equivalent_rules")) {
        return SymbolOptional();
    }
    
    PredicateGroup excuse = excuse_generator.get_excuse();
    SymbolOptional planning_res;
    RuleSetPtr best_ruleset;

    std::vector<RuleSet> extra_rulesets = excuse_generator.get_rules_that_overcome_excuse(scenario.get_rules(), excuse, transitions);

    // Plan with newer rulesets
    if (!extra_rulesets.empty()) {
        CINFO("excuses") << "Testing " << extra_rulesets.size() << " rules: " << NOENDL;
    }
    double min_score = -10000;
    for (std::vector<RuleSet>::iterator it_ruleset= extra_rulesets.begin(); it_ruleset!=extra_rulesets.end(); ++it_ruleset) {
        float score;
        RexdScenario alt_scenario(scenario);
        alt_scenario.set_rules(*it_ruleset);
        SymbolOptional new_res = planner->plan(alt_scenario, score, false, false);
        if (new_res && (new_res->name != "teacher") && (score > min_score)) {
            min_score = score;
            planning_res = new_res;
            best_ruleset = boost::make_shared<RuleSet>(*it_ruleset);
            CINFO("excuses") << "*" << NOENDL;
            CINFO("excuses");
            CINFO("excuses") << "Better ruleset found! Result is " << *new_res;
        }
        else {
            CINFO("excuses") << "." << NOENDL;
        }
    }
    if (!extra_rulesets.empty())
        CINFO("excuses");

    // if found good rules -> save them
    if (planning_res) {
        this->expected_reward = min_score;
//         *rules_ptr = *best_ruleset; // TODO THINK: Is it really needed? We are learning that action anyway...
        best_ruleset->write_rules_to_file(scenario.config_reader.get_path("nid_rules_path"));
        transition_context = "ExploitingAlternatives"; 
    }

    return planning_res;
}

SymbolOptional HighLevelPlanner::planning_with_subgoals(const Scenario& scenario, ExcuseGenerator& excuse_generator)
{
    SymbolOptional planning_res;
    
    if (scenario.config_reader.get_variable<bool>("rule_analysis_subgoals")) {
        if (scenario.config_reader.get_variable<bool>("use_vmin")) {
            planning_res = planning_with_subgoals_vmin(scenario);
        }
        else if (scenario.config_reader.get_variable<bool>("rule_analysis")) {
            planning_res = plan_with_subgoals_excuses(scenario, excuse_generator);
        }
        transition_context = "ExploitingSubgoals";
    }
    
    return planning_res;
}

SymbolOptional HighLevelPlanner::planning_with_subgoals_vmin(const Scenario& scenario)
{
    SymbolOptional planning_res;
    
    CINFO("planning") << "Decreasing Vmin by " << scenario.config_reader.get_variable<float>("auto_experiments_increasing_vmin_quantity");
    RexdScenario new_scenario(scenario);
    new_scenario.set_vmin(-1 * scenario.config_reader.get_variable<float>("auto_experiments_increasing_vmin_quantity"));
    planning_res = planner->plan(new_scenario, expected_reward, true, true);
    if (planning_res && (*planning_res == MAKE_SYMBOL_FINISH)) {
        // Subgoals are for approaching the goal, so
        // don't finish with subgoals!
        planning_res = SymbolOptional();
    }
    return planning_res;
}

SymbolOptional HighLevelPlanner::plan_with_subgoals_excuses(const Scenario& scenario, ExcuseGenerator& excuse_generator)
{
    SymbolOptional planning_res;
    
    // this type of subgoals require rule_analysis
    PredicateGroup excuse = excuse_generator.get_excuse();
    CINFO("planning") << "Planning using subgoals with excuse " << excuse;
    if (scenario.get_goals().size() > 1) {
        CWARNING("planning") << "Goals > 1 not working with subgoals!!!";
    }
    Goal new_goal = excuse_generator.get_subgoal(scenario.get_goals().get_predicates(), excuse);
    GoalGroup new_goals;
    new_goals.push_back(new_goal);
    CINFO("planning") << "New goal is " << new_goal;
    Scenario new_scenario(scenario);
    new_scenario.set_goals(new_goals);
    planning_res = planner->plan(new_scenario, expected_reward, false);
    return planning_res;
}


std::vector<std::string> HighLevelPlanner::get_actions_requiring_excuse(const Scenario& scenario, ExcuseGenerator& excuse_generator)
{
    std::vector<std::string> actions_requiring_excuse;
    
    PredicateGroup excuse = excuse_generator.get_excuse();
    if (!excuse.empty()) {
        RuleSet rules_requiring_excuse = scenario.get_rules().find_rules_with_a_precondition(excuse);
        rules_requiring_excuse.remove_nop_rules();
        std::vector<std::string> actions_requiring_excuse = rules_requiring_excuse.get_action_names();
    }
    
    return actions_requiring_excuse;
}
