/*
 * High level planner to select actions in a RL setup.
 * Copyright (C) 2014 David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef HIGHLEVELPLANNER_H
#define HIGHLEVELPLANNER_H

#include <rexd/planner.h>
#include <rexd/excuses.h>
#include <rexd/explorer.h>
#include <rexd/rexd_scenario.h>

/*!
\class HighLevelPlanner
\brief High level planning methods. Uses a planner in combination with exploration and excuses.

It does all the planning related to a reinforcement learning algorithm, including the exploration vs exploitation. Also planning with the help of excuses is included.
*/
class HighLevelPlanner
{
public:
    /**
     * \brief Constructor
     * 
     * \param config_reader configuration.
     * \param logging_file_ptr_ pointer to ofstream for logging to a file the planner results.
     */
    HighLevelPlanner(const ConfigReader& config_reader, ofstream* logging_file_ptr_);
    
    /**
     * \brief Decides between exploration and exploitation. Uses excuses to plan if enabled.
     * 
     * It does the following steps until an action is obtained:
     * - Explores if state is unknown.
     * - Plans.
     * - Plans with alternative rules (based on excuses).
     * - Plans with subgoals (based on excuses / or decreases Vmin).
     * - Confirms dangerous actions.
     * - Returns no plan.
     * 
     * \param scenario scenario in which to plan.
     * \param transitions used to see if state-action pairs are known, and validate rule alternatives.
     * \retval excuse excuse obtained if exploration and standard planning failed.
     * \retval actions_requiring_excuse actions requiring excuse if any (used to display information to the teacher).
     * \return Planned action if any.
     */
    SymbolOptional exploit_or_explore(const Scenario& scenario, 
                                      const TransitionGroup& transitions, 
                                      PredicateGroup& excuse,
                                      std::vector<std::string>& actions_requiring_excuse);
    
    /**
     * \brief Looks for state-action pairs not known already. If any are found, a exploration action is selected.
     * 
     * \param scenario scenario
     * \param transitions previous experiences to check if state-action pairs are known.
     * \param forced if true a exploration action is selected even if state is known.
     * \return Action selected to be explored if any.
     */
    SymbolOptional explore(const Scenario& scenario, const TransitionGroup& transitions, bool forced);
    
    /**
     * \brief Returns a pointer to the task planner used internally.
     */
    inline PlannerPtr get_planner_ptr() { return planner; };
    
private:
    PlannerPtr planner;
    ofstream* logging_file_ptr;
    
    float expected_reward;
    bool already_confirmed_as_safe; // TODO remove
    bool subgoals_disabled;
    
    std::string transition_context;
    
    /**
     * \brief Plans using the standard task planner.
     * 
     * \param scenario scenario.
     * \param transitions used for R-MAX to know if state-action pairs are known.
     * \return Planned action if any.
     */
    SymbolOptional standard_plan(const Scenario& scenario, const TransitionGroup& transitions);
    
    // TODO redo after reworking dangerous rules
    void avoid_exploration_of_dangerous_rules(const Scenario& scenario, SymbolOptional& planning_res);
    
    // TODO redo after reworking dangerous rules
    void confirm_dangerous_actions(const Scenario& scenario, SymbolOptional& planning_res);
    
    /**
     * \brief Plans with excuses to generate alternative rule sets and to try to obtain an alternative plan.
     * 
     * Using an excuse, searches for alternative equivalent rule sets that get a similar score using Pasula's function.
     */
    void check_teacher_demonstration_requested(SymbolOptional& planning_res);
    
    /**
     * \brief Plans with excuses to generate alternative rule sets and to try to obtain an alternative plan.
     * 
     * Using an excuse, searches for alternative equivalent rule sets that get a similar score using Pasula's function.
     */
    SymbolOptional planning_with_alternative_rules(const Scenario& scenario, const TransitionGroup& transitions, ExcuseGenerator& excuse_generator);
    
    /**
     * \brief Plans with subgoals depending on the algorithm.
     * 
     * Possible algorithms:
     *  - E3 - a set of subgoals (obtained with the excuse) as the reward state for the planner.
     *  - V-MIN - decreases temporary Vmin.
     */
    SymbolOptional planning_with_subgoals(const Scenario& scenario, ExcuseGenerator& excuse_generator);
    SymbolOptional planning_with_subgoals_vmin(const Scenario& scenario);
    SymbolOptional plan_with_subgoals_excuses(const Scenario& scenario, ExcuseGenerator& excuse_generator);
    
    /**
     * \brief Gets action that have the excuse predicates in their preconditions.
     */
    std::vector<std::string> get_actions_requiring_excuse(const Scenario& scenario, ExcuseGenerator& excuse_generator);
};

typedef boost::shared_ptr<HighLevelPlanner> HighLevelPlannerPtr;


#endif // HIGHLEVELPLANNER_H
